unit CompleteForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, TntStdCtrls, NotePadView, TntClasses, TntForms;

type
  TFormComplete = class(TTntForm)
    TntListBox1: TTntListBox;
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    ListProc1: TWndMethod;
    FNotePad: TNotePad;
    FList: TTntStringList;
    FCompString: WideString;
    FClosed: boolean;
    FCompleteStart: integer;
    FExt: WideString;
    F2D: boolean;
    procedure ListProc2(var Msg:TMessage);
    procedure ListChar(var Msg: TWMKey);
    procedure ListKeyDown(var Msg: TWMKey);
    procedure CompleteCancel;
    procedure CompleteEnd(const Value: WideString);
    function InitList: boolean;
    function InitBox(Auto: boolean): boolean;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    procedure CompleteStart(Value: TNotePad; const fn: WideString);
    { Public 錾 }
  end;

var
  FormComplete: TFormComplete;

implementation

uses DM, WideStrUtils, wsFileOp, TextFileEncoding;

{$R *.dfm}

{ TFormComplete }

procedure TFormComplete.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.WndParent := ManaMain.MMGetHandle;
  Params.Style := WS_DLGFRAME or WS_POPUP;
end;

procedure TFormComplete.ListProc2(var Msg: TMessage);
var
  km: TWMKey;
begin
  case Msg.Msg of
    WM_CHAR:begin
      km := TWMKey(Msg);
      ListChar(km);
      Msg := TMessage(km);
    end;
    WM_KEYDOWN: begin
      km := TWMKey(Msg);
      ListKeyDown(km);
      Msg := TMessage(km);
    end;
  end;
  ListProc1(Msg);
  case Msg.Msg of
    WM_SETFOCUS: begin
      if TntListBox1.Focused then
        FNotePad.CompleteFocusStart;
    end;
    WM_KILLFOCUS: begin
      FNotePad.CompleteFocusCancel;
      if FClosed then exit;
      CompleteCancel;
    end;
  end;
end;

procedure TFormComplete.FormCreate(Sender: TObject);
begin
  ListProc1 := TntListBox1.WindowProc;
  TntListBox1.WindowProc := ListProc2;
  FList := TTntStringList.Create;
end;

procedure TFormComplete.CompleteStart(Value: TNotePad; const fn: WideString);
var
  P: TPoint;
begin
  FNotePad := Value;
  FExt := ExtractFileExtW(fn);
  if (not InitList) or (not InitBox(true)) then exit;
  FClosed := false;

  P := FNotePad.ClientToScreen(FNotePad.CompletePoint);
  Left := P.X;
  Top := P.Y;
  Show;
end;

procedure TFormComplete.ListChar(var Msg: TWMKey);
var
  ws: WideString;
begin
  case Msg.CharCode of
    VK_ESCAPE: CompleteCancel;
    VK_RETURN: begin
      if TntListBox1.ItemIndex < 0 then exit;
      ws := TntListBox1.Items[TntListBox1.ItemIndex];
      if Length(ws) = 0 then exit;
      CompleteCancel;
      CompleteEnd(ws);
    end;
    VK_BACK: begin
      if Length(FCompString) = 0 then begin
      end else begin
        SetLength(FCompString, Length(FCompString) - 1);
        FNotePad.InputKeyChar(Msg);
        InitBox(false);
      end;
      Msg.CharCode := 0;
    end;
    $002D:begin
      if F2D then begin
        if Length(FCompString) = 0 then begin
        end else begin
          FCompString := FCompString + WideString(WideChar(Msg.CharCode));
          FNotePad.InputKeyChar(Msg);
          InitBox(false);
          Msg.CharCode := 0;
        end;
      end;
    end;
    $0030..$0039:begin
      if Length(FCompString) = 0 then begin
      end else begin
        FCompString := FCompString + WideString(WideChar(Msg.CharCode));
        FNotePad.InputKeyChar(Msg);
        InitBox(false);
        Msg.CharCode := 0;
      end;
    end;
    $0041..$005A,
    $005F,
    $0061..$007A:begin
      FCompString := FCompString + WideString(WideChar(Msg.CharCode));
      FNotePad.InputKeyChar(Msg);
      InitBox(false);
      Msg.CharCode := 0;
    end;
  end;
end;

procedure TFormComplete.ListKeyDown(var Msg: TWMKey);
var
  len: integer;
  C: WideChar;
begin
  case Msg.CharCode of
    VK_LEFT:begin
      if Length(FCompString) = 0 then begin
      end else begin
        SetLength(FCompString, Length(FCompString) - 1);
        FNotePad.InputKey(Msg);
        InitBox(false);
      end;
      Msg.CharCode := 0;
    end;
    VK_RIGHT:begin
      len := Length(FCompString);
      if FCompleteStart + len < FNotePad.TextLength then begin
        C := (FNotePad.ToPointer(FCompleteStart + len))^;
        case C of
          '-':begin
            if F2D then begin
              if Length(FCompString) = 0 then begin
              end else begin
                FNotePad.InputKey(Msg);
                if FNotePad.SelectionStart > FCompleteStart + len then
                  FCompString := FCompString + WideString(C);
                InitBox(false);
              end;
            end;
          end;
          '0'..'9':begin
            if Length(FCompString) = 0 then begin
            end else begin
              FNotePad.InputKey(Msg);
              if FNotePad.SelectionStart > FCompleteStart + len then
                FCompString := FCompString + WideString(C);
              InitBox(false);
            end;
          end;
          'A'..'Z', '_', 'a'..'z':begin
            FNotePad.InputKey(Msg);
            if FNotePad.SelectionStart > FCompleteStart + len then
              FCompString := FCompString + WideString(C);
            InitBox(false);
          end;
        end;//case
      end;
      Msg.CharCode := 0;
    end;
  end;
end;

function TFormComplete.InitList: boolean;
const
  ds: WideString = '';

var
  P, P1: PWChar;
  i, j, len, len1: integer;
  ws: WideString;
  fc: TFileCode;
  b: boolean;
  ws1, ws2: WideString;
  procedure proc1;
  var
    l: integer;
  begin
    if P1 <> PWChar(ds) then begin
      if (i - 1) <> FNotePad.SelectionStart then begin
        len1 := P - P1;
        SetLength(ws, len1);
        for l := 1 to len1 do ws[l] := (P1 + l - 1)^;
        FList.Add(ws);
      end;
      P1 := PWChar(ds);
    end;
  end;
begin
  Result := false;
  FList.Clear;
  ws := '';
  ws1 := ExtractFilePathW(ParamStrW(0)) + 'ext\';
  ws2 := '.txt';
  if Length(FExt) > 1 then begin
    FExt := ws1 + Copy(FExt, 2, Length(FExt) - 1) + ws2;
  end;
  fc := fcANSI;
  if FileExists2(FExt) then begin
    ws := LoadFromFile(FExt, fc, false, b);
  end;
  for i := 0 to FNotePad.CurrentMode.ExtCount - 1 do begin
    if Length(ws) > 0 then break;
    FExt := FNotePad.CurrentMode.Exts[i];
    FExt := ws1 + Copy(FExt, 2, Length(FExt) - 1) + ws2;
    if FileExists2(FExt) then ws := LoadFromFile(FExt, fc, false, b);
  end;

  FList.Text := ws;
  FList.Sorted := true;
  F2D := false;
  for i := 1 to Length(ws) do begin
    case ws[i] of
      '-': begin
        F2D := true;
        break;
      end;
    end;
  end;
  //I
  P := FNotePad.ToPointer(0);
  P1 := PWChar(ds);
  ws := '';
  i := 1;
  len := FNotePad.TextLength;
  while i <= len do begin
    case P^ of
      '-':begin
        if not F2D then proc1;
      end;
      '0'..'9': begin
      end;
      'A'..'Z','_','a'..'z':begin
        if P1 = PWChar(ds) then P1 := P;
      end;
      else proc1;
    end;
    inc(i);
    inc(P);
  end;
  proc1;
  if FList.Count = 0 then exit;
  P := FNotePad.ToPointer(0);
  i := FNotePad.SelectionStart + FNotePad.SelectionLength;
  j := i;
  while j > 0 do begin
    case (p + j - 1)^ of
      '-': if F2D then dec(j) else break;
      '0'..'9','A'..'Z','_','a'..'z': dec(j);
      else break;
    end;
  end;
  while j < i do begin
    case (p + j)^ of
      '-', '0'..'9': inc(j);
      else break;
    end;
  end;
  FCompleteStart := j;
  len := i - j;
  SetLength(FCompString, len);
  for i := 1 to len do FCompString[i] := (p + j + i - 1)^;
  Result := true;
end;

procedure TFormComplete.FormDestroy(Sender: TObject);
begin
  if Assigned(FList) then FList.Free;
end;

function TFormComplete.InitBox(Auto: boolean): boolean;
var
  i, len: integer;
  AList: TTntStringList;
  ws: WideString;
begin
  Result := true;
  AList := TTntStringList.Create;
  try
    AList.Duplicates := dupIgnore;
    AList.Sorted := true;
    len := Length(FCompString);
    if len > 0 then ws := CharLowerW(PWChar(FCompString));
    for i := 0 to FList.Count - 1 do begin
      if len  = 0 then AList.Append((FList[i])) else
      if Length(FList[i]) >= len then begin
        if ws = CharLowerW(PwChar(Copy(FList[i], 1, len))) then begin
          AList.Append((FList[i]));
        end;
      end;
    end;
    //ȂAList1sڂɋs邩菜
    while (AList.Count > 0) and (Length(AList[0]) = 0) do AList.Delete(0);
    TntListBox1.Items.Text := AList.Text;
    if Auto and (TntListBox1.Count = 1) then begin
      CompleteEnd(TntListBox1.Items[0]);
      Result := false;
      exit;
    end;
    //̃\bhɓ_ł͌₪ĂA
    //̎_ł͌₪ȂȂĂꍇB
    if TntListBox1.Items.Count = 0 then begin
      Result := false;
      exit;
    end;
    //͕KvȂ
    //if (len > 0) and (TntListBox1.Items.IndexOf(FCompString) = -1) then TntListBox1.Items.Append(FCompString);
    if TntListBox1.Items.Count > 0 then TntListBox1.ItemIndex := 0;
  finally
    AList.Free;
  end;
end;

procedure TFormComplete.CompleteCancel;
begin
  FClosed := true;
  SendMessage(TntListBox1.Handle, WM_KILLFOCUS, 0 , 0);
  Close;
end;

procedure TFormComplete.CompleteEnd(const Value: WideString);
var
  i, j, len: integer;
  b: boolean;
  p: PWideChar;
begin
  P := FNotePad.ToPointer(0);
  i := FCompleteStart;
  len := FNotePad.TextLength;
  while i < len do begin
    case (p + i)^ of
      '-': if F2D then inc(i) else break;
      '0'..'9','A'..'Z','_','a'..'z': inc(i);
      else break;
    end;
  end;
  if (Length(Value) = i - FCompleteStart) and (i = FNotePad.SelectionStart) then begin
    b := true;
    for j := 0 to i - FCompleteStart - 1 do begin
      if Value[j + 1] <> (p + FCompleteStart + j)^ then begin
        b := false;
        break;
      end;
    end;
    if b then exit;
  end;
  FNotePad.ReplaseWithoutSelect(Value, FCompleteStart, i - FCompleteStart);
end;

end.
