{********************************************************************}
{                                                                    }
{    TFormStrPreview for TStringsPrinter                             }
{                                                                    }
{    start  1999/10/20                                               }
{    update 2003/11/28                                               }
{                                                                    }
{    Copyright (c) 1999, 2003 {cF <katsuhiko.honda@nifty.ne.jp> }
{                                                                    }
{********************************************************************}

unit StringsPrinterPreview;

{$I heverdef.inc}

interface

{$R STRINGSPRINTEROPENHAND} // crOpenHand Cursor

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, ExtCtrls, ComCtrls, StdCtrls, Menus, NotePadPrinter, TntForms,
  TntComCtrls, TntStdCtrls, TntExtCtrls, TntButtons;

const
  OpenHandResourceIdent = 'OpenHand';
var
  crOpenHand: TCursor = (1960);

type
  TPreviewMode = (pmHundred, pmWidth, pmHeight);

  TFormStringsPrinterPreview = class(TTntForm)
    Panel1: TTntPanel;
    SpeedButton1: TTntSpeedButton;
    SpeedButton2: TTntSpeedButton;
    SpeedButton3: TTntSpeedButton;
    SpeedButton4: TTntSpeedButton;
    SpeedButton5: TTntSpeedButton;
    SpeedButton6: TTntSpeedButton;
    SpeedButton7: TTntSpeedButton;
    SpeedButton8: TTntSpeedButton;
    StatusBar1: TTntStatusBar;
    Panel2: TTntPanel;
    ScrollBox1: TTntScrollBox;
    Image1: TTntImage;
    Bevel1: TTntBevel;
    Label1: TTntLabel;
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure SpeedButton3Click(Sender: TObject);
    procedure SpeedButton4Click(Sender: TObject);
    procedure SpeedButton5Click(Sender: TObject);
    procedure SpeedButton6Click(Sender: TObject);
    procedure SpeedButton7Click(Sender: TObject);
    procedure SpeedButton8Click(Sender: TObject);
    procedure Image1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);

    {$IFDEF COMP4_UP}
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    {$ENDIF}

  private
    FH: Integer;
    FW: Integer;
    FStringsPrinter: TNotePadPrinter;
    FMetaFileList: TMetaFileList;
    FMetaFileIndex: Integer;
    FPreviewMode: TPreviewMode;
    FOrigin: TPoint;
    procedure SetPreviewMode(Value: TPreviewMode);
    procedure SetMetaFileIndex(Value: Integer);
    procedure AdjustImagePosition;
  public
    class procedure Execute(StringsPrinter: TNotePadPrinter; MetaFileList: TMetaFileList);
    property PreviewMode: TPreviewMode read FPreviewMode write SetPreviewMode;
    property MetaFileIndex: Integer read FMetaFileIndex write SetMetaFileIndex;
  end;

implementation


{$R *.DFM}

uses
  Printers, Math;

{ TFormPreview }

class procedure TFormStringsPrinterPreview.Execute(StringsPrinter: TNotePadPrinter;
  MetaFileList: TMetaFileList);
var
  Form: TFormStringsPrinterPreview;
begin
  Form := TFormStringsPrinterPreview.Create(StringsPrinter);
  try
    Form.FStringsPrinter := StringsPrinter;
    Form.FMetaFileList := MetaFileList;
    Form.MetaFileIndex := 0;
    Form.PreviewMode := pmHundred;
    Form.ShowModal;
  finally
    Form.Free;
  end;
end;

procedure TFormStringsPrinterPreview.AdjustImagePosition;
var
  H, W: Integer;
begin
  ScrollBox1.HorzScrollBar.Position := 0;
  ScrollBox1.VertScrollBar.Position := 0;
  // Top
  H := ScrollBox1.Height;
  if ScrollBox1.HorzScrollBar.Range > ScrollBox1.Width then
    Dec(H, FH);
  if Image1.Height > H then
    Image1.Top := 0
  else
    Image1.Top := (H - Image1.Height) div 2;
  // Left
  W := ScrollBox1.Width;
  if ScrollBox1.VertScrollBar.Range > ScrollBox1.Height then
    Dec(W, FW);
  if Image1.Width > W then
    Image1.Left := 0
  else
    Image1.Left := (W - Image1.Width) div 2;
end;

procedure TFormStringsPrinterPreview.SetPreviewMode(Value: TPreviewMode);
var
  MetaFile: TMetaFile;
  Offset: TRect;
  X, Y: Integer;
begin
  FPreviewMode := Value;
  // active metafile
  MetaFile := FMetaFileList.MetaFiles[FMetaFileIndex];
  // paper offset
  Offset := FStringsPrinter.PaperOffset;
  // paper size
  X := Offset.Left + MetaFile.Width + Offset.Right;
  Y := Offset.Top + MetaFile.Height + Offset.Bottom;
  case FPreviewMode of
    pmHundred:
      begin
        Image1.Width := Round(X * FStringsPrinter.Ratio);
        Image1.Height := Round(Y * FStringsPrinter.Ratio);
      end;
    pmWidth:
      begin
        Image1.Width := ScrollBox1.Width - FW;
        Image1.Height := Round(Image1.Width * Y / X);
      end;
    pmHeight:
      begin
        Image1.Height := ScrollBox1.Height - FH;
        Image1.Width := Round(Image1.Height * X / Y);
      end;
  end;
  Image1.Picture.Graphic := nil;
  Image1.Canvas.StretchDraw(
    Rect(
      Round(Offset.Left * Image1.Width / X),
      Round(Offset.Top * Image1.Height / Y),
      Round((Offset.Left + MetaFile.Width) * Image1.Width / X),
      Round((Offset.Top + MetaFile.Height) * Image1.Height / Y)
    ),
    MetaFile
  );
  AdjustImagePosition;
end;

procedure TFormStringsPrinterPreview.SetMetaFileIndex(Value: Integer);
begin
  FMetaFileIndex := Max(0, Min(Value, FMetaFileList.Count - 1));
  Label1.Caption := Format('%3d / %3d', [FMetaFileIndex + 1, FMetaFileList.Count]);
  SpeedButton1.Enabled := FMetaFileIndex > 0;
  SpeedButton2.Enabled := FMetaFileIndex > 0;
  SpeedButton3.Enabled := FMetaFileIndex < FMetaFileList.Count - 1;
  SpeedButton4.Enabled := FMetaFileIndex < FMetaFileList.Count - 1;
  UpdateWindow(Handle);
  SetPreviewMode(PreviewMode);
end;

procedure TFormStringsPrinterPreview.FormShow(Sender: TObject);
begin
  FW := GetSystemMetrics(SM_CXVSCROLL);
  FH := GetSystemMetrics(SM_CYHSCROLL);
//  Width := Min(Screen.Width, Image1.Width + FW + 16);
//  Height := Screen.WorkAreaHeight div 6 * 5;
  {$IFDEF COMP4_UP}
   Self.OnMouseWheel := FormMouseWheel;
  {$ENDIF}
end;

procedure TFormStringsPrinterPreview.FormResize(Sender: TObject);
begin
  AdjustImagePosition;
end;

procedure TFormStringsPrinterPreview.SpeedButton1Click(Sender: TObject);
begin
  MetaFileIndex := 0;
end;

procedure TFormStringsPrinterPreview.SpeedButton2Click(Sender: TObject);
begin
  MetaFileIndex := MetaFileIndex - 1;
end;

procedure TFormStringsPrinterPreview.SpeedButton3Click(Sender: TObject);
begin
  MetaFileIndex := MetaFileIndex + 1;
end;

procedure TFormStringsPrinterPreview.SpeedButton4Click(Sender: TObject);
begin
  MetaFileIndex := FMetaFileList.Count - 1;
end;

procedure TFormStringsPrinterPreview.SpeedButton5Click(Sender: TObject);
begin
  PreviewMode := pmHeight;
end;

procedure TFormStringsPrinterPreview.SpeedButton6Click(Sender: TObject);
begin
  PreviewMode := pmWidth;
end;

procedure TFormStringsPrinterPreview.SpeedButton7Click(Sender: TObject);
begin
  PreviewMode := pmHundred;
end;

procedure TFormStringsPrinterPreview.SpeedButton8Click(Sender: TObject);
begin
  Close;
end;

procedure TFormStringsPrinterPreview.Image1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FOrigin := Point(X, Y);
end;

procedure TFormStringsPrinterPreview.Image1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if (ScrollBox1.HorzScrollBar.Range > ScrollBox1.Width) or
     (ScrollBox1.VertScrollBar.Range > ScrollBox1.Height) then
    if csLButtonDown in Image1.ControlState then
    begin
      ScrollBox1.HorzScrollBar.Position := ScrollBox1.HorzScrollBar.Position + (FOrigin.X - X);
      ScrollBox1.VertScrollBar.Position := ScrollBox1.VertScrollBar.Position + (FOrigin.Y - Y);
    end
    else
      Image1.Cursor := crOpenHand
  else
    Image1.Cursor := crDefault;
end;

{$IFDEF COMP4_UP}
procedure TFormStringsPrinterPreview.FormMouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
begin
  ScrollBox1.VertScrollBar.Position := ScrollBox1.VertScrollBar.Position - WheelDelta;
end;
{$ENDIF}

procedure LoadCursors;
begin
  // crOpenHand Cursor Resource
  Screen.Cursors[crOpenHand] := LoadCursor(HInstance, PChar(OpenHandResourceIdent));
end;

initialization
  LoadCursors;
end.
