unit ModeDelphi;

interface

uses Windows, NotePadMode, NotePadString;

type
  TModeDelphi = class(TNotePadMode)
  protected
    procedure InitReserveList;override;
  public
    function NewParser(target: TNotePadString): TNotePadParser;override;
  end;

  TParserDelphi = class(TNotePadParser)
  private
  protected
    procedure ParseToken(AHttp: boolean);overload;override;
    procedure ParseToken(const C: WideChar);overload;override;
  public
    constructor Create(target: TNotePadString; mode: TModeDelphi);
  end;

implementation

uses NotePadClasses, ParserConsts;

const
  ss1: array[0..96] of WideString =(
    'absolute',
    'abstract',
    'and',
    'array',
    'as',
    'asm',
    'assembler',
    'automated',
    'begin',
    'case',
    'cdecl',
    'class',
    'const',
    'constructor',
    'contains',
    'default',
    'destructor',
    'dispid',
    'dispinterface',
    'div',
    'do',
    'downto',
    'dynamic',
    'else',
    'end',
    'except',
    'export',
    'exports',
    'external',
    'far',
    'file',
    'finalization',
    'finally',
    'for',
    'forward',
    'function',
    'goto',
    'if',
    'implementation',
    'in',
    'inherited',
    'initialization',
    'inline',
    'interface',
    'is',
    'label',
    'library',
    'message',
    'mod',
    'near',
    'nil',
    'nodefault',
    'not',
    'object',
    'of',
    'or',
    'out',
    'overload',
    'override',
    'packed',
    'pascal',
    'private',
    'procedure',
    'program',
    'property',
    'protected',
    'public',
    'published',
    'raise',
    'readonly',
    'record',
    'register',
    'repeat',
    'requires',
    'resident',
    'resourcestring',
    'safecall',
    'set',
    'shl',
    'shr',
    'stdcall',
    'stored',
    'string',
    'then',
    'threadvar',
    'to',
    'try',
    'type',
    'unit',
    'until',
    'uses',
    'var',
    'virtual',
    'while',
    'with',
    'writeonly',
    'xor');

type
  THash1 = class(THashTable)
  protected
    procedure InitTable; override;
  public
    function isReserbed(const Value: WideString): boolean; override;
  end;

{ THash1 }

procedure THash1.InitTable;
var
  aa1: array of word;
  i, j: integer;
begin
  FSize := 197;{}
  FMaxLen := 1;
  SetLength(aa1, Length(ss1{}));
  for i := 0 to  Length(aa1) - 1 do begin
    aa1[i] := AnkToHash(ss1{}[i]);
    if Length(ss1{}[i]) > FMaxLen then FMaxLen := Length(ss1{}[i]);
  end;
  SetLength(FTable, FSize);
  for i := 0 to  Length(FTable) - 1 do begin
    FTable[i] := 0;
  end;
  for i := 0 to  Length(aa1) - 1 do begin
    j := aa1[i];
    while FTable[j] > 0 do begin
      inc(j);
      if j >= Length(FTable) then dec(j, Length(FTable) - 1);
    end;
    FTable[j] := i + 1;
  end;
end;

function THash1.isReserbed(const Value: WideString): boolean;
var
  i: integer;
begin
  Result := false;
  if Length(Value) > FMaxLen then exit;
  i := AnkToHash(Value);
  while FTable[i] > 0 do begin
    if CharLowerW(PWChar(ss1{}[FTable[i] - 1])) = Value then begin
      Result := true;
      exit;
    end;
    inc(i);
    if i >= FSize then dec(i, FSize);
  end;
end;

var
  FHash1: THash1;

type
  TDelphiList = class(TReserveList)
  public
    function Find(const s: WideString): boolean; override;
  end;

{ TDelphiList }

function TDelphiList.Find(const s: WideString): boolean;
begin
  Result := FHash1.isReserbed(CharLowerW(PWChar(s)));
end;

{ TModeDelphi }

procedure TModeDelphi.InitReserveList;
var
  a: array of WideString;
begin
  FReserveList := TDelphiList.Create(a, false);
  FReserve2List := TReserveList.Create(a, true);
  FReserve3List := TReserveList.Create(a, true);
end;

function TModeDelphi.NewParser(target: TNotePadString): TNotePadParser;
begin
  Result := TParserDelphi.Create(target, self);
end;

{ TParserDelphi }

constructor TParserDelphi.Create(target: TNotePadString;
  mode: TModeDelphi);
begin
  inherited Create(target, mode);
end;

procedure TParserDelphi.ParseToken(AHttp: boolean);
var
  C2: WideChar;
begin
  if FToken = toCommentBrock2Start then begin
    ParseToken(toCommentBrock2);
    exit;
  end;
  case FSb[FParsePos] of
    '''': begin
      if FParsePos < FSb.StringLength - 1 then begin
        C2 := FSb[FParsePos + 1];
        if C2 <> #10 then begin
          inc(FParsePos);
          ParseToken(toSingleQuotation);
          exit;
        end;
      end;
      ParseToken(toSingleQuotation);
      exit;
    end;
    '{': begin
      ParseToken(toCommentBrock0);
      exit;
    end;
    '/': begin
      if FParsePos < FSb.StringLength - 1 then begin
        C2 := FSb[FParsePos + 1];
        if C2 = '/' then begin
          inc(FParsePos);
          ParseToken(toComment);
          exit;
        end;
      end;
    end;
    #$28: begin
      if FParsePos < FSb.StringLength - 1 then begin
        C2 := FSb[FParsePos + 1];
        if C2 = #$2A then begin
          inc(FParsePos);
          ParseToken(toCommentBrock2Start);
          exit;
        end;
      end;
    end;
    #$24: begin
      ParseToken(toHex);
      exit;
    end;
    #$23: begin
      if FParsePos < FSb.StringLength - 1 then begin
        C2 := FSb[FParsePos + 1];
        if C2 = #$24 then begin
          inc(FParsePos);
          ParseToken(toDelphiCharHex);
          exit;
        end;
      end;
      ParseToken(toDelphiChar);
      exit;
    end;
  end;
  inherited ParseToken(false);
end;

procedure TParserDelphi.ParseToken(const C: WideChar);
begin
  case C of
    toSingleQuotation: begin
      FToken := C;
      SingleQuotationProc;
    end;
    toCommentBrock0: begin
      FToken := C;
      CommentBrock0Proc;
    end;
    toComment: begin
      FToken := C;
      CommentProc;
    end;
    toCommentBrock2Start: begin
      FToken := C;
      CommentBrock2StartProc;
    end;
    toCommentBrock2: begin
      FToken := C;
      CommentBrock2Proc(0);
    end;
    toCommentBrock2sp: begin
      FToken := toCommentBrock2;
      CommentBrock2Proc(1);
    end;
    toHex: begin
      FToken := C;
      HexProc;
    end;
    toDelphiChar: begin
      FToken := C;
      IntegerProc;
    end;
    toDelphiCharHex: begin
      FToken := C;
      HexProc;
    end;
    else inherited;
  end;
end;


initialization
  FHash1 := THash1.Create;
finalization
  FHash1.Free;
end.
