unit NotePadString;

interface

uses Windows, Classes, Graphics, NotePadClasses, StringBuffer;

type
  TNotePadString = class
  private
    FBuffer: TStringBuffer;
    FStart: TCaretPosition;
    FEnd: TCaretPosition;
    FChangedStart: integer;
    FChangedEnd: integer;
    FCountBeforeChanged: integer;
    FStrings: TLineList;
    FUndoList: TUndoList;
    FWrapByte: integer;
    FTabCount: integer;
    FBoxMode: boolean;
    FModified: boolean;
    FUpdateCount: integer;
    FNeedWrapAll: boolean;
    FParseDone: boolean;
    FMarks: AInteger;
    FLineEnds: TLineEnds;
    //Cxg
    FOnChanged: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    FOnModified: TNotifyEvent;
    FOnNeedParse: TNotifyEvent;

    procedure Changing; 
    procedure Changed;
    procedure ClearString; 
    procedure DeleteString(ABox: boolean); 
    procedure InsertString(Value: WideString; ABox: boolean);
    procedure WrapAll;
    procedure WrapFast;
    procedure CheckWrapAll;
    procedure MarksOnDeleteString(AStart, ALength: integer);
    procedure MarksOnInsertString(AStart, ALength: integer);
    procedure MarksRemoveRange(AStart, ALength: integer);
    //vpeB
    function GetCanUndo: boolean;
    function GetCanRedo: boolean;
    function GetSelStartPos: TPoint;
    procedure SetSelStartPos(Value: TPoint);
    function GetSelEndPos: TPoint;
    procedure SetSelEndPos(Value: TPoint);
    function GetSelStart: integer;
    procedure SetSelStart(Value: integer);
    function GetSelEnd: integer;
    procedure SetSelEnd(Value: integer);
    function GetSelText: WideString;
    function GetSelTextIf: WideString;
    procedure SetSelText(const Value: WideString);
    function GetCount: integer;
    function GetTextLength: integer;
    function GetSelLength: integer;
    procedure SetWrapByte(Value: integer);
    procedure SetTabCount(Value: integer);
    function GetText: WideString;
    function GetTextIf: WideString;
    procedure SetText(const Value: WideString);
    function GetItems(Index: integer): PLineData;
    procedure SetItems(Index: integer; data: PLineData);
    function GetMarkCount: integer;
    function GetMarks(Index: Integer): integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure ClearAll;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure ClearUndo;
    function GetLineLength(Value: integer): integer;
    procedure ClearModified;
    function GetSelectedArea(Y: integer): TPoint;
    function GetCaretX(X, Y: integer): integer;
    function FromCaretX(X, Y: integer; out over: boolean): integer;
    procedure RedrawCurrent;
    procedure SelEndToSelStart;
    procedure Undo;
    procedure Redo;
    function GetTextCR: WideString;
    function GetTextLF: WideString;
    function GetTextCRLF: WideString;
    procedure SetSelTextBox(const Value: WideString);
    function GetReturnedString: WideString;
    function GetIndentString(X, Y: integer): WideString;
    function GetLineString(X, Y: integer; out XX: integer): WideString;
    function LinesPosToWrapPos(X, Y: integer): TPoint;
    procedure InputDelKey;
    procedure InputBackKey;
    function DoTabIndent(unindent: boolean): boolean;
    procedure DragMove(dstpos: TPoint);
    procedure AddMark(Value: integer);
    procedure DeleteMark(Value: integer);
    procedure ReplaseWithoutSelect(const Value: WideString; AStart, ALength: integer);
    //vpeB
    property CanUndo: boolean read GetCanUndo;
    property CanRedo: boolean read GetCanRedo;
    property SelStartPos: TPoint read GetSelStartPos write SetSelStartPos;
    property SelEndPos: TPoint read GetSelEndPos write SetSelEndPos;
    property SelStart: integer read GetSelStart write SetSelStart;
    property SelEnd: integer read GetSelEnd write SetSelEnd;
    property SelText: WideString read GetSelText write SetSelText;
    property SelTextIf: WideString read GetSelTextIf;
    property Count: integer read GetCount;
    property BoxMode: boolean read FBoxMode write FBoxMode;
    property InternalSB: TStringBuffer read FBuffer;
    property CountBeforeChanged: integer read FCountBeforeChanged;
    property ChangedStart: integer read FChangedStart;
    property ChangedEnd: integer read FChangedEnd;
    property TextLength: integer read GetTextLength;
    property SelLength: integer read GetSelLength;
    property WrapByte: integer read FWrapByte write SetWrapByte;
    property TabCount: integer read FTabCount write SetTabCount;
    property Text: WideString read GetText write SetText;
    property TextIf: WideString read GetTextIf;
    property Items[Index: Integer]: PLineData read GetItems write SetItems; default;
    property ParseDone: boolean read FParseDone;
    property Marks[Index: Integer]: integer read GetMarks;
    property MarkCount: integer read GetMarkCount;
    property LineEnds: TLineEnds read FLineEnds;
    //Cxg
    property OnChanged: TNotifyEvent read FOnChanged write FOnChanged;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
    property OnModified: TNotifyEvent read FOnModified write FOnModified;
    property OnNeedParse: TNotifyEvent read FOnNeedParse write FOnNeedParse;
  end;

//eLXg̓ŕKLFϊ̂ŁA
//̃NX̓f[^CR邱Ƃ͐΂ɂȂdlB

implementation

uses Types, Math, NotePadConsts, UnicodeWrap, FontUtils;

{ TNotePadString }

procedure TNotePadString.AddMark(Value: integer);
var
  i: integer;
  ps, pd: Cardinal;
begin
  while Value > 0 do begin
    if Value < FBuffer.StringLength then begin
      case FBuffer[Value - 1] of
        #10: break;
      end;
    end;
    dec(Value)
  end;
  if Length(FMarks) = 0 then begin
    SetLength(FMarks, 1);
    FMarks[0] := Value;
  end else begin
    for i := 0 to Length(FMarks) - 1 do begin
      if Value < FMarks[i] then begin
        SetLength(FMarks, Length(FMarks) + 1);
        ps := Cardinal(@(FMarks[0])) + Cardinal(i) * 4;
        pd := ps + 4;
        Move(Pointer(ps)^, Pointer(pd)^, (Length(FMarks) - i - 1) * 4);
        FMarks[i] := Value;
        break;
      end else if Value > FMarks[i] then begin
        if i = Length(FMarks) - 1 then begin
          SetLength(FMarks, i + 2);
          FMarks[i + 1] := Value;
          break;
        end;
      end else break;
    end;
  end;
end;

procedure TNotePadString.BeginUpdate;
begin
  inc(FUpdateCount);
  if FUpdateCount = 1 then Changing;
end;

procedure TNotePadString.Changed;
begin
  if (FUpdateCount = 0) then begin
    CheckWrapAll;
    if Assigned(FOnChanged) then begin
      FOnChanged(self);
    end;
  end;
end;

procedure TNotePadString.Changing;
begin
  if Assigned(FOnChanging) then FOnChanging(self);
  FChangedStart := Min(FStart.Y, FEnd.Y);
  FChangedEnd := Max(FStart.Y, FEnd.Y);
  FCountBeforeChanged := FStrings.Count;
  FParseDone := false;
end;

procedure TNotePadString.CheckWrapAll;
begin
  if not FNeedWrapAll then exit;
  //EndUpdateɌĂ΂Ă̂BeginUpdate͗vȂ
  FNeedWrapAll := false;
  FChangedStart := 0;
  FChangedEnd := FStrings.Count - 1;
  WrapAll;
end;

procedure TNotePadString.ClearAll;
begin
  BeginUpdate;
  ClearString;
  ClearUndo;
  FModified := false;
  EndUpdate;
end;

procedure TNotePadString.ClearModified;
begin
  FModified := false;
end;

procedure TNotePadString.ClearString;
var
  ALine: TLineData;
begin
  FChangedStart := 0;
  FChangedEnd := Max(FChangedEnd, (FStrings.Count - 1));
  inc(FCountBeforeChanged);
  FStart.Position := 0;
  FStart.X := 0;
  FStart.Y := 0;
  FEnd := FStart;
  FStrings.Clear();
//  FBuffer.Free;
//  FBuffer := TStringBuffer.Create(0);
  FBuffer.Remove(0, FBuffer.StringLength);
  SetLength(FMarks, 0);
  ALine.NeedWrap := false;
  ALine.LineEnd := leNone;
  ALine.Length := 0;
  ALine.Start := 0;
  FStrings.Add(ALine);
  if not FModified then begin
    FModified := true;
    if Assigned(FOnModified) then FOnModified(self);
  end;
end;

procedure TNotePadString.ClearUndo;
begin
  FUndoList.Clear;
end;

constructor TNotePadString.Create;
begin
  FBuffer := TStringBuffer.Create(0);
  FStrings := TLineList.Create(0);
  FUndoList := TUndoList.Create;
  FWrapByte := MAX_WRAPBYTE;
  FTabCount := 8;
  FNeedWrapAll := false;
  ClearAll;
  FLineEnds := lsCRLF;
end;

procedure TNotePadString.DeleteMark(Value: integer);
begin
  MarksRemoveRange(Value, 1);
end;

procedure TNotePadString.DeleteString(ABox: boolean);
var
  i: integer;
  swap: TCaretPosition;
  AStart: TCaretPosition;
  p: array of TPoint;
  DelLength: integer;
  ALine: PLineData;
  NewLine: TLineData;
begin
  if FStart.Position = FEnd.Position then exit;
  if FStart.Position > FEnd.Position then begin
    swap := FStart; FStart := FEnd; FEnd := swap;
  end;
  if ABox then begin
    AStart := FStart;
    SetLength(p, FEnd.Y - FStart.Y + 1);
    for i := FStart.Y to FEnd.Y do
      p[i - AStart.Y] := GetSelectedArea(i);
    for i :=FStart.Y to FEnd.Y do begin
      SelStartPos := Point(p[i - AStart.Y].X, i);
      SelEndPos := Point(p[i - AStart.Y].Y, i);
      DeleteString(false);
    end;
    FStart := AStart;
    if p[0].X < AStart.X then begin
      FStart.X := p[0].X;
      dec(FStart.Position, AStart.X - p[0].X);
    end;
  end else begin
    DelLength := FEnd.Position - FStart.Position;
    FBuffer.Remove(FStart.Position, DelLength);
    MarksOnDeleteString(FStart.Position, DelLength);
    FChangedStart := Min(FChangedStart, FStart.Y);
    FChangedEnd := Max(FChangedEnd, FEnd.Y);
    if FStart.Y = FEnd.Y then begin
      ALine := FStrings[FStart.Y];
      ALine.NeedWrap := true;
      dec(ALine.Length, DelLength);
      for i := FStart.Y + 1 to FStrings.Count -1 do begin
        dec(FStrings[i].Start, DelLength);
      end;
    end else begin
      ALine := FStrings[FEnd.Y];
      NewLine.NeedWrap := true;
      NewLine.LineEnd := ALine.LineEnd;
      NewLine.Start := FStrings[FStart.Y].Start;
      NewLine.Length := FStart.X + ALine.Length - FEnd.X;
      FStrings.RemoveRange(FStart.Y + 1, FEnd.Y - FStart.Y);
      FStrings[FStart.Y]^ := NewLine;
      for i := FStart.Y + 1 to FStrings.Count -1 do begin
        dec(FStrings[i].Start, DelLength);
      end;
    end;
  end;
  FEnd := FStart;
  if not FModified then begin
    FModified := true;
    if Assigned(FOnModified) then FOnModified(self);
  end;
end;

destructor TNotePadString.Destroy;
begin
  FBuffer.Free;
  FStrings.Free;
  FUndoList.Free;
  inherited;
end;

function TNotePadString.DoTabIndent(unindent: boolean): boolean;
var
  ALine: PLineData;
  start1, end1: TCaretPosition;
  start2, end2: integer;
  i, j: integer;
  s: WideString;
  b: boolean;
  sl: array of WideString;
  sb: TStringBuffer;
begin
  Result := false;
	if FBoxMode then exit;//Box͋AĂ
	if FEnd.Position = FStart.Position then exit;//IĂȂfalse
  if FStart.Y = FEnd.Y then begin//s܂łȂ
    if FStart.X = 0 then begin
      ALine := FStrings[FEnd.Y];
      if ALine.Length <> FEnd.X then exit;
    end else if FEnd.X = 0 then begin
      ALine := FStrings[FStart.Y];
      if ALine.Length <> FStart.X then exit;
    end else exit; //sȂ̂Ɏn_I_X=0Ȃ
  end;
  BeginUpdate;
  //ZNg
  if FEnd.Position > FStart.Position then begin
    start1 := FStart;
    end1 := FEnd;
  end else begin
    start1 := FEnd;
    end1 := FStart;
  end;
  start2 := start1.Position;
  dec(start1.Position, start1.X);
  i := start1.Y;
  while (i > 0) and (FStrings[i -1].LineEnd = leWrap) do begin
    dec(i);
    dec(start1.Position, FStrings[i].Length);
  end;
  end2 := end1.Position;
  if end1.X = 0 then i := end1.Y - 1 else begin
    inc(end1.Position, FStrings[end1.Y].Length - end1.X);
    i := end1.Y;
  end;
  while (i < FStrings.Count) and (FStrings[i].LineEnd = leWrap) do begin
    inc(i);
    inc(end1.Position, FStrings[i].Length);
  end;
  if (start1.Position < start2) or (end1.Position > end2) then begin
    SelStart := start1.Position;
    SelEnd := end1.Position;
  end;
  //ZNg܂
  RedrawCurrent;
  s := SelText;
  b := false;
  j := 0;
  i := 0;
  while i < Length(s) do begin
  //for i := 0 to Length(s) - 1 do begin
    case s[i + 1] of
      #10:begin
        SetLength(sl, Length(sl) + 1);
        sl[Length(sl) - 1] := copy(s, j + 1, i - j + 1);
        j := i + 1;
      end;
      else begin
        if (i = Length(s) - 1) then begin
          SetLength(sl, Length(sl) + 1);
          sl[Length(sl) - 1] := copy(s, j + 1, i - j + 1);
        end;
      end;
    end;//switch
    inc(i);
  end;//for
  if unindent then begin
    for i := 0 to Length(sl) - 1 do begin
      if (Length(sl[i]) > 0) and (sl[i][1] = #09) then begin
        b := true;
        sl[i] := copy(sl[i], 2, Length(sl[i]) - 1);
      end;
    end;
  end else begin
    for i := 0 to Length(sl) - 1 do begin
      if (Length(sl[i]) > 0) and (sl[i][1] <> #10) then begin
        b := true;
        sl[i] := #09 + sl[i];
      end;
    end;
  end;
  if b then begin
    sb := TStringBuffer.Create(Length(s));
    for i := 0 to Length(sl) - 1 do sb.Append(sl[i]);
    i := Min(SelStart, SelEnd);
    SelText := sb.ToString;
    sb.Free;
    SelStart := i;
    RedrawCurrent;
  end;
  EndUpdate;
  Result := true;//CfgsłZNg͂܂
end;

procedure TNotePadString.DragMove(dstpos: TPoint);
var
  AUndo: TUndoData;
  AStart, AEnd: TCaretPosition;
  dst: integer;
  ALine: PLineData;
begin
  if FBoxMode then exit;
  if FStart.Y < FEnd.Y then begin
    AStart := FStart;
    AEnd := FEnd;
  end else if FEnd.Y < FStart.Y then begin
    AStart := FEnd;
    AEnd := FStart;
  end else if FStart.X < FEnd.X then begin
    AStart := FStart;
    AEnd := FEnd;
  end else if FEnd.X < FStart.X then begin
    AStart := FEnd;
    AEnd := FStart;
  end else begin
    AStart := FStart;
    AEnd := FEnd;
  end;
  ALine := FStrings[dstpos.Y];
  dst := ALine.Start + dstpos.X;
  if (AStart.Position <= dst) and (dst <= AEnd.Position) then Exit;
  BeginUpdate;
  AUndo.MarksBefore := Copy(FMarks);
  if dst < AStart.Position then begin
    AUndo.Position := dst;
    AUndo.SelStartBefore := FStart.Position;
    AUndo.SelEndBefore := FEnd.Position;
    AUndo.SelStartAfter := dst + (AEnd.Position - AStart.Position);
    AUndo.SelEndAfter := dst;
    AUndo.Deleted := FBuffer.ToString(dst, AEnd.Position - dst);
    AUndo.Added := FBuffer.ToString(AStart.Position, AEnd.Position - AStart.Position) +
      FBuffer.ToString(dst, AStart.Position - dst);
    SelStart := AEnd.Position;
    SelEnd := dst;
    RedrawCurrent;
    DeleteString(false);
    InsertString(AUndo.Added, false);
    SelEnd := dst;
    SelStart := dst + (AEnd.Position - AStart.Position);
    RedrawCurrent;
  end else begin
    AUndo.Position := AStart.Position;
    AUndo.SelStartBefore := FStart.Position;
    AUndo.SelEndBefore := FEnd.Position;
    AUndo.SelStartAfter := dst - (AEnd.Position - AStart.Position);
    AUndo.SelEndAfter := dst;
    AUndo.Deleted := FBuffer.ToString(AStart.Position, dst - AStart.Position);
    AUndo.Added := FBuffer.ToString(AEnd.Position, dst - AEnd.Position) +
      FBuffer.ToString(AStart.Position, AEnd.Position - AStart.Position);
    SelStart := AStart.Position;
    SelEnd := dst;
    RedrawCurrent;
    DeleteString(false);
    InsertString(AUndo.Added, false);
    SelStart := SelStart - (AEnd.Position - AStart.Position);
    RedrawCurrent;
  end;
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndBefore := FLineEnds;
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.EndUpdate;
begin
  dec(FUpdateCount);
  Changed;
end;

function TNotePadString.FromCaretX(X, Y: integer;
  out over: boolean): integer;
var
  i, sum, sumtab: integer;
  ALine: PLineData;
  s: WideString;
begin
  sum := 0;
  over := false;
  ALine := FStrings[Min(FStrings.Count - 1, Y)];
  //if ALine = nil then Result := Y;
  //if ALine <> nil then
  s := FBuffer.ToString(ALine.Start, ALine.Length);
  Result := -1;//G[ɂȂ鐔
  //ȂX}CiXĂ܂B[ȏザȂƔj]
  if X < 0 then X := 0;
  for i := 0 to Length(s) do begin
    Result := i;
    if sum = X then exit;
    if sum > X then begin
      Result := i - 1;
      exit;
    end;
    if i = Length(s) then begin
      over := true;
      exit;
    end;
    case s[i + 1] of
      #$0009: begin
        sumtab := FTabCount * GetGlyphAveWidth - (sum mod (FTabCount * GetGlyphAveWidth));
        if sumtab < GetGlyphAveWidth then inc(sumtab, FTabCount * GetGlyphAveWidth);
        inc(sum, sumtab);
      end;
      else inc(sum, GetGlyphWidth(s[i + 1]));
    end;
  end;
end;

function TNotePadString.GetCanRedo: boolean;
begin
  Result := FUndoList.CanRedo;
end;

function TNotePadString.GetCanUndo: boolean;
begin
  Result := FUndoList.CanUndo;
end;

function TNotePadString.GetCaretX(X, Y: integer): integer;
var
  ALine: PLineData;
  s: WideString;
  i, sumtab: integer;
begin
  ALine := FStrings[Min(FStrings.Count - 1, Y)];
  s := FBuffer.ToString(ALine.Start, ALine.Length);
  Result := 0;
  for i := 1 to Min(X, Length(s)) do begin
    case s[i] of
      #$0009: begin
        sumtab := FTabCount * GetGlyphAveWidth - (Result mod (FTabCount * GetGlyphAveWidth));
        if sumtab < GetGlyphAveWidth then inc(sumtab, FTabCount * GetGlyphAveWidth);
        inc(Result, sumtab);
      end;
      else  inc(Result, GetGlyphWidth(s[i]));
    end;
  end;
  if X = Length(s) + 1 then inc(Result, GetGlyphAveWidth); //wi`̉s
end;

function TNotePadString.GetCount: integer;
begin
  Result := FStrings.Count;
end;

function TNotePadString.GetIndentString(X, Y: integer): WideString;
var
  ALine: PLineData;
  s: WideString;
  i, j:integer;
begin
  while Y > 0 do begin
    ALine := FStrings[Y - 1];
    if ALine.LineEnd <> leWrap then break;
    dec(Y);
    inc(X, GetLineLength(Y));
  end;
  ALine := FStrings[Y];
  s := FBuffer.ToString(ALine.Start, ALine.Length);
  j := -1;
  for i := 0 to Length(s) -1 do begin
    if i >= X then break;
    case s[i + 1] of
      #09, #$0020, #$3000: j := i;
      else break;
    end;
  end;
  inc(j);
  if (j = Length(s)) and (ALine.LineEnd = leWrap) then j := 0;
  Result := copy(s, 1, j);
end;

function TNotePadString.GetItems(Index: integer): PLineData;
begin
  Result := FStrings[Index];
end;

function TNotePadString.GetLineLength(Value: integer): integer;
var
  ALine: PLineData;
begin
  ALine := FStrings[Value];
  case ALine.LineEnd of
    leLF:
      Result := ALine.Length + 1;
    else
      Result := ALine.Length;
  end;
end;

function TNotePadString.GetLineString(X, Y: integer;
  out XX: integer): WideString;
var
  ALine: PLineData;
  i: integer;
  P: PWChar;
begin
  while Y > 0 do begin
    ALine := FStrings[Y - 1];
    if ALine.LineEnd <> leWrap then break;
    dec(Y);
    inc(X, GetLineLength(Y));
  end;
  ALine := FStrings[Y];
  XX := X;
  P := FBuffer.ToPointer(ALine.Start);
  i := 0;
  while not ((P + i)^ in [WideChar(#0), WideChar(#10)]) do begin
    inc(i);
  end;
  Result := FBuffer.ToString(ALine.Start, i);
end;

function TNotePadString.GetMarkCount: integer;
begin
  Result := Length(FMarks);
end;

function TNotePadString.GetMarks(Index: Integer): integer;
begin
  Result := FMarks[Index];
end;

function TNotePadString.GetReturnedString: WideString;
//AppendŉsR[hĂ̂͌񂶂ȂH
var
  i: integer;
  ALine: PLineData;
  sb: TStringBuffer;
begin
  sb := TStringBuffer.Create(2);
  for i := 0 to FStrings.Count - 1 do begin
    ALine := FStrings[i];
    case ALine.LineEnd of
		  leLF,//:begin
			  //sb.Append(FBuffer.ToString(ALine.Start, ALine.Length));
			  //sb.Append(#10);
			//end;
		  leWrap:begin
		    case FLineEnds of
          lsCR:begin
				    sb.Append(FBuffer.ToString(ALine.Start, ALine.Length));
				    sb.Append(#13);
				  end;
			    lsLF:begin
				    sb.Append(FBuffer.ToString(ALine.Start,ALine.Length));
				    sb.Append(#10);
				  end;
			    lsCRLF:begin
				    sb.Append(FBuffer.ToString(ALine.Start,ALine.Length));
				    sb.Append(#13#10);
				  end;
        end;
      end;
      else sb.Append(FBuffer.ToString(ALine.Start, ALine.Length));
    end;
  end;
  result := sb.ToString();
  sb.Free;
end;

function TNotePadString.GetSelectedArea(Y: integer): TPoint;
var
  L, R: integer;
  AStart, AEnd: TCaretPosition;
  over1, over2: boolean;
  ALine: PLineData;
begin
  Result := Point(0, 0);
  if FStart.Y < FEnd.Y then begin
    AStart := FStart;
    AEnd := FEnd;
  end else if FEnd.Y < FStart.Y then begin
    AStart := FEnd;
    AEnd := FStart;
  end else if FStart.X < FEnd.X then begin
    AStart := FStart;
    AEnd := FEnd;
  end else if FEnd.X < FStart.X then begin
    AStart := FEnd;
    AEnd := FStart;
  end else exit;
  if (AEnd.Y < Y) or (AStart.Y > Y) then exit;
  if FBoxMode then begin
    L := GetCaretX(AStart.X, AStart.Y);
    R := GetCaretX(AEnd.X, AEnd.Y);
    if L = R then exit;
    if L < R then
      Result := Point(FromCaretX(L, Y, over1), FromCaretX(R, Y, over2))
    else
      Result := Point(FromCaretX(R, Y, over1), FromCaretX(L, Y, over2));
    if (Result.X <> Result.Y) or over2 then exit;
    inc(Result.Y);
  end else begin
    if AStart.Y < Y then
      Result := Point(-1, 0)
    else
      Result := Point(AStart.X, 0);
    ALine := FStrings[Y];
    if AEnd.Y > Y then begin
      Result.Y := ALine.Length;
      case ALine.LineEnd of
        leLF: inc(Result.Y);
      end;
      exit;
    end;
    Result.Y := AEnd.X;
  end;
end;

function TNotePadString.GetSelEnd: integer;
begin
  Result := FEnd.Position;
end;

function TNotePadString.GetSelEndPos: TPoint;
begin
  Result := Point(FEnd.X, FEnd.Y);
end;

function TNotePadString.GetSelLength: integer;
begin
  Result := Abs(FStart.Position - FEnd.Position);
end;

function TNotePadString.GetSelStart: integer;
begin
  Result := FStart.Position;
end;

function TNotePadString.GetSelStartPos: TPoint;
begin
  Result := Point(FStart.X, FStart.Y);
end;

function TNotePadString.GetSelText: WideString;
var
  AStart, AEnd: TCaretPosition;
  sb: TStringBuffer;
  i: integer;
  ALine: PLineData;
  p: TPoint;
begin
  if FStart.Position = FEnd.Position then begin
    Result := '';
    exit;
  end;
  if FStart.Position < FEnd.Position then begin
    AStart := FStart;
    AEnd := FEnd;
  end else begin
    AStart := FEnd;
    AEnd := FStart;
  end;
  if FBoxMode and (AStart.Y <> AEnd.Y) then begin //sȂȂBOXɂȂėǂ
    sb := TStringBuffer.Create(2);
    for i := AStart.Y to AEnd.Y do begin
      ALine := FStrings[i];
      p := GetSelectedArea(i);
      sb.Append(copy(FBuffer.ToString(ALine.Start, ALine.Length), p.X + 1, p.Y - p.X));
      sb.Append(#10);
    end;
    Result := sb.ToString;
    sb.Free;
  end else Result := FBuffer.ToString(AStart.Position, AEnd.Position - AStart.Position);
end;

function TNotePadString.GetSelTextIf: WideString;
var
  ws: WideString;
  i, j, len: integer;
begin
  ws := SelText;
  len := Length(ws);
  SetLength(Result, len * 2);
  i := 1;
  j := 1;
  while i <= len do begin
    if ws[i] = #10 then begin
      case FLineEnds of
        lsCR: begin
          Result[j] := #13;
          inc(j);
        end;
        lsLF: begin
          Result[j] := #10;
          inc(j);
        end;
        lsCRLF: begin
          Result[j] := #13;
          inc(j);
          Result[j] := #10;
          inc(j);
        end;
      end;
    end else begin
      Result[j] := ws[i];
      inc(j);
    end;
    inc(i);
  end;//while
  SetLength(Result, j - 1);
end;

function TNotePadString.GetText: WideString;
begin
  Result := FBuffer.ToString;
end;

function TNotePadString.GetTextIf: WideString;
var
  ws: WideString;
  i, j, len: integer;
begin
  ws := FBuffer.ToString;
  len := Length(ws);
  SetLength(Result, len * 2);
  i := 1;
  j := 1;
  while i <= len do begin
    if ws[i] = #10 then begin
      case FLineEnds of
        lsCR: begin
          Result[j] := #13;
          inc(j);
        end;
        lsLF: begin
          Result[j] := #10;
          inc(j);
        end;
        lsCRLF: begin
          Result[j] := #13;
          inc(j);
          Result[j] := #10;
          inc(j);
        end;
      end;
    end else begin
      Result[j] := ws[i];
      inc(j);
    end;
    inc(i);
  end;//while
  SetLength(Result, j - 1);
end;

function TNotePadString.GetTextCR: WideString;
var
  ABuffer: TStringBuffer;
  i: integer;
  ALine: PLineData;
begin
  ABuffer := TStringBuffer.Create(FBuffer.Capacity);
  for i := 0 to FStrings.Count - 1 do begin
    ALine := FStrings[i];
    case ALine.LineEnd of
      leLF:
        ABuffer.Append(FBuffer.ToString(ALine.Start, ALine.Length) + #13);
      else
        ABuffer.Append(FBuffer.ToString(ALine.Start, ALine.Length));
    end;
  end;
  Result := ABuffer.ToString;
  ABuffer.Free;
end;

function TNotePadString.GetTextCRLF: WideString;
var
  ABuffer: TStringBuffer;
  i: integer;
  ALine: PLineData;
begin
  ABuffer := TStringBuffer.Create(FBuffer.Capacity);
  for i := 0 to FStrings.Count - 1 do begin
    ALine := FStrings[i];
    case ALine.LineEnd of
      leLF:
        ABuffer.Append(FBuffer.ToString(ALine.Start, ALine.Length) + #13#10);
      else
        ABuffer.Append(FBuffer.ToString(ALine.Start, ALine.Length));
    end;
  end;
  Result := ABuffer.ToString;
  ABuffer.Free;
end;

function TNotePadString.GetTextLength: integer;
begin
  Result := FBuffer.StringLength;
end;

function TNotePadString.GetTextLF: WideString;
var
  ABuffer: TStringBuffer;
  i: integer;
  ALine: PLineData;
begin
  ABuffer := TStringBuffer.Create(FBuffer.Capacity);
  for i := 0 to FStrings.Count - 1 do begin
    ALine := FStrings[i];
    case ALine.LineEnd of
      leLF:
        ABuffer.Append(FBuffer.ToString(ALine.Start, ALine.Length) + #10);
      else
        ABuffer.Append(FBuffer.ToString(ALine.Start, ALine.Length));
    end;
  end;
  Result := ABuffer.ToString;
  ABuffer.Free;
end;

procedure TNotePadString.InputBackKey;
var
  ALine: PLineData;
  AUndo: TUndoData;
  AStoreBox: boolean;
begin
  if FStart.Position <> FEnd.Position then begin
    SelText := '';
    exit;
  end;
  FEnd := FStart;//ɂ͂邯ǂ́H
  while FStart.Position = FEnd.Position do begin
    if FEnd.X = 0 then begin
      if FEnd.Y = 0 then begin
        FEnd := FStart;
        exit;
      end;
      ALine := FStrings[FEnd.Y - 1];
      case ALine.LineEnd of
        leLF: SelEnd := SelEnd - 1;
        else begin//wrap̍ŏIn_ꍇ
          dec(FEnd.Y);
          FEnd.X := ALine.Length;
        end;
      end;
    end else SelEnd := SelEnd - 1;
  end;
  BeginUpdate;
  AUndo.MarksBefore := Copy(FMarks);
  AUndo.Position := FEnd.Position;
  AUndo.SelStartBefore := FStart.Position;
  AUndo.SelEndBefore := FStart.Position;
  AUndo.SelStartAfter := FEnd.Position;
  AUndo.SelEndAfter := FEnd.Position;
  AStoreBox := FBoxMode;
  FBoxMode := false;
  AUndo.Deleted := SelText;
  FBoxMode := AStoreBox;
  AUndo.Added := '';
  DeleteString(false);
  InsertString('', false);
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndBefore := FLineEnds;
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.InputDelKey;
var
  ALine: PLineData;
  AUndo: TUndoData;
  AStoreBox: boolean;
begin
  if FStart.Position <> FEnd.Position then begin
    SelText := '';
    exit;
  end;
  //SelEndtoSelStart;//ɂ͂Ȃǂ́H
  ALine := FStrings[FEnd.Y];
  while FStart.Position = FEnd.Position do begin
    if FEnd.X = ALine.Length then begin
      case ALine.LineEnd of
        leLF: SelEnd := SelEnd + 1;
        leNone: begin
          FEnd := FStart;
          exit;
        end;
        else begin//wrap̍ŏIn_ꍇ
          inc(FEnd.Y);
          FEnd.X := 0;
          ALine := FStrings[FEnd.Y];
        end;
      end;
    end else SelEnd := SelEnd + 1;
  end;
  BeginUpdate;
  AUndo.MarksBefore := Copy(FMarks);
  AUndo.Position := FStart.Position;
  AUndo.SelStartBefore := FStart.Position;
  AUndo.SelEndBefore := FStart.Position;
  AUndo.SelStartAfter := FEnd.Position;
  AUndo.SelEndAfter := FEnd.Position;
  AStoreBox := FBoxMode;
  FBoxMode := false;
  AUndo.Deleted := SelText;
  FBoxMode := AStoreBox;
  AUndo.Added := '';
  DeleteString(false);
  InsertString('', false);
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndBefore := FLineEnds;
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.InsertString(Value: WideString; ABox: boolean);
var
  i, j: integer;
  C: WideChar;
  AStart: TCaretPosition;
  AX: integer;
  LineInsertCount: integer;
  ANeedInsert: boolean;
  ABoxText: WideString;
  over1: boolean;
  InsLength: integer;
  NewSelStart: integer;
  ALine: PLineData;
  NewLine: TLineData;
begin
  if ABox then begin
    AStart := FStart;
    AX := GetCaretX(FStart.X, FStart.Y);
    j := 0;
    LineInsertCount := 0;
    for i := 0 to Length(Value) do begin
      ANeedInsert := i = Length(Value);
      if not ANeedInsert then begin
        C := Value[i + 1];
        if C = #10 then
          ANeedInsert := true;
      end;
      if ANeedInsert then begin
        if i > j then begin
          ABoxText := copy(Value, j + 1, i - j);
          while AStart.Y + LineInsertCount >= FStrings.Count do begin
            SelStart := FBuffer.StringLength;
            InsertString(#10, false);
          end;
          SelStartPos := Point(
            FromCaretX(AX, AStart.Y + LineInsertCount, over1),
            AStart.Y + LineInsertCount);
          InsertString(ABoxText, false);
        end;
        j := i + 1;
        inc(LineInsertCount);
      end;//if
    end;//for
  end else begin
    InsLength := Length(Value);
    if InsLength = 0 then exit;
    FBuffer.Insert(FStart.Position, Value, InsLength);
    MarksOnInsertString(FStart.Position, InsLength);
    FChangedStart := Min(FChangedStart, FStart.Y);
    NewSelStart := FStart.Position + InsLength;
    ALine := FStrings[FStart.Y];
    ALine.NeedWrap := true;
    inc(ALine.Length, InsLength);
    j := ALine.Start;
    for i := FStart.Position to NewSelStart - 1 do begin
      C := FBuffer[i];
      if C = #10 then begin
        NewLine.NeedWrap := true;
        NewLine.Start := ALine.Start;
        NewLine.Length := i - j;
        NewLine.LineEnd := leLF;
        FStrings.Insert(FStart.Y, NewLine);
        inc(ALine.Start, i - j + 1);
        dec(ALine.Length, i - j + 1);
        inc(FStart.Y);
        j := i + 1;
      end;//cr
    end;//for
    for i := FStart.Y + 1 to FStrings.Count - 1 do begin
      inc(FStrings[i].Start, InsLength);
    end;
    FStart.Position := NewSelStart;
    FStart.X := FStart.Position - FStrings[FStart.Y].Start;
    FEnd := FStart;
    FChangedEnd := Max(FChangedEnd, FStart.Y);
  end;
  if not FModified then begin
    FModified := true;
    if Assigned(FOnModified) then FOnModified(self);
  end;
end;

function TNotePadString.LinesPosToWrapPos(X, Y: integer): TPoint;
var
  ALine: PLineData;
  i, j: integer;
begin
  Result := Point(0, -1);
  j := 0;
  for i := 0 to FStrings.Count - 1 do begin
    if j = Y then begin
      Result.Y := i;
      break;
    end;
    ALine := FStrings[i];
    case ALine.LineEnd of
      leLF: inc(j);
    end;
  end;
  if Result.Y = -1 then begin
    result.Y := FStrings.Count - 1;
    result.X := FStrings[result.Y].Length;
    exit;
  end;
  while true do begin
    ALine := FStrings[result.Y];
    if ALine.Length < X then begin
      dec(X, ALine.Length);
      inc(result.Y);
    end else begin
      result.X := X;
      break;
    end;
  end;
end;

procedure TNotePadString.MarksRemoveRange(AStart, ALength: integer);
var
  ps, pd: Cardinal;
begin
  if AStart + ALength < Length(FMarks) then begin
    pd := Cardinal(@(FMarks[0])) + Cardinal(AStart) * 4;
    ps := pd + Cardinal(ALength) * 4;
    Move(Pointer(ps)^, Pointer(pd)^, (Length(FMarks)  - ALength - AStart) * 4);
  end;
  SetLength(FMarks, Length(FMarks)  - ALength);
end;

procedure TNotePadString.MarksOnInsertString(AStart, ALength: integer);
var
  i: integer;
begin
  for i := 0 to Length(FMarks) - 1 do begin
    //AStart > FMarks[i] ̓PAȂ
    if AStart <= FMarks[i] then inc(FMarks[i], ALength);
    while FMarks[i] > 0 do begin
      if FMarks[i] < FBuffer.StringLength then begin
        case FBuffer[FMarks[i] - 1] of
          #10: break;
        end;
      end;
      dec(FMarks[i])
    end;
  end;
end;

procedure TNotePadString.MarksOnDeleteString(AStart, ALength: integer);
var
  i: integer;
begin
  if ALength = 0 then exit;
  i := 0;
  while i < Length(FMarks) do begin
    //AStart > FMarks[i] ̓PAȂ
    if (AStart + ALength) < FMarks[i] then begin
      FMarks[i] := FMarks[i] - ALength;
    end else if AStart <= FMarks[i] then begin
      FMarks[i] := AStart;
      while FMarks[i] > 0 do begin
        if FMarks[i] < FBuffer.StringLength then begin
          case FBuffer[FMarks[i] - 1] of
            #10: break;
          end;
        end;
        dec(FMarks[i])
      end;
    end;
    if (i > 0) and (FMarks[i] = FMarks[i -1]) then
      MarksRemoveRange(i, 1)
    else
      inc(i);
  end;
end;

procedure TNotePadString.Redo;
var
  AUndo: TUndoData;
begin
  if not CanRedo then exit;
  BeginUpdate;
  AUndo := FUndoList.GetLastRedo;
  SelStart := AUndo.Position;
  SelEnd := FStart.Position + Length(AUndo.Deleted);
  DeleteString(false);
  InsertString(AUndo.Added, false);
  FUndoList.Redo;
  SelStart := AUndo.SelStartAfter;
  SelEnd := AUndo.SelEndAfter;
  FLineEnds := AUndo.LineEndAfter;
  RedrawCurrent;
  FMarks := Copy(AUndo.MarksAfter);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.RedrawCurrent;
begin
  FChangedStart := Min(FChangedStart, Min(FStart.Y, FEnd.Y));
  FChangedEnd := Max(FChangedEnd, Max(FStart.Y, FEnd.Y));
end;

procedure TNotePadString.ReplaseWithoutSelect(const Value: WideString; AStart,
  ALength: integer);
var
  AUndo: TUndoData;
  L: integer;
  ws: WideString;
  i, j: integer;
begin
  L := Length(Value);
  SetLength(ws, L);
  i := 1; //]Ƃ
  j := 1; //K
  while i <= L do begin
    case Value[i] of
      #13:begin
        ws[j] := #10;
        inc(j);
        inc(i);
        if (i <= L) and (Value[i] = #10) then begin
          inc(i);
        end;
      end;
      #10:begin
        ws[j] := #10;
        inc(j);
        inc(i);
      end;
      else begin
        ws[j] := Value[i];
        inc(j);
        inc(i);
      end;
    end;
  end;
  SetLength(ws, j - 1);
  BeginUpdate;
  AUndo.SelStartBefore := FStart.Position;
  AUndo.SelEndBefore := FEnd.Position;
  SelStart := AStart;
  SelEnd := AStart + ALength;
  AUndo.MarksBefore := Copy(FMarks);
  AUndo.Position := Min(FStart.Position, FEnd.Position);
  AUndo.SelStartAfter := AUndo.Position + Length(Value);
  AUndo.SelEndAfter := AUndo.SelStartAfter;
  AUndo.Deleted := SelText;
  AUndo.Added := ws;
  DeleteString(false);
  InsertString(ws, false);
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndBefore := FLineEnds;
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.SelEndToSelStart;
begin
  FEnd := FStart;
end;

procedure TNotePadString.SetItems(Index: integer; data: PLineData);
begin
  FStrings[Index] := data;
end;

procedure TNotePadString.SetSelEnd(Value: integer);
var
  NewPos, NewX, NewY: integer;
  ALine: PLineData;
begin
  if Value < FBuffer.StringLength then begin
    NewPos := Value;
  end else NewPos := FBuffer.StringLength;
  if NewPos = FEnd.Position then exit;
  if NewPos = FStart.Position then begin
    FEnd := FStart;
    exit;
  end;
  NewX := FEnd.X + NewPos - FEnd.Position;
  NewY := FEnd.Y;
  ALine := FStrings[NewY];
  while (ALine.Length < NewX) do begin
    dec(NewX, ALine.Length);
    case ALine.LineEnd of
      leLF: dec(NewX);
    end;
    inc(NewY);
    ALine := FStrings[NewY];
  end;
  while NewX < 0 do begin
    dec(NewY);
    ALine := FStrings[NewY];
    inc(NewX, ALine.Length);
    case ALine.LineEnd of
      leLF: inc(NewX);
    end;
  end;
  FEnd.Position := NewPos;
  FEnd.X := NewX;
  FEnd.Y := NewY;
end;

procedure TNotePadString.SetSelEndPos(Value: TPoint);
var
  NewY, NewX: integer;
  ALine: PLineData;
begin
  NewY := Min(Value.Y, FStrings.Count - 1);
  ALine := FStrings[NewY];
  NewX := Min(ALine.Length, Value.X);
  FEnd.X := NewX;
  FEnd.Y := NewY;
  FEnd.Position := ALine.Start + NewX;
end;

procedure TNotePadString.SetSelStart(Value: integer);
var
  NewPos, NewX, NewY: integer;
  ALine: PLineData;
begin
  if Value < FBuffer.StringLength then begin
    NewPos := Value;
  end else NewPos := FBuffer.StringLength;
  if NewPos = FStart.Position then exit;
  if NewPos = FEnd.Position then begin
    FStart := FEnd;
    exit;
  end;
  NewX := FStart.X + NewPos - FStart.Position;
  NewY := FStart.Y;
  ALine := FStrings[NewY];
  while (ALine.Length < NewX) do begin
    dec(NewX, ALine.Length);
    case ALine.LineEnd of
      leLF: dec(NewX);
    end;
    inc(NewY);
    ALine := FStrings[NewY];
  end;
  while NewX < 0 do begin
    dec(NewY);
    ALine := FStrings[NewY];
    inc(NewX, ALine.Length);
    case ALine.LineEnd of
      leLF: inc(NewX);
    end;
  end;
  FStart.Position := NewPos;
  FStart.X := NewX;
  FStart.Y := NewY;
end;

procedure TNotePadString.SetSelStartPos(Value: TPoint);
var
  NewY, NewX: integer;
  ALine: PLineData;
begin
  NewY := Min(Value.Y, FStrings.Count - 1);
  ALine := FStrings[NewY];
  NewX := Min(ALine.Length, Value.X);
  FStart.X := NewX;
  FStart.Y := NewY;
  FStart.Position := ALine.Start + NewX;
end;

procedure TNotePadString.SetSelText(const Value: WideString);
var
  AUndo: TUndoData;
  AStart, AEnd: TCaretPosition;
  L: integer;
  ws: WideString;
  i, j: integer;
begin
  L := Length(Value);
  SetLength(ws, L);
  i := 1; //]Ƃ
  j := 1; //K
  while i <= L do begin
    case Value[i] of
      #13:begin
        ws[j] := #10;
        inc(j);
        inc(i);
        if (i <= L) and (Value[i] = #10) then begin
          inc(i);
        end;
      end;
      #10:begin
        ws[j] := #10;
        inc(j);
        inc(i);
      end;
      else begin
        ws[j] := Value[i];
        inc(j);
        inc(i);
      end;
    end;
  end;
  SetLength(ws, j - 1);
  BeginUpdate;
  AUndo.MarksBefore := Copy(FMarks);
  if FBoxMode then begin
    if FStart.Y < FEnd.Y then begin
      AStart := FStart;
      AEnd := FEnd;
    end else if FEnd.Y < FStart.Y then begin
      AStart := FEnd;
      AEnd := FStart;
    end else if FStart.X < FEnd.X then begin
      AStart := FStart;
      AEnd := FEnd;
    end else if FEnd.X < FStart.X then begin
      AStart := FEnd;
      AEnd := FStart;
    end else begin
      AStart := FStart;
      AEnd := FEnd;
    end;
    L := FStrings[AStart.Y].Start;
    AUndo.Position := L;
    AUndo.Deleted := FBuffer.ToString(L, FBuffer.StringLength - L);
    AUndo.SelStartBefore := FStart.Position;
    AUndo.SelEndBefore := FEnd.Position;
    DeleteString(true);
    InsertString(ws, false);
    AUndo.Added := FBuffer.ToString(L, FBuffer.StringLength - L);
    AUndo.SelStartAfter := FStart.Position;
    AUndo.SelEndAfter := FStart.Position;
  end else begin
    AUndo.Position := Min(FStart.Position, FEnd.Position);
    AUndo.SelStartAfter := AUndo.Position + Length(Value);
    AUndo.SelEndAfter := AUndo.SelStartAfter;
    AUndo.Deleted := SelText;
    AUndo.Added := ws;
    AUndo.SelStartBefore := FStart.Position;
    AUndo.SelEndBefore := FEnd.Position;
    DeleteString(false);
    InsertString(ws, false);
  end;
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndBefore := FLineEnds;
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.SetSelTextBox(const Value: WideString);
var
  AStart, AEnd: TCaretPosition;
  L: integer;
  AUndo: TUndoData;
  ws: WideString;
  i, j: integer;
begin
  L := Length(Value);
  SetLength(ws, L);
  i := 1; //]Ƃ
  j := 1; //K
  while i <= L do begin
    case Value[i] of
      #13:begin
        ws[j] := #10;
        inc(j);
        inc(i);
        if (i <= L) and (Value[i] = #10) then begin
          inc(i);
        end;
      end;
      #10:begin
        ws[j] := #10;
        inc(j);
        inc(i);
      end;
      else begin
        ws[j] := Value[i];
        inc(j);
        inc(i);
      end;
    end;
  end;
  SetLength(ws, j - 1);
  BeginUpdate;
  AUndo.MarksBefore := Copy(FMarks);
  if FStart.Y < FEnd.Y then begin
    AStart := FStart;
    AEnd := FEnd;
  end else if FEnd.Y < FStart.Y then begin
    AStart := FEnd;
    AEnd := FStart;
  end else if FStart.X < FEnd.X then begin
    AStart := FStart;
    AEnd := FEnd;
  end else if FEnd.X < FStart.X then begin
    AStart := FEnd;
    AEnd := FStart;
  end else begin
    AStart := FStart;
    AEnd := FEnd;
  end;
  L := FStrings[AStart.Y].Start;
  AUndo.Position := L;
  AUndo.Deleted := FBuffer.ToString(L, FBuffer.StringLength - L);
  AUndo.SelStartBefore := FStart.Position;
  AUndo.SelEndBefore := FEnd.Position;
  DeleteString(FBoxMode);
  InsertString(ws, true);
  AUndo.Added := FBuffer.ToString(L, FBuffer.StringLength - L);
  AUndo.SelStartAfter := FStart.Position;
  AUndo.SelEndAfter := FStart.Position;
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndBefore := FLineEnds;
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  WrapFast;
  EndUpdate;
end;

procedure TNotePadString.SetTabCount(Value: integer);
begin
  if (Value <> 2) and (Value <> 4) and (Value <> 6) then Value := 8;
  if Value = FTabCount then exit;
  BeginUpdate();
  FChangedStart := 0;
  FChangedEnd := FStrings.Count - 1;
  FTabCount := Value;
  FNeedWrapAll := true;
  EndUpdate();
end;

procedure TNotePadString.SetText(const Value: WideString);
var
  AUndo: TUndoData;
  ws: WideString;
  i, j, len: integer;
begin
  AUndo.LineEndBefore := FLineEnds;
  len := Length(Value);
  SetLength(ws, len);
  i := 1; //]Ƃ
  j := 1; //K
  while i <= len do begin
    case Value[i] of
      #13:begin
        ws[j] := #10;
        inc(j);
        inc(i);
        if (i <= len) and (Value[i] = #10) then begin
          inc(i);
          FLineEnds := lsCRLF;
        end else FLineEnds := lsCR;
      end;
      #10:begin
        ws[j] := #10;
        inc(j);
        inc(i);
        FLineEnds := lsLF;
      end;
      else begin
        ws[j] := Value[i];
        inc(j);
        inc(i);
      end;
    end;
  end;
  SetLength(ws, j - 1);
  BeginUpdate;
  AUndo.MarksBefore := Copy(FMarks);
  AUndo.Position := 0;
  AUndo.SelStartBefore := FStart.Position;
  AUndo.SelEndBefore := FEnd.Position;
  AUndo.SelStartAfter := 0;
  AUndo.SelEndAfter := 0;
  AUndo.Deleted := Text;
  AUndo.Added := ws;
  ClearString;
  InsertString(ws, false);
  AUndo.MarksAfter := Copy(FMarks);
  AUndo.LineEndAfter := FLineEnds;
  FUndoList.Add(AUndo);
  SelStart := 0;
  FEnd := FStart;
  FNeedWrapAll := true;
  EndUpdate;
end;

procedure TNotePadString.SetWrapByte(Value: integer);
begin
  Value := Max(MIN_WRAPBYTE, Min(MAX_WRAPBYTE, Value));
  BeginUpdate;
  FChangedStart := 0;
  FChangedEnd := FStrings.Count - 1;
  FWrapByte := Value;
  FNeedWrapAll := true;
  EndUpdate;
end;

procedure TNotePadString.Undo;
var
  AUndo: TUndoData;
begin
  if not CanUndo then exit;
  BeginUpdate;
  AUndo := FUndoList.GetLastUndo;
  SelStart := AUndo.Position;
  SelEnd := FStart.Position + Length(AUndo.Added);
  DeleteString(false);
  InsertString(AUndo.Deleted, false);
  FUndoList.Undo();
  SelStart := AUndo.SelStartBefore;
  SelEnd := AUndo.SelEndBefore;
  FLineEnds := AUndo.LineEndBefore;
  RedrawCurrent;
  WrapFast;
  FMarks := Copy(AUndo.MarksBefore);
  EndUpdate;
end;

procedure TNotePadString.WrapAll;
var
  AList: TLineList;
  i, sum, suminc: integer;
  ALine: PLineData;
  NewLine: PLineData;
  BLine: TLineData;
  LoopFlag1: boolean;
  AX, AX2: integer;
  C: WideChar;
begin
  AList := TLineList.Create(FStrings.Count);
  i := 0;
  while i < FStrings.Count do begin
    NewLine := FStrings[i];
    while NewLine.LineEnd = leWrap do begin
      inc(i);
      ALine := FStrings[i];
      inc(NewLine.Length, ALine.Length);
      NewLine.LineEnd := ALine.LineEnd;
    end;
    repeat
      LoopFlag1 := false;
      AX := 0;
      sum := 0;
      while AX < NewLine.Length do begin
        C := FBuffer[AX + NewLine.Start];
        if C = #$0009 then begin
          suminc := FTabCount * GetGlyphAveWidth - (sum mod (FTabCount * GetGlyphAveWidth));
          if suminc < GetGlyphAveWidth then inc(suminc, FTabCount * GetGlyphAveWidth);
        end else suminc := GetGlyphWidth(C);
        inc(sum, suminc);
        if sum > FWrapByte * GetGlyphAveWidth then begin
          if not IsPuncChar(C) then begin
            //sԂ牺Ȃꍇ͌炷
            dec(sum, suminc);
            dec(AX);
          end;// not punc
          AX2 := AX;
          if AX + NewLine.Start + 1 < FBuffer.StringLength then
          while true do begin
            //obNE𒴂猳̈ʒu֖߂ĔB
            if sum <= FWrapByte * GetGlyphAveWidth div 2 then begin
              AX := AX2;
              break;
            end;
            C := FBuffer[AX + NewLine.Start];
            if CheckLineBreak(C, FBuffer[AX + NewLine.Start + 1]) then break;
            if C = #$0009 then begin
              suminc := FTabCount * GetGlyphAveWidth - (sum mod (FTabCount * GetGlyphAveWidth));
              if suminc < GetGlyphAveWidth then inc(suminc, FTabCount * GetGlyphAveWidth);
            end else suminc := GetGlyphWidth(C);
            dec(sum, suminc);
            dec(AX);
          end;
          if AX < NewLine.Length - 1 then begin
            BLine.LineEnd := NewLine.LineEnd;
            BLine.Length := NewLine.Length - (AX + 1);
            BLine.Start := NewLine.Start + AX + 1;
            NewLine.Length := AX + 1;
            NewLine.LineEnd := leWrap;
            NewLine.NeedWrap := false;
            AList.Add(NewLine^);
            NewLine^ := BLine;
            LoopFlag1 := true;
          end;
          break;
        end;// sum>FWrapByte
        inc(AX);
      end;//while ax<newline.length
      if not LoopFlag1 then begin
        NewLine.NeedWrap := false;
        AList.Add(NewLine^);
      end;
    until not LoopFlag1;
    inc(i);
  end;//while
  FStrings.Free;
  FStrings := AList;
  sum := FStart.Position;
  for i := 0 to FStrings.Count - 1 do begin
    ALine := FStrings[i];
    if sum <= ALine.Length then begin
      FStart.Y := i;
      FStart.X := sum;
      break;
    end;
    case ALine.LineEnd of
      leLF: dec(sum, ALine.Length + 1);
      else dec(sum, ALine.Length);
    end;
  end;
  if FStart.Position = FEnd.Position then
    FEnd := FStart
  else begin
    sum := FEnd.Position;
    for i := 0 to FStrings.Count - 1 do begin
      ALine := FStrings[i];
      if sum <= ALine.Length then begin
        FEnd.Y := i;
        FEnd.X := sum;
        break;
      end;
      case ALine.LineEnd of
        leLF: dec(sum, ALine.Length + 1);
        else dec(sum, ALine.Length);
      end;
    end;
  end;
  FChangedStart := 0;
  FChangedEnd := Max(FChangedEnd, FStrings.Count - 1);
  if Assigned(FOnNeedParse) then FOnNeedParse(self);
  FParseDone := true;
end;

procedure TNotePadString.WrapFast;
var
  i, j, sum, suminc, FirstFound, LastFound: integer;
  DoneNeedWrap: boolean;
  NewLine: PLineData;
  ALine: TLineData;
  NeedRemove, NeedRemoveSum: integer;
  AX, AX2: integer;
  C: WideChar;
begin
  DoneNeedWrap := false;
  FirstFound := -1;
  LastFound := 1;
  for i := Max(1, FChangedStart) to Min(FStrings.Count - 1, FChangedEnd) do begin
    if FStrings[i].NeedWrap then begin
      if FirstFound = -1 then FirstFound := i;
      LastFound := i;
      j := i - 1;
      NewLine := FStrings[j];
      if (NewLine.LineEnd = leWrap) and (not NewLine.NeedWrap) then begin
        NewLine.NeedWrap := true;
        if FChangedStart >= j then FChangedStart := j;
        if FirstFound >= j then FirstFound := j;
      end;
    end;
  end;
  if FirstFound < 2 then FirstFound := FChangedStart;//1sOĂ邵A-1A
  i := FirstFound;
  repeat
    NewLine := FStrings[i];
    if i > 0 then NewLine.Start := FStrings[i - 1].Start + GetLineLength(i - 1);
    if NewLine.NeedWrap then begin
      NewLine.NeedWrap := false;
      NeedRemove := 0;
      while NewLine.LineEnd = leWrap do begin
        ALine := FStrings[i + 1 + NeedRemove]^;
        inc(NewLine.Length, ALine.Length);
        NewLine.LineEnd := ALine.LineEnd;
        inc(NeedRemove);
      end;
      NeedRemoveSum := NeedRemove;
      AX := 0;
      sum := 0;
      while AX < NewLine.Length do begin
        C := FBuffer[AX + NewLine.Start];
        if C = #$0009 then begin
          suminc := FTabCount * GetGlyphAveWidth - (sum mod (FTabCount * GetGlyphAveWidth));
          if suminc < GetGlyphAveWidth then inc(suminc, FTabCount * GetGlyphAveWidth);
        end else suminc := GetGlyphWidth(C);
        inc(sum, suminc);
        if sum > FWrapByte * GetGlyphAveWidth then begin
          if not IsPuncChar(C) then begin
            //sԂ牺Ȃꍇ͌炷
            dec(sum, suminc);
            dec(AX);
          end;// not punc
          AX2 := AX;
          if AX + NewLine.Start + 1 < FBuffer.StringLength then
          while true do begin
            //obNE𒴂猳̈ʒu֖߂ĔB
            if sum <= FWrapByte * GetGlyphAveWidth div 2 then begin
              AX := AX2;
              break;
            end;
            C := FBuffer[AX + NewLine.Start];
            if CheckLineBreak(C, FBuffer[AX + NewLine.Start + 1]) then break;
            if C = #$0009 then begin
              suminc := FTabCount * GetGlyphAveWidth - (sum mod (FTabCount * GetGlyphAveWidth));
              if suminc < GetGlyphAveWidth then inc(suminc, FTabCount * GetGlyphAveWidth);
            end else suminc := GetGlyphWidth(C);
            dec(sum, suminc);
            dec(AX);
          end;
          if AX < NewLine.Length - 1 then begin
            ALine.NeedWrap := true;
            ALine.LineEnd := NewLine.LineEnd;
            ALine.Length := NewLine.Length - (AX + 1);
            ALine.Start := NewLine.Start + AX + 1;
            if NeedRemove > 0 then begin
              FStrings[i + 1]^ := ALine;
              dec(NeedRemove);
            end else begin
              FStrings.Insert(i + 1, ALine);
              inc(LastFound);
            end;
            NewLine.Length := AX + 1;
            NewLine.LineEnd := leWrap;
          end;
          break;
        end;//if (sum > FWrapByte)
        inc(AX);
      end;//while (AX < NewLine.Length)
      if NeedRemove > 0 then FStrings.RemoveRange(i + 1 + NeedRemoveSum - NeedRemove, NeedRemove);
      DoneNeedWrap := true;
    end;//if (NewLine.NeedWrap)
    inc(i);
  until (i >= FStrings.Count) or (i > LastFound + 1);
  //Ń[hbvJ[\ʒu낦Ȃ
  if DoneNeedWrap then begin
    sum := FStart.Position;
    for i := 0 to FStrings.Count - 1 do begin
      NewLine := FStrings[i];
      if sum <= NewLine.Length then begin
        FStart.Y := i;
        FStart.X := sum;
        break;
      end;
      case NewLine.LineEnd of
        leLF: dec(sum, NewLine.Length + 1);
        else dec(sum, NewLine.Length);
      end;
    end;
    if FStart.Position = FEnd.Position then
      FEnd := FStart
    else begin
      sum := FEnd.Position;
      for i := 0 to FStrings.Count - 1 do begin
        NewLine := FStrings[i];
        if sum <= NewLine.Length then begin
          FEnd.Y := i;
          FEnd.X := sum;
          break;
        end;
        case NewLine.LineEnd of
          leLF: dec(sum, NewLine.Length + 1);
          else dec(sum, NewLine.Length);
        end;
      end;
    end;
  end;//if DoneNeedWrap
  FChangedStart := Min(FChangedStart, Min(FStart.Y, FEnd.Y));
  FChangedEnd := Max(FChangedEnd, Max(FStart.Y, FEnd.Y));
  if Assigned(FOnNeedParse) then FOnNeedParse(self);
  FParseDone := true;
end;

end.


