//GfB^GWTNotePad
//zz
//http://wantech.ikuto.com/

unit NotePadView;

interface

uses Windows, Messages, Classes, Graphics, Controls, ActiveX, Dialogs,
  UniControl, NotePadString, NotePadMode, DragDropClasses;

type
  TDropFileEvent = procedure(Sender: TObject; Files: array of WideString) of object;
  TRulerEvent = procedure(Sender: TObject; Col: integer) of object;

  TNotePadOptions = record
    AutoIndent: integer;
    ColorBack: TColor;
    ColorCaretLine: TColor;
    ColorDBSpace: TColor;
    ColorEOF: TColor;
    ColorFore: TColor;
    ColorLineEnd: TColor;
    ColorLeftBar: TColor;
    ColorRuler: TColor;
    ColorSBSpace: TColor;
    ColorTab: TColor;
    ColorWrap: TColor;
    ColorWrapLine: TColor;
    ColorWideChar: TColor;
    ColorMark: TColor;
    Font: TFont;
    LeftBarCountWrap: boolean;
    LineMargin: integer;
    ShowDBSpace: boolean;
    ShowEOF: boolean;
    ShowLeftBar: boolean;
    ShowLineEnd: boolean;
    ShowRuler: boolean;
    ShowSBSpace: boolean;
    ShowTab: boolean;
    ShowWrap: boolean;
    ShowWrapLine: boolean;
    TabCount: integer;
    BackGroundBMP: TGraphic;
    AlphaBlend: integer;
    DrawNot: boolean;
    KeepCol: boolean;
    RulerEnabled: boolean;

    ColorControl,
    ColorHighlight,
    ColorSymbol,
    ColorString,
    ColorInteger,
    ColorComment,
    ColorReserve,
    ColorReserve2,
    ColorReserve3,
    ColorAnk,
    ColorAlphabetic,
    ColorDBSymbol,
    ColorDBInt,
    ColorDBAlph,
    ColorDBHira,
    ColorDBKana,
    ColorDBKanji,
    ColorSBKana,
    ColorUnicode,
    ColorLink,
    ColorHtmlTag,
    ColorPrefix,
    ColorHere:
    TColor;
    StyleControl,
    StyleHighlight,
    StyleSymbol,
    StyleString,
    StyleInteger,
    StyleComment,
    StyleReserve,
    StyleReserve2,
    StyleReserve3,
    StyleAnk,
    StyleAlphabetic,
    StyleDBSymbol,
    StyleDBInt,
    StyleDBAlph,
    StyleDBHira,
    StyleDBKana,
    StyleDBKanji,
    StyleSBKana,
    StyleUnicode,
    StyleLink,
    StyleHtmlTag,
    StylePrefix,
    StyleHere:
    TFontStyles;
  end;

  TWordWrap = (wwNone, wwCount, wwRight);

  THighlightMode = (hmDefault, hmHtml, hmCSharp, hmDelphi, hmJScript,
                    hmVBScript, hmPython, hmCpp, hmJava, hmPerl,
                    hmRuby, hmVisualBasic, hmCss, hmBat);

type
  TCodePage = (cpJapanese = 1, cpAscii = 2, cp1251 = 4,
               cpBig5 = 8, cpGB2312 = 16, cpEUCKR = 32,
               cpEUCJP = 64, cpNone = 0);

  TNotePad = class(TTntCustomControl, IDropTarget)
  private
    FNotePadString: TNotePadString;
    FFirstVisiblePix: integer;
    FFirstVisibleLine: integer;
    FLeftBarWidth: integer;
    FWordWrap: TWordWrap;
    FWordWrapCount: integer;
    FShowLeftBar: integer;
    FShowRuler: integer;
    FModified: boolean;
    FColKeeping: integer;
    FCaretHided: boolean;
    FDoubleClicked: boolean;
    FFocusClick: boolean;
    FDragCancel: boolean;
    FDDString: TDDString;
    FDDEnter: IDataObject;
    FDDSource: TDDSource;
    FDDReady: boolean;
    FDDEffect: integer;
    FDragCaret: TPoint;
    FDraging: boolean;
    FFontWidth: integer;
    FLineHeight: integer;
    FAscent: integer;
    FDescent: integer;
    FLeading: integer;
    FLineSelect: integer;
    FRulerDrag: boolean;
    FVisibleByteCount: integer;
    FVisibleLineCount: integer;
    FCurrentMode: TNotePadMode;
    FDocumentParent: TNotePad;
    FDocumentChild: TNotepad;
    FSaveMode: TNotePadMode;
    FCodePage: TCodePage;
    FInitOnHandle: boolean;
    FFontChanged: boolean;
    FForceScroll: boolean;
    FCompleting: boolean;
    FMouseWheelDelta: Smallint;

    FFont: TFont;
    FSearchText: WideString;
    FSearchIgCase: boolean;
    FFormatEtc: TFormatEtc;
    FAlphaBlend: byte;
    FSubSelectStart, FSubSelectEnd: integer;

    //Cxg
    FViewChanged: TNotifyEvent;
    FModifiedChanged: TNotifyEvent;
    FFilesDroped: TDropFileEvent;
    FRulerDown, FRulerMove, FRulerUp: TRulerEvent;

    procedure CalcFont;
    procedure CalcLeftBar;
    procedure initScroll;
    function NeedWrap: boolean;
    procedure initCaretPosition;
    procedure DrawCanvas(ACanvas: TCanvas; ARgnData: PRgnData);
    procedure DrawLine(ALineNum: integer; ACaretLine: boolean; parser: TNotePadParser; ACanvas: TCanvas);
    procedure DrawRuler(ACanvas: TCanvas; r: TRect);
    procedure CaretReset;
    procedure CaretHide;
    procedure CaretShow;
    procedure SelectWordSub(X, Y: integer; out APrevious, AStart, ALength: integer; out canJump: boolean);
    function DoubleClickDown(X: integer): boolean;
    function PointToPos(X, Y: integer; Split: boolean): TPoint;
    function CreateBitmap: boolean;
    procedure InitBitmap;
    procedure InputChar(C: Word);
    procedure TrackMouseEventSub;

    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMImeRequest(var Msg: TMessage); message WM_IME_REQUEST;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMImeChar(var Message: TWMChar); message WM_IME_CHAR;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
    procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBlCLK;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMMouseHover(var Message: TWMMouse); message WM_MOUSEHOVER;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    //nh
    procedure StringOnNeedParse(Sender: TObject);
    procedure StringOnChanged(Sender: TObject);
    procedure StringOnModified(Sender: TObject);
    //vpeB
    function GetBoxSelectMode: boolean;
    procedure SetBoxSelectMode(Value: boolean);
    function GetSelectionLength: integer;
    procedure SetSelectionLength(Value: integer);
    function GetSelectionStart: integer;
    procedure SetSelectionStart(Value: integer);
    function GetEditMode: THighlightMode;
    procedure SetEditMode(Value: THighlightMode);
    function GetOptions: TNotePadOptions;
    procedure SetOptions(const Value: TNotePadOptions);
    procedure SetWordWrapCount(Value: integer);
    procedure SetWordWrap(Value: TWordWrap);
    procedure SetModified(Value: boolean);
    function GetSelectedText: WideString;
    procedure SetSelectedText(Value: WideString);
    function GetCanUndo: boolean;
    function GetCanRedo: boolean;
    function GetTextLength: integer;
    function GetText: WideString;
    function GetTextIf: WideString;
    procedure SetText(const Value: WideString);
    procedure SetSearchText(const Value: WideString);
    procedure SetSearchigCase(Value: boolean);
    procedure SetDocumentParent(const Value: TNotePad);
    procedure SetCodePage(const Value: TCodePage);
    function GetMarks(index: integer): integer;
    function GetMarkCount: integer;
    function GetBoxSelected: boolean;
    function GetSelectedWithoutCRLF: boolean;
    function GetCaretline: integer;
    procedure SetCaretLine(const Value: integer);
    function GetCompletePoint: TPoint;
  protected
    //ŃI[o[ChĂ݂āAprotectedKvȂ̂
    FOptions: TNotePadOptions;
    FStr: TNotePadString;
    procedure DoViewChanged;
    //Ǝǉ̂߂̉z
    //ɒ`g[NԍZbgׂꏊ
    procedure SearchHighlight(Value: integer); virtual;
  protected
    procedure CreateHandle; override;
    procedure WndProc(var Message: TMessage); override;
    procedure Paint; override;


    //TControl̓\bhƏՓ˂̂Ń\bhς
    function IDropTarget.DragOver = DTDragOver;
    function DTDragOver(grfKeyState: Longint; pt: TPoint;
      var dwEffect: Longint): HResult; stdcall;
    function IDropTarget.DragEnter = DTDragEnter;
    function DTDragEnter(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; stdcall;
    function IDropTarget.DragLeave = DTDragLeave;
    function DTDragLeave: HResult; stdcall;
    function IDropTarget.Drop = DTDrop;
    function DTDrop(const dataObj: IDataObject; grfKeyState: Longint;
      pt: TPoint; var dwEffect: Longint): HResult; stdcall;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DefaultHandler(var Message); override;
    procedure ClearAll;
    procedure Clear;
    procedure ClearUndo;
    procedure Undo;
    procedure Redo;
    procedure Cut;
    procedure Copy;
    procedure Paste;
    function CanPaste: boolean;
    procedure SelectAll;
    procedure ScrollToCaret;
    procedure LoadFromString(const Value: WideString);
    procedure Select(Start, Len: integer);
    function GetTextCR: WideString;
    function GetTextLF: WideString;
    function GetTextCRLF: WideString;
    function GetReturnedString: WideString;
    function ToPointer(index: integer): PWChar;
    procedure ImeRegWord;
    procedure SetModeFromPath(pathName: WideString);
    procedure SelectWord(X, Y: integer);
    procedure SelectWordFromCaret;
    function GetWordFromCaret: WideString;
    function GetModeExt(Value: THighlightMode): WideString;
    procedure SetModeExt(mode: THighlightMode; const Value: WideString);
    procedure AddMark; overload;
    procedure AddMark(Value: integer); overload;
    procedure DeleteMark(Value: integer);
    procedure Reconvert;
    procedure InputKey(Message: TWMKey);//Alt+ړL[̂߂̎
    procedure InputKeyChar(Message: TWMKey);//͕⊮̂߂̎
    procedure SubSelect(AStart, AEnd: integer);//^I𔽓]
    procedure CompleteFocusStart;
    procedure CompleteFocusCancel;
    procedure ReplaseWithoutSelect(const Value: WideString; AStart, ALength: integer);
    //Ǝǉ̂߂̉z
    //ɒ`g[Nԍϊׂꏊ
    function ColorFromToken(const C: WideChar): TColor; virtual;
    function StyleFromToken(const C: WideChar): TFontStyles; virtual;

    //vpeB
    property BoxSelectMode: boolean read GetBoxSelectMode write SetBoxSelectMode;
    property SelectionLength: integer read GetSelectionLength write SetSelectionLength;
    property SelectionStart: integer read GetSelectionStart write SetSelectionStart;
    property CurrentMode: TNotePadMode read FCurrentMode;
    property EditMode: THighlightMode read GetEditMode write SetEditMode;
    property Options: TNotePadOptions read GetOptions write SetOptions;
    property WordWrapCount: integer read FWordWrapCount write SetWordWrapCount;
    property WordWrap: TWordWrap read FWordWrap write SetWordWrap;
    property Modified: boolean read FModified write SetModified;
    property SelectedText: WideString read GetSelectedText write SetSelectedText;
    property CanUndo: boolean read GetCanUndo;
    property CanRedo: boolean read GetCanRedo;
    property TextLength: integer read GetTextLength;
    property Text: WideString read GetText write SetText;
    property TextIf: WideString read GetTextIf;
    property SearchText: WideString read FSearchText write SetSearchText;
    property SearchIgCase: boolean read FSearchIgCase write SetSearchigCase;
    property DocumentParent: TNotePad read FDocumentParent write SetDocumentParent;
    property NotePadString: TNotePadString read FStr;
    property CodePage: TCodePage read FCodePage write SetCodePage;
    property Marks[index: integer]: integer read GetMarks;
    property MarkCount: integer read GetMarkCount;
    property BoxSelected: boolean read GetBoxSelected;
    property SelectedWithoutCRLF: boolean read GetSelectedWithoutCRLF;
    property CaretLine: integer read GetCaretline write SetCaretLine;
    property CompletePoint: TPoint read GetCompletePoint;
  published
    property Align;
    property Anchors;
    property Enabled;
    property ImeMode;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop default True;
    property Visible;
    //Cxg
    property ViewChanged: TNotifyEvent read FViewChanged write FViewChanged;
    property ModifiedChanged: TNotifyEvent read FModifiedChanged write FModifiedChanged;
    property FilesDroped: TDropFileEvent read FFilesDroped write FFilesDroped;
    property RulerDown: TRulerEvent read FRulerDown write FRulerDown;
    property RulerMove: TRulerEvent read FRulerMove write FRulerMove;
    property RulerUp: TRulerEvent read FRulerUp write FRulerUp;
  end;

procedure Register;

implementation

uses sysutils, Math, Imm, Forms, Clipbrd, ClipbrdW, ShellAPI, ShlObj,
  ModeDefault, ModeHtml, ModeCSharp, ModeDelphi, ModeJScript, ModeVBScript,
  ModePython, ModeCpp, ModeJava, ModePerl, ModeRuby, ModeVisualBasic, ModeCss,
  ModeBat,
  LineParser, NotePadConsts, NotePadClasses, ParserConsts, NotePadSearch,
  FontUtils, MultiByteMap, StringBuffer;

procedure Register;
begin
  RegisterComponents('Samples', [TNotePad]);
end;

var
  FInitialized: boolean;

  FModeDefault: TModeDefault;
  FModeHtml: TModeHtml;
  FModeCSharp: TModeCSharp;
  FModeDelphi: TModeDelphi;
  FModeJScript: TModeJScript;
  FModeVBScript: TModeVBScript;
  FModePython: TModePython;
  FModeCpp: TModeCpp;
  FModeJava: TModeJava;
  FModePerl: TModePerl;
  FModeRuby: TModeRuby;
  FModeVisualBasic: TModeVisualBasic;
  FModeCss: TModeCss;
  FModeBat: TModeBat;

  FBGWindow: TBitmap;
  FTileBitmap: TBitmap;
  FNeedInitBMP: boolean;
  FNeedInitImeFont: boolean;

  FRgn: HRGN;

procedure UnitInitialize;
begin
  //AvNĈԍŏɂ邱Ƃ͓dÑ`FbNȂ̂A
  //InitializationŎdƂ̑Oɏ邱ƂɂȂĂ܂B
  if FInitialized then exit;
  FInitialized := true;
  FontUtils.UnitInitialize;
  NotePadConsts.UnitInitialize;
  NotePadSearch.UnitInitialize;
  ClipbrdW.UnitInitialize;
  
  FModeDefault := TModeDefault.Create;
  FModeHtml := TModeHtml.Create;
  FModeCSharp := TModeCSharp.Create;
  FModeDelphi := TModeDelphi.Create;
  FModeJScript := TModeJScript.Create;
  FModeVBScript := TModeVBScript.Create;
  FModePython := TModePython.Create;
  FModeCpp := TModeCpp.Create;
  FModeJava := TModeJava.Create;
  FModePerl := TModePerl.Create;
  FModeRuby := TModeRuby.Create;
  FModeVisualBasic := TModeVisualBasic.Create;
  FModeCss := TModeCss.Create;
  FModeBat := TModeBat.Create;
  FModeHtml.FileExts := '.htm;.html;.php;.xml';
  FModeCSharp.FileExts := '.cs';
  FModeDelphi.FileExts := '.pas;.dpr;.dpk';
  FModeJScript.FileExts := '.js';
  FModeVBScript.FileExts := '.vbs';
  FModePython.FileExts := '.py;.pys';
  FModeCpp.FileExts := '.c;.cpp;.h;.hpp';
  FModeJava.FileExts := '.java;.class;.jsl';
  FModePerl.FileExts := '.cgi;.pl;.pls;.pm;.pod';
  FModeRuby.FileExts := '.rb';
  FModeVisualBasic.FileExts := '.vb';
  FModeCss.FileExts := '.css';
  FModeBat.FileExts := '.bat';
  //16rbgJ[ɉĂ񂶂Ȃ낤
  //8rbgJ[̓pbgǗKvȂ̂Ń_B
  //ԂA]Ɠ]ŃJ[ȂL
  FBGWindow := TBitmap.Create;
  FBGWindow.PixelFormat := pf16bit;
  FTileBitmap := TBitmap.Create;
  FTileBitmap.PixelFormat := pf16bit;
  FNeedInitBMP := true;

end;

function CreateOptions: TNotePadOptions;
begin
  result.ColorFore := clBlack;
  result.ColorBack := clWhite;
  result.ColorLeftBar := clGreen;
  result.ColorRuler := clGreen;
  result.ColorCaretLine := clBlue;//ꂾclNoneŔ\Ȃ̂̓}Y
  result.ColorSBSpace := clSilver;
  result.ColorDBSpace := clBlue;
  result.ColorTab := clBlue;
  result.ColorEOF := clSilver;
  result.ColorLineEnd := clTeal;
  result.ColorWrap := clFuchsia;
  result.ColorWrapLine := clGreen;
  Result.ColorWideChar  := clNone;
  Result.ColorMark      := clYellow;
  result.ShowRuler := true;//eLXg̈Ô̂͌Ăǂ
  result.ShowLeftBar := true;//eLXg̈Ô̂͌Ăǂ
  result.ShowSBSpace := false;//Ȃ̂Kv͂Ȃ
  result.ShowDBSpace := false;//Ȃ̂Kv͂Ȃ
  result.ShowTab := false;//Ȃ̂Kv͂Ȃ
  result.ShowEOF := false;//EOFȂďɏĂӖs
  result.ShowLineEnd := true;//s͕\邱ƂOɔp1Ă邩
  result.ShowWrap := false;//ɉƎvĂ邩
  result.ShowWrapLine := true;//eLXg̈Ô̂͌Ăǂ
  result.TabCount := 8;
  result.LeftBarCountWrap := true;
  result.LineMargin := 2;
  result.Font := FDefaultFont;
  result.AutoIndent := 1;
  result.BackGroundBMP := nil;
  result.AlphaBlend := 70;
  result.DrawNot := false;
  result.KeepCol := false;
  result.RulerEnabled := true;

  Result.ColorControl   := clNone;
  Result.ColorHighlight := clNone;
  Result.ColorSymbol    := clNone;
  Result.ColorString    := clNone;
  Result.ColorInteger   := clNone;
  Result.ColorComment   := clNone;
  Result.ColorReserve   := clNone;
  Result.ColorReserve2  := clNone;
  Result.ColorReserve3  := clNone;
  Result.ColorAnk       := clNone;
  Result.ColorAlphabetic:= clNone;
  Result.ColorDBSymbol  := clNone;
  Result.ColorDBInt     := clNone;
  Result.ColorDBAlph    := clNone;
  Result.ColorDBHira    := clNone;
  Result.ColorDBKana    := clNone;
  Result.ColorDBKanji   := clNone;
  Result.ColorSBKana    := clNone;
  Result.ColorUnicode   := clNone;
  Result.ColorLink      := clNone;
  Result.ColorHtmlTag   := clNone;
  Result.ColorPrefix    := clNone;
  Result.ColorHere      := clNone;
  Result.StyleControl   := [];
  Result.StyleHighlight := [];
  Result.StyleSymbol    := [];
  Result.StyleString    := [];
  Result.StyleInteger   := [];
  Result.StyleComment   := [];
  Result.StyleReserve   := [];
  Result.StyleReserve2  := [];
  Result.StyleReserve3  := [];
  Result.StyleAnk       := [];
  Result.StyleAlphabetic:= [];
  Result.StyleDBSymbol  := [];
  Result.StyleDBInt     := [];
  Result.StyleDBAlph    := [];
  Result.StyleDBHira    := [];
  Result.StyleDBKana    := [];
  Result.StyleDBKanji   := [];
  Result.StyleSBKana    := [];
  Result.StyleUnicode   := [];
  Result.StyleLink      := [];
  Result.StyleHtmlTag   := [];
  Result.StylePrefix    := [];
  Result.StyleHere      := [];

end;

function ObjectFromHighlight(Value: THighlightMode): TNotePadMode;
begin
  case Value of
    hmHtml:    Result := FModeHtml   ;
    hmCSharp:  Result := FModeCSharp ;
    hmDelphi:  Result := FModeDelphi ;
    hmJScript: Result := FModeJScript;
    hmVBScript: Result := FModeVBScript;
    hmPython: Result := FModePython;
    hmCpp: Result := FModeCpp;
    hmJava: Result := FModeJava;
    hmPerl: Result := FModePerl;
    hmRuby: Result := FModeRuby;
    hmVisualBasic: Result := FModeVisualBasic;
    hmCss: Result := FModeCss;
    hmBat: Result := FModeBat;
    else       Result := FModeDefault;
  end;
end;


{ TNotePad }

procedure TNotePad.AddMark(Value: integer);
begin
  FStr.AddMark(Value);
  Invalidate;
end;

procedure TNotePad.AddMark;
begin
  AddMark(FStr.SelEnd);
end;

procedure TNotePad.CalcFont;
var
  h : HDC;
  TM: TEXTMETRICW;
  p: TSize;
begin
//[[A܂Ԃ߂A^u̕߂A
//֘Ȁ邽߂ɁAtHg̕ϕ𒲂ׂKvB
//tHgł̔pꕶ̕擾̂A
//v|[VitHgƂꂪ₽Lݒ肳Ă肷̂ŁA
//̃[̕AŎg{̃tHgɂB
  h := GetDC(Handle);
  try
    SelectObject(h, FFont.Handle);
    GetTextMetricsW(h, TM);
    GetTextExtentPoint32W(h, '0', 1, p);
    FFontWidth := p.cx;//Max(TM.tmAveCharWidth, p.cx);
    FLineHeight := TM.tmHeight + TM.tmExternalLeading + 1 + FOptions.LineMargin;
    FLeading := TM.tmInternalLeading div 2;//ANZgLƂ̍(AscentɊ܂܂)
                                           //ǁA0ɒ[ɑ傫Ȃ̂ŔŁB
    FAscent := TM.tmAscent;//A_[C̍
    FDescent := TM.tmDescent;//A_[C牺̍
  finally
    ReleaseDC(Handle, h);
  end;
  InitFontWidthMap(FFont, FFontWidth, (tm.tmPitchAndFamily and 1) <> 1);
  FFontChanged := true;
  FNeedInitImeFont := true;
end;

procedure TNotePad.CalcLeftBar;
var
  i: integer;
begin
  i := (Length(inttostr(FStr.Count)) + 1) * FFontWidth * FShowLeftBar;
  if i <> FLeftBarWidth then begin
    FLeftBarWidth := i;
    Invalidate;
  end;
end;

function TNotePad.CanPaste: boolean;
begin
  result := Clipboard.HasFormat(CF_UNICODETEXT);
end;

procedure TNotePad.CaretHide;
begin
  if HandleAllocated {and Focused} and (not FCaretHided) then
  begin
    FCaretHided := true;
    HideCaret(Handle);
    DestroyCaret;
  end;
end;

procedure TNotePad.CaretReset;
begin
  //z͕2炢ǁAv|[VitHgɔׂđ肷
  //CreateCaret(Handle, 0, 0{2}, {Abs(FFont.Height) + 1}FAscent + FDescent);
  //initCaretPosition;
end;

procedure TNotePad.CaretShow;
begin
  if HandleAllocated and FCaretHided then
  begin
    FCaretHided := false;
    if FDraging then
      CreateCaret(Handle, 1, 1{2}, FAscent + FDescent)
    else
      CreateCaret(Handle, 0, 2{2}, FAscent + FDescent);
    ShowCaret(Handle);
  end;
end;

procedure TNotePad.Clear;
begin
  FColKeeping := -1;
  FStr.Text := '';
end;

procedure TNotePad.ClearAll;
begin
  FColKeeping := -1;
  FStr.OnModified := nil;//-= new EventHandler(StringOnModified);
  FStr.ClearAll;
  FModified := false;
  FStr.OnModified := StringOnModified;
end;

procedure TNotePad.ClearUndo;
begin
  FStr.ClearUndo;
end;

function TNotePad.ColorFromToken(const C: WideChar): TColor; 
begin
  case C of
    toControl:
      Result := FOptions.ColorControl;
    toHighlight:
      Result := FOptions.ColorHighlight;
    toSBSymbol,
    toHereStart,
    toSymbolNoDiv:
      Result := FOptions.ColorSymbol;
    toQString:
      Result := FOptions.ColorString;
    toSBInt,
    toHex:
      Result := FOptions.ColorInteger;
    toReserve:
      Result := FOptions.ColorReserve;
    toReserve2:
      Result := FOptions.ColorReserve2;
    toReserve3:
      Result := FOptions.ColorReserve3;
    toAnk,
    toAnkNoReserve:
      Result := FOptions.ColorAnk;
    toAlphabetic,
    toHangul:
      Result := FOptions.ColorAlphabetic;
    toDBSymbol:
      Result := FOptions.ColorDBSymbol;
    toDBInt:
      Result := FOptions.ColorDBInt;
    toDBAlph:
      Result := FOptions.ColorDBAlph;
    toDBHira:
      Result := FOptions.ColorDBHira;
    toDBKana:
      Result := FOptions.ColorDBKana;
    toDBKanji:
      Result := FOptions.ColorDBKanji;
    toSBKana,
    toSBKanaSymbol:
      Result := FOptions.ColorSBKana;
    toUnicode:
      Result := FOptions.ColorUnicode;
    toUrl,
    toUrlStart,
    toFtpStart,
    toHttpsStart,
    toRtspStart,
    toPnmStart,
    toMmsStart,
    toNewsStart,
    toNntpStart:
      Result := FOptions.ColorLink;
    toComment,
    toVBRem,
    toCommentBrock0,
    toCommentBrock1,
    toCommentBrock1Start,
    toCommentBrock2,
    toCommentBrock2Start,
    toCommentBrock3,
    toCommentBrock3Start,
    toPodComment,
    toPodCommentEnd:
      Result := FOptions.ColorComment;
    toSingleQuotation,
    toDoubleQuotation,
    toAtDoubleQuot,
    toAtDoubleQuotStart,
    toDelphiChar,
    toDelphiCharHex,
    toRegExp,
    toMatchStart,
    toSubStart,
    toSubMatch:
      Result := FOptions.ColorString;
    toTagStart,
    toTagEnd,
    toTagQuestStart,
    toTagQuestEnd,
    toInsideTag:
      Result := FOptions.ColorHtmlTag;
    toDirective:
      Result := FOptions.ColorPrefix;
    toHereStartWord,
    toHereEnd:
      Result := FOptions.ColorHere;
    else Result := clNone;

  end;
end;

procedure TNotePad.CompleteFocusCancel;
begin
  if not FCompleting then exit;
  FCompleting := false;
  Invalidate;
end;

procedure TNotePad.CompleteFocusStart;
begin
  if FCompleting then exit;
  FCompleting := true;
  Invalidate;
end;

procedure TNotePad.Copy;
var
  s: WideString;
begin
  s := FStr.SelTextIf;
  if Length(s) > 0 then begin
    PutToClipBrd(s, BoxSelected);
    FStr.BeginUpdate;
    FStr.EndUpdate;
  end;
end;

constructor TNotePad.Create(AOwner: TComponent);
begin
  UnitInitialize;
  inherited;
  FNotePadString := TNotePadString.Create;
  FStr := FNotePadString;
  FFirstVisiblePix := 0;
  FFirstVisibleLine := 0;
  FLeftBarWidth := 0;
  FWordWrap := wwNone;
  FWordWrapCount := 70;
  FShowLeftBar := 1;
  FShowRuler := 1;
  FModified := false;
  FColKeeping := -1;
  FCaretHided := true;
  FDoubleClicked := false;
  FFocusClick := false;
  FDragCancel :=false;
  FDDString := nil;
  FDDReady := false;
  FDragCaret := Point(0, 0);
  FDraging := false;
  FCompleting := false;
  FMouseWheelDelta := 0;
  FCurrentMode := FModeDefault;

  FOptions := CreateOptions;
  FFont := TFont.Create;
  FFont.Assign(FDefaultFont);
  FOptions.Font := FFont;

  FAlphaBlend := FOptions.AlphaBlend;
  FNeedInitBMP := true;
  FSubSelectStart := -1;
  FSubSelectEnd := -1;
  
  FCodePage := cpNone;
  FInitOnHandle := true;
end;

function TNotePad.CreateBitmap: boolean;
begin
  Result := false;
  if (FBGWindow.Width < ClientWidth) or (FBGWindow.Height < ClientHeight) then begin
  end else exit;
  FBGWindow.Height := ClientHeight;
  FBGWindow.Width := ClientWidth;
  FNeedInitBMP := true;
  Result := true;
  Invalidate;
end;

procedure TNotePad.CreateHandle;
begin
  inherited;
  if not FInitOnHandle then exit;
  FInitOnHandle := false;
  FStr.OnChanged := StringOnChanged;
  FStr.OnNeedParse := StringOnNeedParse;
  CalcFont;
  ShowScrollBar(Handle, SB_BOTH, true);
  ClearAll;
end;

procedure TNotePad.Cut;
var
  s: WideString;
begin
  s := FStr.SelTextIf;
  if (Length(s) > 0) and PutToClipBrd(s, BoxSelected) then begin
    FColKeeping := -1;
    FStr.SelText := '';
  end;
end;

procedure TNotePad.DefaultHandler(var Message);
begin
  case TMessage(Message).Msg of
    WM_LBUTTONDOWN, WM_MBUTTONDOWN, WM_RBUTTONDOWN: begin
      if not (csDesigning in ComponentState) and not Focused then
        Windows.SetFocus(Handle);
      inherited;
    end;
    WM_NCLBUTTONDOWN, WM_NCMBUTTONDOWN, WM_NCRBUTTONDOWN: begin
      if not (csDesigning in ComponentState) and not Focused and
         (TWMNCHitMessage(Message).HitTest = HTCLIENT) then
        Windows.SetFocus(Handle);
      inherited;
    end;
    else inherited;
  end;
end;

procedure TNotePad.DeleteMark(Value: integer);
begin
  FStr.DeleteMark(Value);
  Invalidate;
end;

destructor TNotePad.Destroy;
begin
  if Assigned(FDocumentParent) then DocumentParent := nil;
  if Assigned(FDocumentChild) then FDocumentChild.DocumentParent := nil;
  FFont.Free;
  FNotePadString.Free;
  inherited;
end;

function TNotePad.DoubleClickDown(X: integer): boolean;
var
  p: TPoint;
  APrevious, AStart, ALength: integer;
  AJump: boolean;
  s: WideString;
begin
  p := FStr.SelEndPos;
  //IEł́AP̋E_uNbNƂAȍ~̒PꂪIB
  //ł͂܂ł̒PꂪIB
  //SelectWordSub͍ĕϊɂg邽߁Aȍ~̒PIdlB
  //āA_uNbNdl݊ɂ邽߁AňZKvBij
  //ǂ烁́A_uNbNǂŔ͖āA
  //Lbgꏊ܂ł̒PI悤B
  //TNotePad̎dlŎgĂA͂_uNbN܂ޒPꂪIB
  SelectWordSub(Max(0, X), p.Y, APrevious, AStart, ALength, AJump);
  //SelectWordSub(Max(0, p.X - 1), p.Y, AStart, ALength, AJump);
  if AJump then begin
    s := FStr.InternalSB.ToString(AStart, ALength);
    ShellExecuteW(0, 'open', PWChar(s), nil, nil, SW_SHOW);
    result := true;
  end else begin
    FStr.BeginUpdate;
    FStr.SelStart := AStart;
    FStr.SelEnd := FStr.SelStart + ALength;
    FStr.RedrawCurrent;
    FStr.EndUpdate;
    result := false;
  end;
end;

procedure TNotePad.DoViewChanged;
begin
  //j[Ac[o[AXe[^Xo[XV邽߂̃Cxg
  if Assigned(FViewChanged) then FViewChanged(self);
  if Assigned(FDocumentParent) and Assigned(FDocumentParent.FViewChanged)
    then FDocumentParent.FViewChanged(self);
  if Assigned(FDocumentChild) and Assigned(FDocumentChild.FViewChanged)
    then FDocumentChild.FViewChanged(self);
end;

procedure TNotePad.DrawCanvas(ACanvas: TCanvas; ARgnData: PRgnData);
var
  r:TRect;
  i, j: integer;
  parser: TNotePadParser;
  ALineNum: integer;
  s: WideString;
  ACaretLine: boolean;
  y1: integer;
  //ALog: integer;
begin
  ACanvas.Brush.Color := FOptions.ColorBack;
  ACanvas.Brush.Style := bsSolid;
  if Assigned(FOptions.BackGroundBMP) then begin
    for i := 0 to ARgnData.rdh.nCount - 1 do begin
      r := PRect(integer(@ARgnData.Buffer) + i * sizeof(TRect))^;
      BitBlt(ACanvas.Handle, r.Left, r.Top,
             r.Right - r.Left, r.Bottom - r.Top,
             FTileBitmap.Canvas.Handle, r.Left, r.Top, SRCCOPY);
    end;
  end else begin
    FillRgn(ACanvas.Handle, FRgn, ACanvas.Brush.Handle);
  end;
  ACanvas.Font := FFont;
  //[[`
  r := Rect(0, 0, ClientWidth, FLineHeight);
  if (FShowRuler > 0) and RectInRegion(FRgn, r) then begin
    DrawRuler(ACanvas, r);
  end;
  //stgo[`
  parser := FCurrentMode.NewParser(FStr);
  //ɑSsp[Xς݂ŁA`ɕKvȂƂ낾J[p[X΂
  ALineNum := 1;
  for i := 0 to FFirstVisibleLine - 1 do
    if FStr[i].LineEnd <> leWrap then inc(ALineNum);
  //ALog := IntLength(FStr.Count);
  for i := 1 to FVisibleLineCount do begin
    ACaretLine := (FStr.SelLength = 0) and (FOptions.ColorCaretLine <> clNone) and
      (FStr.SelEndPos.Y = FFirstVisibleLine + i - 1);
    y1 := FLineHeight * (i - 1 + FShowRuler);
    r := Rect(FLeftBarWidth + 6, y1, ClientWidth, y1 + FLineHeight);
    if RectInRegion(FRgn, r) and (FFirstVisibleLine + i <= FStr.Count) then begin
      parser.ParseLineColor(FFirstVisibleLine + i - 1);
      SearchHighlight(FFirstVisibleLine + i - 1);
      DrawLine(i, ACaretLine, parser, ACanvas); 
    end;
    r := Rect(0, y1, FLeftBarWidth + 3, y1 + FLineHeight);
    if RectInRegion(FRgn, r) or FOptions.LeftBarCountWrap then begin
      if Assigned(FOptions.BackGroundBMP) then begin
        BitBlt(ACanvas.Handle, r.Left, r.Top,
             r.Right - r.Left, r.Bottom - r.Top,
             FTileBitmap.Canvas.Handle, r.Left, r.Top, SRCCOPY);
      end else begin
        ACanvas.Brush.Color := FOptions.ColorBack;
        ACanvas.Brush.Style := bsSolid;
        ACanvas.FillRect(r);
      end;
      if (FFirstVisibleLine + i <= FStr.Count) then begin
        for j := 0 to FStr.MarkCount - 1 do begin
          if FStr.Marks[j] = FStr[FFirstVisibleLine + i - 1].Start then begin
            ACanvas.Brush.Color := FOptions.ColorMark;
            ACanvas.Brush.Style := bsSolid;
            ACanvas.FillRect(r);
            break;
          end;
        end;
      end;
      if FShowLeftBar > 0 then begin
        ACanvas.Brush.Style := bsClear;
        ACanvas.Pen.Color := FOptions.ColorLeftBar;
        ACanvas.Pen.Width := 1;
        ACanvas.Pen.Style := psSolid;
        ACanvas.MoveTo(r.Right, r.Top);
        ACanvas.LineTo(r.Right, r.Bottom);
      end;
      if (FFirstVisibleLine + i <= FStr.Count) and (FShowLeftBar > 0) then begin
        if FOptions.LeftBarCountWrap then begin
          s := IntToStr(FFirstVisibleLine + i);
          r.Left := FFontWidth - (FFontWidth div 3);
          inc(r.Top, FOptions.LineMargin div 2);
          if ACaretLine then begin
            ACanvas.Font.Color := FOptions.ColorCaretLine;
          end else begin
            ACanvas.Font.Color := FOptions.ColorLeftBar;
          end;
          ACanvas.Font.Style := [];
          //TextOutW(ACanvas.Handle, r.Left + 2 + (ALog - Length(s)) * FFontWidth, r.Top, PWChar(s), Length(s));
          r.Right := r.Right - 1;
          DrawTextW(ACanvas.Handle, PWChar(s), Length(s), r, DT_RIGHT or DT_SINGLELINE or DT_NOPREFIX);
        end else if (FFirstVisibleLine + i < 2) or (FStr[FFirstVisibleLine + i - 2].LineEnd <> leWrap) then begin
          s := IntToStr(ALineNum);
          r.Left := FFontWidth - (FFontWidth div 3);
          inc(r.Top, FOptions.LineMargin div 2);
          if ACaretLine then begin
            ACanvas.Font.Color := FOptions.ColorCaretLine;
          end else begin
            ACanvas.Font.Color := FOptions.ColorLeftBar;
          end;
          ACanvas.Font.Style := [];
          //TextOutW(ACanvas.Handle, r.Left + 2 + (ALog - Length(s)) * FFontWidth, r.Top, PWChar(s), Length(s));
          r.Right := r.Right - 1;
          DrawTextW(ACanvas.Handle, PWChar(s), Length(s), r, DT_RIGHT or DT_SINGLELINE or DT_NOPREFIX);
        end;
      end;
    end;
    if (FFirstVisibleLine + i <= FStr.Count) and (FStr[FFirstVisibleLine + i - 1].LineEnd <> leWrap) then inc(ALineNum);
  end;//for
  parser.Free;
  if FOptions.DrawNot then begin
    for i := 0 to ARgnData.rdh.nCount - 1 do begin
        r := PRect(integer(@ARgnData.Buffer) + i * sizeof(TRect))^;
        BitBlt(ACanvas.Handle, r.Left, r.Top,
               r.Right - r.Left, r.Bottom - r.Top,
               ACanvas.Handle, r.Left, r.Top, NOTSRCCOPY);
    end;
  end;
end;

procedure TNotePad.DrawLine(ALineNum: integer; ACaretLine: boolean;
  parser: TNotePadParser; ACanvas: TCanvas);
var
  ALine: PLineData;
  sum: integer;
  ATabPix: integer;
  ps: array of TPoint;
  ASelArea: TPoint;
  r: TRect;
  HalfHeight: integer;
  HalfWidth: integer;
  BaseX, BaseY: integer;
  i, j: integer;
  C: PWideChar;
  AColor: TColor;
  AStyle: TFontStyles;
  Skip: integer;
  suminc:integer;
  AWrapLine: TPoint;
  AIntLeading: integer;
  ws: WideString;
//var
//  usedefchar:bool;
  function IsSingleDraw(const AWChar: WideChar): boolean;
  begin
    case AWChar of
      #$0009, #$0020, #$3000: Result := true;
      else case TokenFromFirstChar(AWChar) of
        #0, toDBSymbol, toAlphabetic, toUnicode: Result := true;
        else Result := false;
      end;
    end;
  end;
begin
  //܂ԂC
  //CtHg㏑ƌhǂȂ̂ŁAɂĒuB
  //wihԂŏ邩mȂ̂ŁAAWrapLineɓĂ̓sxB
  AWrapLine.X := FLeftBarWidth + 6 + FFontWidth * FStr.WrapByte - FFirstVisiblePix - 1;
  AWrapLine.Y := (ALineNum - 1 + FShowRuler) * FLineHeight;
  if FOptions.ShowWrapLine then begin
    ACanvas.Pen.Color := FOptions.ColorWrapLine;
    ACanvas.MoveTo(AWrapLine.X, AWrapLine.Y);
    ACanvas.LineTo(AWrapLine.X, AWrapLine.Y + FLineHeight);
  end;
  ALine := FStr[FFirstVisibleLine + ALineNum - 1];
  ASelArea := FStr.GetSelectedArea(FFirstVisibleLine + ALineNum - 1);
  sum := - FFirstVisiblePix;
  ATabPix := FStr.TabCount * FFontWidth;
  HalfHeight := ({Abs(FFont.Height) - 2}FAscent - FLeading) div 2;
  HalfWidth := (FFontWidth - 2) div 2;
  BaseY := (ALineNum - 1 + FShowRuler) * FLineHeight + (FOptions.LineMargin div 2);
  AIntLeading := FAscent - HalfHeight * 2;
  C := FStr.InternalSB.ToPointer(ALine.Start);
  Skip := 0;
  for i := 1 to ALine.Length do begin
    if Skip > 0 then dec(Skip) else begin
      BaseX := FLeftBarWidth + 6 - 1 + (sum{ * FFontWidth});
      case C^ of
        #0009: begin
          if (sum >= -(FFontWidth * 2)) and (FOptions.ShowTab) then begin
            ACanvas.Pen.Color := FOptions.ColorTab;
            SetLength(ps, 4);
            ps[0] := Point(BaseX + 1, BaseY + AIntLeading);
            ps[1] := Point(ps[0].X, ps[0].Y + HalfHeight * 2);
            ps[2] := Point(ps[0].X + FFontWidth - 2, ps[0].Y + HalfHeight);
            ps[3] := ps[0];
            ACanvas.Polyline(ps);
          end;
          suminc := ATabPix - ((sum + FFirstVisiblePix) mod ATabPix);
          if suminc < FFontWidth then inc(suminc, ATabPix);
          inc(sum, suminc);
        end;
        #$0020: begin
          if (sum >= -(FFontWidth * 2)) and FOptions.ShowSBSpace then begin
            ACanvas.Pen.Color := FOptions.ColorSBSpace;
            SetLength(ps, 4);
            ps[0] := Point(BaseX + 1, BaseY + AIntLeading + HalfHeight * 3 div 2);
            ps[1] := Point(ps[0].X, BaseY + AIntLeading + HalfHeight * 2);
            ps[2] := Point(ps[1].X + GetGlyphWidth(WideChar($0020)) - 2 , ps[1].Y);
            ps[3] := Point(ps[2].X, ps[0].Y - 1);
            ACanvas.Polyline(ps);
          end;
          inc(sum, GetGlyphWidth(WideChar($0020)));
        end;
        #$3000: begin
          if (sum >= -(FFontWidth * 2)) and FOptions.ShowDBSpace then begin
            ACanvas.Pen.Color := FOptions.ColorDBSpace;
            SetLength(ps, 5);
            ps[0] := Point(BaseX + 1, BaseY + AIntLeading);
            ps[1] := Point(ps[0].X, ps[0].Y + HalfHeight * 2);
            ps[2] := Point(ps[1].X + GetGlyphWidth(WideChar($3000)) - 2, ps[1].Y);
            ps[3] := Point(ps[2].X, ps[0].Y);
            ps[4] := ps[0];
            ACanvas.Polyline(ps);
          end;
          inc(sum, GetGlyphWidth(WideChar($3000)));
        end;
        #$0000..#$0008, #$000A..#$001F,
        #$0080..#$009F: begin
          suminc := GetGlyphWidth(C^);
          if sum >= -suminc then begin
            AStyle := StyleFromToken(LineTokens[i - 1]);
            ACanvas.Font.Style := AStyle - [fsUnderline];
            AColor := ColorFromToken(LineTokens[i - 1]);
            if AColor = clNone then ACanvas.Font.Color := FOptions.ColorFore else
              ACanvas.Font.Color := ColorFromToken(LineTokens[i - 1]);
            ACanvas.Brush.Style := bsClear;
            ws := CC2String(C^);
            for j := 1 to Length(ws) do WidthArray[j-1] := GetGlyphWidth(ws[j]);
            ExtTextOutW(ACanvas.Handle, BaseX, BaseY, {ETO_GLYPH_INDEX}0, nil, PWChar(ws), Length(ws), Pointer(WidthArray));
            if fsUnderline in AStyle then begin
              ACanvas.Pen.Color := ACanvas.Font.Color;
              ACanvas.MoveTo(BaseX, BaseY + FAscent{Abs(FFont.Height) - 1});
              ACanvas.LineTo(BaseX + suminc, BaseY + FAscent{Abs(FFont.Height) - 1});
            end;
          end;
          inc(sum, suminc);
        end;
        else begin
          suminc := GetGlyphWidth(C^);
          if sum >= -(FFontWidth * 2) then begin
            AStyle := StyleFromToken(LineTokens[i - 1]);
//݂̎dlł͕`悪{glbNɂȂꍇA
//@ɕ`炷̃JMɂȂB
//ɁA̕`ɂRXǵȂ̏ɂRXgꡂɑ傫B
//TextOut̃R[񌸂炷߂ɁA100s̃R[hĂ͂ȂB
//Ƃ킯ŁA܂Ƃ߂ĕ`ł镔Ă݂A
//܂Ƃ߂ĕ`悵ĂPÂ`悵ĂႢ킩ȂxBiBj
//ɁAFFontWidthƂ̖肪̂ňꗘȂBiBj
//v|[ViΉFFontWidthEp̂ŁA߂ĔrĂ݂ƑB
//R̓eɂč͂邾낤A͂܂Ƃ߂ĕ`ׂ͂B
//uāv̕ƁuĂv̕ႤꍇāATextOutł͔j]I
//ExtTextOutɂāAz𓱓邱ƂŉAXs[h͂ǂ낤H
//TextOut() is essentially a call to ExtTextOut() with wOptions, lpRect, and lpDx set to zero.
//ExtTextOutmȂA܂Ƃ߂ĕ`悵B
//ĂƂTextOuẗ`ĎɂȂ͂B
            //Cɕ`łꍇ̏
            if (not(fsBold in AStyle)) and (not(fsItalic in AStyle)) and
               (not IsSingleDraw(C^))  then begin
              while (i + Skip < ALine.Length) and
                    (sum + suminc <= FVisibleByteCount * FFontWidth) do begin
                if IsSingleDraw((PWChar(C + Skip + 1))^) then break;
                if (ColorFromToken(LineTokens[i - 1]) <> ColorFromToken(LineTokens[i - 1 + Skip + 1]))
                  or (StyleFromToken(LineTokens[i - 1]) <> StyleFromToken(LineTokens[i - 1 + Skip + 1])) then break;
                inc(Skip);
                inc(suminc, GetGlyphWidth((PWChar(C + Skip))^));
              end;
            end;
            //WideCharToMultiByteŃ`FbNĂƂĂxB
            //SJISe[upӂăXs[hAbvB
            //܂AEUC(CP51932)JIS(CP50222)usedefchartrueɂȂĂȂB
            if FOptions.ColorWideChar <> clNone then begin
              r.Left := BaseX;
              r.Top  := (ALineNum - 1 + FShowRuler) * FLineHeight;
              r.Bottom := r.Top + FLineHeight;
              ACanvas.Brush.Color := FOptions.ColorWideChar;
              ACanvas.Brush.Style := bsSolid;
              for j := 0 to skip do begin
                r.Right := r.Left + GetGlyphWidth((PWChar(C + j))^);
                if (MBMAP[Word((PWChar(C + j))^)] and byte(FCodePage)) <> 0 then begin
                  ACanvas.FillRect(r);
                  if (r.Right >= AWrapLine.X) and FOptions.ShowWrapLine then begin
                    ACanvas.Pen.Color := FOptions.ColorWrapLine;
                    ACanvas.MoveTo(AWrapLine.X, AWrapLine.Y);
                    ACanvas.LineTo(AWrapLine.X, AWrapLine.Y + FLineHeight);
                  end;
                end;
                r.Left := r.Right;
              end;
            end;
            ACanvas.Font.Style := AStyle - [fsUnderline];
            AColor := ColorFromToken(LineTokens[i - 1]);
            if AColor = clNone then ACanvas.Font.Color := FOptions.ColorFore else
              ACanvas.Font.Color := ColorFromToken(LineTokens[i - 1]);
            ACanvas.Brush.Style := bsClear;
            for j := 0 to Skip do WidthArray[j] := GetGlyphWidth((PWChar(C + j))^);
            ExtTextOutW(ACanvas.Handle, BaseX, BaseY, 0, nil, C, Skip + 1, Pointer(WidthArray));
            if fsUnderline in AStyle then begin
              ACanvas.Pen.Color := ACanvas.Font.Color;
              ACanvas.MoveTo(BaseX, BaseY + FAscent);
              ACanvas.LineTo(BaseX + suminc, BaseY + FAscent);
            end;
          end;
          inc(sum, suminc);
        end;
      end;//case
      if sum > FVisibleByteCount * FFontWidth then break;
    end;//skip
    inc(C);
  end; //for
  if sum >= -(FFontWidth * 2) then case ALine.LineEnd of
    leNone: begin
      if FOptions.ShowEOF then begin
        ACanvas.Font.Style := [];
        ACanvas.Font.Color := FOptions.ColorEOF;
        ACanvas.Brush.Style := bsClear;
        ACanvas.TextOut(FLeftBarWidth + 6 + sum, BaseY,'[EOF]');
      end;
    end;
    leLF: begin
      case FStr.LineEnds of
        lsCR:begin
          if FOptions.ShowLineEnd then begin
            ACanvas.Pen.Color := FOptions.ColorLineEnd;
            SetLength(ps, 5);
            ps[0] := Point(FLeftBarWidth + 6 + sum + HalfWidth * 2, BaseY + AIntLeading + 1 + HalfHeight * 2 - HalfWidth);
            ps[1] := Point(ps[0].X - HalfWidth * 2, ps[0].Y);
            ps[2] := Point(ps[1].X + HalfWidth, ps[1].Y - HalfWidth);
            ps[3] := Point(ps[2].X, ps[1].Y + HalfWidth);
            ps[4] := ps[1];
            ACanvas.Polyline(ps);
          end;
        end;
        lsLF:begin
          if FOptions.ShowLineEnd then begin
            ACanvas.Pen.Color := FOptions.ColorLineEnd;
            SetLength(ps, 5);
            ps[0] := Point(FLeftBarWidth + 6 + sum + HalfWidth, BaseY + AIntLeading + 1);
            ps[1] := Point(ps[0].X, ps[0].Y + HalfHeight * 2);
            ps[2] := Point(ps[1].X - HalfWidth, ps[1].Y - HalfWidth);
            ps[3] := Point(ps[1].X + HalfWidth, ps[2].Y);
            ps[4] := ps[1];
            ACanvas.Polyline(ps);
          end;
        end;
        lsCRLF:begin
          if FOptions.ShowLineEnd then begin
            ACanvas.Pen.Color := FOptions.ColorLineEnd;
            SetLength(ps, 6);
            ps[0] := Point(FLeftBarWidth + 6 + sum + HalfWidth * 2, BaseY + AIntLeading + 1);
            ps[1] := Point(ps[0].X, ps[0].Y + HalfHeight * 2 - HalfWidth);
            ps[2] := Point(ps[1].X - HalfWidth * 2, ps[1].Y);
            ps[3] := Point(ps[2].X + HalfWidth, ps[2].Y - HalfWidth);
            ps[4] := Point(ps[3].X, ps[2].Y + HalfWidth);
            ps[5] := ps[2];
            ACanvas.Polyline(ps);
          end;
        end;
      end;//case
    end;
    leWrap: begin
      if FOptions.ShowWrap then begin
        ACanvas.Pen.Color := FOptions.ColorWrap;
        SetLength(ps, 4);
        ps[0] := Point(FLeftBarWidth + 6 + sum + HalfWidth * 2, BaseY + AIntLeading + 1);
        ps[1] := Point(ps[0].X - HalfWidth * 2, ps[0].Y + HalfHeight);
        ps[2] := Point(ps[0].X, ps[1].Y + HalfHeight);
        ps[3] := ps[1];
        ACanvas.Polyline(ps);
      end;
    end;
  end;
  if (FSubSelectStart >= 0) and (FSubSelectEnd >= 0) and
     (ALine.Start < FStr.TextLength - FSubSelectEnd) and
     (ALine.Start + ALine.Length >= FSubSelectStart) then begin
    r.Top := (ALineNum - 1 + FShowRuler) * FLineHeight;
    r.Bottom := r.Top + FLineHeight;
    r.Left := FSubSelectStart - ALine.Start;
    r.Right := Min(ALine.Start + ALine.Length + 1, (FStr.TextLength - FSubSelectEnd) - ALine.Start);
    if r.Left < 0 then r.Left := 0 else
      r.Left := FLeftBarWidth + 6 + FStr.GetCaretX(r.Left,
        FFirstVisibleLine + ALineNum - 1) - FFirstVisiblePix - 1;
    r.Right := FLeftBarWidth + 6 + FStr.GetCaretX(r.Right,
      FFirstVisibleLine + ALineNum - 1) - FFirstVisiblePix - 1;
    BitBlt(ACanvas.Handle, r.Left, r.Top,
        r.Right - r.Left, r.Bottom - r.Top,
              ACanvas.Handle, r.Left, r.Top, NOTSRCCOPY);
  end;
  //ł́A͈͑I𒆂̓LbgKI͈͓ɓ悤A
  //E[ɃvX1Ăi_j
  //Lbg1hbgጩɂA2hbgƉE[{1זB
  if ASelArea.X < ASelArea.Y then begin
    r.Top := (ALineNum - 1 + FShowRuler) * FLineHeight;
    r.Bottom := r.Top + FLineHeight;
    if ASelArea.X < 0 then r.Left := 0 else
      r.Left := FLeftBarWidth + 6 + FStr.GetCaretX(ASelArea.X,
        FFirstVisibleLine + ALineNum - 1) - FFirstVisiblePix - 1;
    r.Right := FLeftBarWidth + 6 + FStr.GetCaretX(ASelArea.Y,
      FFirstVisibleLine + ALineNum - 1) - FFirstVisiblePix - 1;
    if (FSubSelectStart >= 0) and (FSubSelectEnd >= 0) then InflateRect(r, -1, -2);
    BitBlt(ACanvas.Handle, r.Left, r.Top,
        r.Right - r.Left {+ 1}, r.Bottom - r.Top,
              ACanvas.Handle, r.Left, r.Top, NOTSRCCOPY);
  end;
  //LbgsC
  //wihԂ̂ŁA`͍ŌɍsB
  if ACaretLine then begin
    r.Top := ((ALineNum - 1 + FShowRuler) + 1) * FLineHeight - 1;
    ACanvas.Pen.Color := FOptions.ColorCaretLine;
    ACanvas.MoveTo(FLeftBarWidth + 6, r.Top);
    ACanvas.LineTo(ClientWidth, r.Top);
  end;
end;

procedure TNotePad.DrawRuler(ACanvas: TCanvas; r: TRect);
var
  i: integer;
  s: WideString;
  x1: integer;
  //r1: TRect;
begin
  if Assigned(FOptions.BackGroundBMP) then begin
    BitBlt(ACanvas.Handle, r.Left, r.Top,
         r.Right - r.Left, r.Bottom - r.Top,
         FTileBitmap.Canvas.Handle, r.Left, r.Top, SRCCOPY);
  end else begin
    ACanvas.Brush.Color := FOptions.ColorBack;
    ACanvas.Brush.Style := bsSolid;
    ACanvas.FillRect(r);
  end;
  ACanvas.Brush.Style := bsClear;
  ACanvas.Pen.Color := FOptions.ColorRuler;
  ACanvas.Pen.Width := 1;
  ACanvas.Pen.Style := psSolid;
  ACanvas.Font.Color := FOptions.ColorRuler;
  ACanvas.Font.Style := [];
  x1 := FLeftBarWidth + 6 - 1;
  for i := 0 to ClientWidth div FFontWidth do begin
    case (i + FFirstVisiblePix div FFontWidth) mod 10 of
      0:begin
        s := inttostr(i + FFirstVisiblePix div FFontWidth);
        TextOutW(ACanvas.Handle, x1 + 2, FOptions.LineMargin div 2, PWChar(s), Length(s));
        ACanvas.MoveTo(x1, 0);
        ACanvas.LineTo(x1, FLineHeight - 1);
      end;
      5:begin
        ACanvas.MoveTo(x1, FLineHeight div 2);
        ACanvas.LineTo(x1, FLineHeight - 1);
      end;
      else begin
        ACanvas.MoveTo(x1, FLineHeight - FLineHeight div 4);
        ACanvas.LineTo(x1, FLineHeight - 1);
      end;
    end;
    inc(x1, FFontWidth);
  end;
  x1 := FStr.GetCaretX(FStr.SelEndPos.X, FStr.SelEndPos.Y);
  x1 := x1 + FLeftBarWidth + 6 - FFirstVisiblePix;
  BitBlt(ACanvas.Handle, x1, 0, FFontWidth - 1,
       FLineHeight, ACanvas.Handle, x1, 0, NOTSRCCOPY);

  ACanvas.MoveTo(r.Left, FLineHeight - 1);
  ACanvas.LineTo(r.Right, FLineHeight - 1);
end;

function TNotePad.DTDragEnter(const dataObj: IDataObject;
  grfKeyState: Integer; pt: TPoint; var dwEffect: Integer): HResult;
var
  AEffect: integer;
  AUnicode: boolean;
  etc: TFormatEtc;
begin
  FDDEnter := dataObj;
  //Windows.SetFocus(Handle);//Lbg\̎擾
  CaretHide;
  CaretReset;
  if (not Enabled) then begin
    dwEffect := DROPEFFECT_NONE;
    FDDEffect := DROPEFFECT_NONE;
    Result := S_OK;
    exit;
  end;
  AUnicode := dataObj.QueryGetData(FormatEtcUniTEXT) = s_OK;
  if (dataObj.QueryGetData(FormatEtcTEXT) = s_OK) or AUnicode then begin
    if AUnicode then FFormatEtc := FormatEtcUniTEXT else
      FFormatEtc := FormatEtcTEXT;
    etc := FormatEtcTEXT;
    etc.cfFormat := RegisterClipboardFormat('UniformResourceLocator');
    if dataObj.QueryGetData(etc) = s_OK then begin
      FFormatEtc := etc;
      AEffect := DROPEFFECT_LINK
    end else begin
      AEffect := DROPEFFECT_MOVE;
      if (FDDString = nil) or
         (dataObj <> IDataObject(FDDString)) or
         (grfKeyState and MK_CONTROL <> 0) then begin
        AEffect := DROPEFFECT_COPY;
        if BoxSelectMode then AEffect := DROPEFFECT_NONE;
      end;
    end;
  end else if dataObj.QueryGetData(FormatEtcHDROP) = s_OK then begin
    AEffect := DROPEFFECT_COPY;
    FFormatEtc := FormatEtcHDROP;
  end else begin
    AEffect := DROPEFFECT_NONE;
  end;//if
  dwEffect := AEffect;
  FDDEffect := AEffect;
  Result := S_OK;
  //OverNƂɂ
  DTDragOver(grfKeyState, pt, dwEffect);
end;

function TNotePad.DTDragLeave: HResult;
begin
  Result := S_OK;
  FDDEnter := nil;
  FDraging := false;
  CaretHide;
  initCaretPosition;
end;

function TNotePad.DTDragOver(grfKeyState: Integer; pt: TPoint;
  var dwEffect: Integer): HResult;
var
  m: TWMScroll;
  over: boolean;
begin
  //MouseMoveN̂DragOver͗ȂƂ̂
  //DragEnter炱Ăł
  dwEffect := FDDEffect;
  Result := S_OK;
  FDragCaret := ScreenToClient(pt);
  if (dwEffect = DROPEFFECT_NONE) or (FFormatEtc.cfFormat = CF_HDROP) then exit;
  if FDragCaret.Y > (ClientHeight - FLineHeight * 2) then begin
    m.ScrollCode := 1;
    WMVScroll(m);
  end else if FDragCaret.Y < (FLineHeight * (2 + FShowRuler)) then begin
    m.ScrollCode := 0;
    WMVScroll(m);
  end;
  FDragCaret := PointToPos(FDragCaret.X, FDragCaret.Y, true);
  FDragCaret.Y := Min(FDragCaret.Y, FStr.Count - 1);
  FDragCaret.X := Max(0, FStr.FromCaretX(FDragCaret.X, FDragCaret.Y, over));
  CaretHide;
  FDraging := true;
  initCaretPosition;
  //^^ւ̃hbOhbv̏ꍇACtrlA^Cf
  if (FDDString <> nil) and (FDDEnter = IDataObject(FDDString)) and
     ((FDDEffect = DROPEFFECT_MOVE) or (FDDEffect = DROPEFFECT_COPY)) then begin
    FDDEffect := DROPEFFECT_MOVE;
    if grfKeyState and MK_CONTROL <> 0 then FDDEffect := DROPEFFECT_COPY;
    dwEffect := FDDEffect;
  end;
end;

function TNotePad.DTDrop(const dataObj: IDataObject; grfKeyState: Integer;
  pt: TPoint; var dwEffect: Integer): HResult;
var
  Medium : TStgMedium;
  ADrop: PDropFiles;
  AFailed: boolean;
  PListA: PAnsiChar;
  PListW: PWChar;
  ss: array of WideString;
  BA, BW: boolean;
  DropText: WideString;
  aSelArea: TPoint;
begin
  dwEffect := FDDEffect;
  Result := S_OK;
  CaretHide;
  AFailed := Failed(dataObj.GetData(fFormatEtc, Medium));
  try
    if not AFailed then begin
      if dataObj.QueryGetData(FormatEtcHDROP) = S_OK then begin
        ADrop := pDropFiles(GlobalLock(Medium.hGlobal));
        try
          SetLength(ss, 0);
          if ADrop.fWide then begin
            PListW := PWChar(ADrop) + 10;
            while PListW^ <> #0 do begin
              SetLength(ss, Length(ss) + 1);
              ss[Length(ss) - 1] := PListW;
              inc(PListW, Length(WideString(PListW)) + 1);
            end;
          end else begin
            PListA := PAnsiChar(ADrop) + 20;
            while PListA^ <> #0 do begin
              SetLength(ss, Length(ss) + 1);
              ss[Length(ss) - 1] := PListA;
              inc(PListA, Length(AnsiString(PListA)) + 1);
            end;
          end;
          if Assigned(FFilesDroped) then FFilesDroped(self, ss);
        finally
          GlobalUnlock(Medium.hGlobal);
        end;
      end else begin
        DropText := '';
        BA := dataObj.QueryGetData(FormatEtcTEXT) = S_OK;
        BW := dataObj.QueryGetData(FormatEtcUniTEXT) = S_OK;
        if BA or BW then begin
          case FFormatEtc.cfFormat of
            CF_UNICODETEXT: DropText := PWChar(GlobalLock(Medium.hGlobal));
            else DropText := PAnsiChar(GlobalLock(Medium.hGlobal));
          end;
          if Length(DropText) > 0 then try
            aSelArea := FStr.GetSelectedArea(FDragCaret.Y);
            if not BoxSelectMode and ((FDragCaret.X <= aSelArea.X) or (FDragCaret.X >= aSelArea.Y)) then begin
              if dwEffect = DROPEFFECT_MOVE then begin
                FStr.DragMove(FDragCaret);
              end else begin
                FStr.BeginUpdate;
                FStr.SelStartPos := FDragCaret;
                FStr.SelEndToSelStart;
                FStr.RedrawCurrent;
                FStr.SelText := DropText;
                FStr.SelStartPos := FDragCaret;
                FStr.RedrawCurrent;
                FStr.EndUpdate;
                Windows.SetFocus(Handle);
              end;
            end;
          finally
            GlobalUnlock(Medium.hGlobal);
          end;
        end;
      end;
    end;
    FDraging := false;
    initCaretPosition;
  finally
    ReleaseStgMedium(Medium);
  end;
end;

function TNotePad.GetBoxSelected: boolean;
begin
  //BoxI𒆂ƑR}ĥŃvpeBKv
  Result := FStr.BoxMode and //(FStr.SelEnd <> FStr.SelStart) and
    (FStr.SelStartPos.Y <> FStr.SelEndPos.Y);
end;

function TNotePad.GetBoxSelectMode: boolean;
begin
  Result := FStr.BoxMode;
end;

function TNotePad.GetCanRedo: boolean;
begin
  result := FStr.CanRedo;
end;

function TNotePad.GetCanUndo: boolean;
begin
  Result := FStr.CanUndo;
end;

function TNotePad.GetCaretline: integer;
var
  i: integer;
begin
  Result := 0;
  for i := 0 to FStr.SelEndPos.Y - 1 do
    if FStr[i].LineEnd <> leWrap then inc(Result);
end;

function TNotePad.GetCompletePoint: TPoint;
var
  EP: TPoint;
begin
  EP := FStr.SelEndPos;
  EP.X := FStr.GetCaretX(EP.X, EP.Y) - FFirstVisiblePix;
  EP.Y := EP.Y - FFirstVisibleLine;
  Result.X := EP.X + FLeftBarWidth + 6 - 1;
  Result.Y := (EP.Y + FShowRuler) * FLineHeight + FOptions.LineMargin div 2 + FAscent + FDescent;
end;

function TNotePad.GetEditMode: THighlightMode;
begin
  if FCurrentMode = FModeHtml then Result := hmHtml else
  if FCurrentMode = FModeCSharp then Result := hmCSharp else
  if FCurrentMode = FModeDelphi then Result := hmDelphi else
  if FCurrentMode = FModeJScript then Result := hmJScript else
  if FCurrentMode = FModeVBScript then Result := hmVBScript else
  if FCurrentMode = FModePython then Result := hmPython else
  if FCurrentMode = FModeCpp then Result := hmCpp else
  if FCurrentMode = FModeJava then Result := hmJava else
  if FCurrentMode = FModePerl then Result := hmPerl else
  if FCurrentMode = FModeRuby then Result := hmRuby else
  if FCurrentMode = FModeVisualBasic then Result := hmVisualBasic else
  if FCurrentMode = FModeCss then Result := hmCss else
  if FCurrentMode = FModeBat then Result := hmBat else
  Result := hmDefault;
end;

function TNotePad.GetMarkCount: integer;
begin
  Result := FStr.MarkCount;
end;

function TNotePad.GetMarks(index: integer): integer;
begin
  Result := FStr.Marks[index];
end;

function TNotePad.GetModeExt(Value: THighlightMode): WideString;
var
  mode: TNotePadMode;
begin
  mode := ObjectFromHighlight(Value);
  result := mode.FileExts;
end;

function TNotePad.GetOptions: TNotePadOptions;
begin
  FOptions.ShowRuler := (FShowRuler = 1);
  FOptions.ShowLeftBar := (FShowLeftBar = 1);
  FOptions.TabCount := FStr.TabCount;//␳邩Kv
  FOptions.Font := FFont;
  FOptions.AlphaBlend := FAlphaBlend;//oCgɋl߂Ă邩
  Result := FOptions;
end;

function TNotePad.GetReturnedString: WideString;
begin
  result := FStr.GetReturnedString;
end;

function TNotePad.GetSelectedText: WideString;
begin
  Result := FStr.SelText;
end;

function TNotePad.GetSelectedWithoutCRLF: boolean;
var
  i: integer;
  AStart, AEnd: integer;
begin
  //I͂ĂAs͊܂܂ȂĂ̂Kv
  Result := false;
  if FStr.SelEnd = FStr.SelStart then exit;
  if FStr.SelEndPos.Y <> FStr.SelStartPos.Y then begin
    if FStr.BoxMode then exit;
    if FStr.SelStartPos.Y < FStr.SelEndPos.Y then begin
      AStart := FStr.SelStartPos.Y;
      AEnd := FStr.SelEndPos.Y;
    end else begin
      AEnd := FStr.SelStartPos.Y;
      AStart := FStr.SelEndPos.Y;
    end;
    for i := AStart to AEnd - 1 do begin
      if FStr[i].LineEnd = leLF then exit;
    end;
  end;
  Result := true;
end;

function TNotePad.GetSelectionLength: integer;
begin
  result := FStr.SelEnd - FStr.SelStart;
end;

function TNotePad.GetSelectionStart: integer;
begin
  result := FStr.SelStart;
end;

function TNotePad.GetText: WideString;
begin
  Result := FStr.Text;
end;

function TNotePad.GetTextCR: WideString;
begin
  Result := FStr.GetTextCR;
end;

function TNotePad.GetTextCRLF: WideString;
begin
  Result := FStr.GetTextCRLF;
end;

function TNotePad.GetTextIf: WideString;
begin
  Result := FStr.TextIf;
end;

function TNotePad.GetTextLength: integer;
begin
  result := FStr.TextLength;
end;

function TNotePad.GetTextLF: WideString;
begin
  Result := FStr.GetTextLF;
end;

function TNotePad.GetWordFromCaret: WideString;
var
  p : TPoint;
  APrevious, AStart, ALength: integer;
  AJump: boolean;
begin
  //IԂɂƂɁAStartEndƁAǂɂׂH
  //SelectWordFromCaretStartɂ̂ŁA킹B
  p := FStr.SelStartPos;
  SelectWordSub(p.X, p.Y, APrevious, AStart, ALength, AJump);
  Result := FStr.InternalSB.ToString(AStart, ALength);
end;

procedure TNotePad.ImeRegWord;
var
  word1: TRegisterWordW;
  s: WideString;
begin
  //http://nienie.com/~masapico/api_ImmConfigureIME.html
  if (SelectionLength <> 0) and (not BoxSelected) then begin
    s := FStr.SelText;
    word1.lpReading := nil;
    word1.lpWord := PWChar(s);
    ImmConfigureIMEW(GetKeyboardLayout(0), Handle, IME_CONFIG_REGISTERWORD, @word1);
  end;
end;

procedure TNotePad.InitBitmap;
var
  bf: _BLENDFUNCTION;
  r1, r2: TRect;
  x, y: integer;
  ABitmap: TBitmap;
begin
  //ɗ̂́A
  //(FBGWindow)傫ȂƂ
  //wiFύXꂽƂ
  //wi摜ύXꂽƂ
  //`悵̂ɏoĂȂƂ
  FNeedInitBMP := false;
  if (FOptions.BackGroundBMP = nil) or FOptions.BackGroundBMP.Empty then begin
  end else begin
    FTileBitmap.Height := FBGWindow.Height;
    FTileBitmap.Width := FBGWindow.Width;
    FTileBitmap.Canvas.Brush.Color := FOptions.ColorBack;
    FTileBitmap.Canvas.Brush.Style := bsSolid;
    FTileBitmap.Canvas.FillRect(FTileBitmap.Canvas.ClipRect);
    bf.BlendOp := AC_SRC_OVER;
    bf.BlendFlags := 0;
    bf.SourceConstantAlpha := FAlphaBlend;
    bf.AlphaFormat :=  0;
    ABitmap := TBitmap.Create;
    try
      ABitmap.Assign(FOptions.BackGroundBMP);
      r2 := Rect(0, 0, ABitmap.Width, ABitmap.Height);
      for x := 0 to FTileBitmap.Width div ABitmap.Width do begin
        for y := 0 to FTileBitmap.Height div ABitmap.Height do begin
          r1 := Rect(x * r2.Right, y * r2.Bottom, r2.Right, r2.Bottom);
          AlphaBlend(FTileBitmap.Canvas.Handle, r1.Left, r1.Top,
                     r1.Right, r1.Bottom,
                     ABitmap.Canvas.Handle, r2.Left, r2.Top,
                     r2.Right, r2.Bottom, bf);
        end;
      end;
    finally
      ABitmap.Free;
    end;
  end;
end;

procedure TNotePad.initCaretPosition;
var
  EP: TPoint;
  CX, CY: integer;
  H: integer;
  CForm: TCompositionForm;
  LFont: TLogFontW;
  i: integer;
  s: WideString;
begin
  if FDraging then begin
    EP := FDragCaret;
    EP.X := FStr.GetCaretX(EP.X, EP.Y) - FFirstVisiblePix;
    dec(EP.Y, FFirstVisibleLine);
    CX := EP.X + FLeftBarWidth + 6 - 1;
    CY := (EP.Y + FShowRuler) * FLineHeight + FOptions.LineMargin div 2;
    if (EP.X >= 0) and (EP.Y >= 0) then begin
      CaretShow;
      SetCaretPos(CX, CY);
    end;
    exit;
  end;
  EP := FStr.SelEndPos;
  EP.X := FStr.GetCaretX(EP.X, EP.Y) - FFirstVisiblePix;
  EP.Y := EP.Y - FFirstVisibleLine;
  CX := EP.X + FLeftBarWidth + 6 - 1;
  CY := (EP.Y + FShowRuler) * FLineHeight + FOptions.LineMargin div 2;
  if (EP.X >= 0) and (EP.Y >= 0) then begin
    if Focused or FCompleting then CaretShow;
    SetCaretPos(CX, CY);
  end;
  H := ImmGetContext(Handle);
  if H <> 0 then begin
    //`̂ptCurrentPos͕ύXΏ
    CForm.dwStyle := CFS_POINT;
    ImmGetCompositionWindow(H, @CForm);
    if (CForm.ptCurrentPos.X <> CX) or (CForm.ptCurrentPos.Y <> CY) then begin
      CForm.ptCurrentPos.X := CX;
      CForm.ptCurrentPos.Y := CY;
      ImmSetCompositionWindow(H, @CForm);
    end;
    CForm.dwStyle := CFS_RECT;
    CForm.rcArea.Left := FLeftBarWidth + 6 - 1;
    CForm.rcArea.Top := 0;
    CForm.rcArea.Bottom := ClientHeight;
    CForm.rcArea.Right := Max(0, ClientWidth - FFontWidth);
    ImmSetCompositionWindow(H, @CForm);
    if (FFont <> nil) and (FNeedInitImeFont) then begin
      //MS-IME2007Ń`̂őΏ悤
      GetObjectW(FFont.Handle, 32{SizeOf(TLogFont)}, @LFont);
      //ȂLFontr[ɖ܂񂾂낤H
      s := FFont.Name;
      for i := 0 to 31 do begin
        if i < Length(s) then
          LFont.lfFaceName[i] := s[i + 1]
        else
          LFont.lfFaceName[i] := #0;
      end;
      ImmSetCompositionFontW(H, @LFont);
      FNeedInitImeFont := false;
    end;
    ImmReleaseContext(Handle, H);
  end;
end;

procedure TNotePad.initScroll;
var
  info1: TScrollInfo;
begin

  FVisibleLineCount := Max(5, (ClientHeight div FLineHeight) + 1 - FShowRuler);
  info1.cbSize := 28;//sizeof(TScrollInfo);
  info1.fMask := 31;//SIF_ALL or SIF_DISABLENOSCROLL;
  info1.nMin := 0;
  info1.nMax := Max(FFirstVisibleLine + FVisibleLineCount - 1, FStr.Count + 2);
  info1.nPage := FVisibleLineCount;
  info1.nPos := FFirstVisibleLine;
  info1.nTrackPos := 0;
  SetScrollInfo(Handle, SB_VERT, info1, true);
  FVisibleByteCount := Max(MIN_WRAPBYTE, (ClientWidth - (FLeftBarWidth + 6)) div FFontWidth + 1);
  info1.cbSize := 28;//sizeof(TScrollInfo);
  //info1.fMask := 31;//SIF_ALL or SIF_DISABLENOSCROLL;
  info1.fMask := 23;//SIF_ALL;
  info1.nMin := 0;
  info1.nMax := Max(FFirstVisiblePix div FFontWidth + FVisibleByteCount - 1, FStr.WrapByte + 2);
  info1.nPage := FVisibleByteCount;
  info1.nPos := FFirstVisiblePix div FFontWidth;
  info1.nTrackPos := 0;
  SetScrollInfo(Handle, SB_HORZ, info1, true);
end;

procedure TNotePad.InputChar(C: Word);
begin
  FColKeeping := -1;
  FStr.SelText := WideChar(C);
end;

procedure TNotePad.InputKey(Message: TWMKey);
//Delphi̎dlȂ̂AAltL[Ȃ牽ĂKeyDownȂB
//FormIsShortCutŁAAltޑ߂炦Ăɓn΂B
//`IAlt+Shift+J[\ōs߂̑[uB
begin
  WMKeyDown(message);
end;

procedure TNotePad.InputKeyChar(Message: TWMKey);
begin
  WMChar(Message);
end;

procedure TNotePad.LoadFromString(const Value: WideString);
begin
  FColKeeping := -1;
  FStr.BeginUpdate;
  FStr.OnModified := nil;
  FStr.Text := Value;
  FStr.ClearUndo;
  FStr.ClearModified;
  FModified := false;
  FStr.OnModified := StringOnModified;
  FStr.EndUpdate;
end;

function TNotePad.NeedWrap: boolean;
var
  i: integer;
begin
  Result := false;
  //ȂƂĂ炾Ǝv
  //if ( (Height == 0) || (Width ==0) ) return false;
  case FWordWrap of
    wwRight:begin
      FFirstVisiblePix := 0;
      i := ((ClientWidth - (FLeftBarWidth + 6)) div FFontWidth) - 3;
    end;
    wwCount:begin
      i := FWordWrapCount;
    end;
    else begin
      i := MAX_WRAPBYTE;
    end;
  end;
  i :=  Max(MIN_WRAPBYTE, Min(MAX_WRAPBYTE, i));
  if (not FFontChanged) and (FStr.WrapByte = i) then exit;
  FFontChanged := false;
  FColKeeping := -1;
  FStr.BeginUpdate;
  FStr.WrapByte := i;
  CalcLeftBar;
  if  i < ((ClientWidth - (FLeftBarWidth + 6)) div FFontWidth) then FFirstVisiblePix := 0;
  initScroll;
  FStr.RedrawCurrent;
  FStr.EndUpdate;
  Invalidate;
  Result := true;
end;

procedure TNotePad.Paint;
var
  r:TRect;
  i: integer;
  ARgnData: PRgnData;
  ARgnSize: Cardinal;
begin
  inherited;
  CaretHide;
  if not IsRectEmpty(Canvas.ClipRect) then begin
    if FNeedInitBMP or (Assigned(FOptions.BackGroundBMP) and
      ((FTileBitmap.Width < FBGWindow.Width) or
      (FTileBitmap.Height < FBGWindow.Height))) then InitBitmap;
    ARgnSize := GetRegionData(FRgn, 0, nil);
    ARgnData := AllocMem(ARgnSize);
    try
      GetRegionData(FRgn, ARgnSize, ARgnData);
      DrawCanvas(FBGWindow.Canvas, ARgnData);
      for i := 0 to ARgnData.rdh.nCount - 1 do begin
        r := PRect(integer(@ARgnData.Buffer) + i * sizeof(TRect))^;
        BitBlt(Canvas.Handle, r.Left, r.Top,
               r.Right - r.Left, r.Bottom - r.Top,
               FBGWindow.Canvas.Handle, r.Left, r.Top, SRCCOPY);
      end;
    finally
      FreeMem(ARgnData, ARgnSize);
    end;
  end;
  if Focused or FDraging or FCompleting then initCaretPosition;
end;

procedure TNotePad.Paste;
var
  s: WideString;
begin
  s := GetFromClipBrd;
  if Length(s) > 0 then begin
    FColKeeping := -1;
    if HasBoxText then begin
      FColKeeping := -1;
      FStr.SetSelTextBox(s);
    end else begin
      FStr.SelText := s;
      FStr.BeginUpdate;
      FStr.EndUpdate;
    end;
  end;
end;

function TNotePad.PointToPos(X, Y: integer; Split: boolean): TPoint;
var
  C: PWideChar;
  w: integer;
  xx, yy:integer;
  over: boolean;
begin
  yy := FFirstVisibleLine + (Y div FLineHeight) - FShowRuler;
  Result := Point(Max(0, X - (FLeftBarWidth + 6) + FFirstVisiblePix),
    Max(0, yy));
  if yy < FFirstVisibleLine then begin
    //ŝȂꏊ͉t[
    w := Result.X mod FFontWidth;
    dec(Result.X, w);
    if Split and (w > FFontWidth div 2) then inc(Result.X, FFontWidth);
    exit;
  end;
  yy := Min(FStr.Count - 1, Result.Y);
  xx := FStr.FromCaretX(Result.X, yy, over);
  if over then exit;//sȏɉEɍs΃t[
  C := FStr.InternalSB.ToPointer(FStr[yy].Start + xx);
  xx := FStr.GetCaretX(xx, yy);
  if C^ = WideChar(#$09) then begin
    w := (FFontWidth * FStr.TabCount) - (xx mod (FFontWidth * FStr.TabCount));
  end else w := GetGlyphWidth(C^);
  if Split then begin
    if Result.X - xx > (w div 2) then
      Result.X := xx + w else Result.X := xx;
  end else Result.X := xx;
end;

procedure TNotePad.Reconvert;
//\Iĕϊ
//ڂ͎󓮓IĕϊWMImeRequestQ
const
  SCS_CAP_SETRECONVERTSTRING = $00000004;
  SCS_SETRECONVERTSTRING     = $00010000;
var
  i: integer;
  r: TReconvertString;
  pr: Pointer;
  s : WideString;
  ss: packed array of WideChar;
  H: HIMC;
begin
  //MS-IME͉s܂łĕϊāACRLF𑗂Ă̂Ŗ肪N
  if not SelectedWithoutCRLF then exit;
  s := FStr.SelText;
  if (ImmGetProperty(GetKeyboardLayout(0), IGP_SETCOMPSTR) and SCS_CAP_SETRECONVERTSTRING) = 0 then exit;
  if SelectionLength > 0 then begin
    i := SelectionLength;
    SelectionStart := SelectionStart + i;
    SelectionLength := - i;
    Update;
  end;
  H := ImmGetContext(Handle);
  if H <> 0 then try
    s := FStr.SelText;
    SetLength(ss, Length(s) + 1);
    for i := 0 to Length(s) - 1 do ss[i] := s[i + 1];
    ss[Length(s)] := #0;
    r.dwSize := 32 + (Length(ss) * 2);
    r.dwVersion := 0;
    r.dwStrOffset := 32;
    r.dwStrLen := Length(s);
    r.dwCompStrOffset := 0;
    r.dwCompStrLen := r.dwStrLen;
    r.dwTargetStrOffset := r.dwCompStrOffset;
    r.dwTargetStrLen := r.dwCompStrLen;
    GetMem(pr, 32 + (Length(ss) * 2));
    try
      CopyMemory(pr, @r, 32);
      CopyMemory(Pointer(integer(pr) + 32), PWChar(ss), Length(ss) * 2);
      ImmSetCompositionStringW(H, SCS_SETRECONVERTSTRING, pr, 32 + Length(ss) * 2 , nil, 0 );
      r := PReconvertString(pr)^;
    finally
      FreeMem(pr);
    end;
  finally
    ImmReleaseContext(Handle, H);
  end;
end;

procedure TNotePad.Redo;
begin
  FColKeeping := -1;
  FStr.Redo;
end;

procedure TNotePad.ReplaseWithoutSelect(const Value: WideString; AStart,
  ALength: integer);
begin
  FColKeeping := -1;
  FStr.ReplaseWithoutSelect(Value, AStart, ALength);
end;

procedure TNotePad.ScrollToCaret;
begin
  FForceScroll := true;
  FStr.BeginUpdate;
  FStr.EndUpdate;
end;

procedure TNotePad.SearchHighlight(Value: integer);
//Osȏɂ܂錟[hT|[gƃ{glbNɂȂ
var
  ALine, ALine2: PLineData;
  Start1, End1: integer;
  p: integer;
  i: integer;
//  AFlag: byte;
begin
  if (FSearchText = '')or
    ((FOptions.ColorHighlight = clNone) and (FOptions.StyleHighlight = [])) then exit;
  ALine := FStr[Value];
  ALine2 := ALine;
  Start1 := ALine.Start;
  End1 := Start1 + ALine.Length;
  if (Value > 0) and (FStr[Value - 1].LineEnd = leWrap) then begin
    dec(Value);
    ALine := FStr[Value];
  end;
  p := ALine.Start;//p͐܂ԂȂs

  SearchBM.SubStr := FSearchText;
  SearchBM.IgCase := FSearchIgCase;

  repeat
    p := SearchBM.Search(FStr.InternalSB.ToPointer(0), FStr.TextLength, p, End1 + Length(FSearchText), true);
    if (p > 0) and (p - 1 - Start1 + Length(FSearchText) - 1 >= 0) then begin
      for i := 0 to Length(FSearchText) - 1 do
        if p - 1 - Start1 + i >= 0 then begin
          if p - 1 - Start1 + i >= ALine2.Length then begin
            p := 0;
            break;
          end else LineTokens[p - Start1 - 1 + i] := toHighlight;
        end;
    end;
  until p = 0;
end;

procedure TNotePad.Select(Start, Len: integer);
var
  i: integer;
begin
  FColKeeping := -1;
  FStr.BeginUpdate;
  i := FStr.TextLength;
  FStr.SelStart := Min(i, Start);
  FStr.SelEnd := Min(i, (Start + Len));
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.SelectAll;
begin
  BoxSelectMode := false;
  FColKeeping := -1;
  FStr.BeginUpdate;
  FStr.SelStart := 0;
  FStr.SelEnd := FStr.TextLength;
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.SelectWord(X, Y: integer);
var
  APrevious, AStart, ALength: integer;
  AJump: boolean;
begin
  SelectWordSub(X, Y, APrevious, AStart, ALength, AJump);
  FStr.BeginUpdate;
  FStr.SelStart := AStart;
  FStr.SelEnd := FStr.SelStart + ALength;
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.SelectWordFromCaret;
var
  p : TPoint;
begin
  FStr.BeginUpdate;
  //AŌĂ΂ꂽꍇStart < EndɂȂĂ͂Ȃ̂ŁA
  //Start <> End̎AStartɍ킹ƑOƓʂɂłB
  FStr.SelEndtoSelStart;
  p := FStr.SelEndPos;
  SelectWord(p.X, p.Y);
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.SelectWordSub(X, Y: integer; out APrevious, AStart,
  ALength: integer; out canJump: boolean);
var
  StartX, StartY, EndX, EndY: integer;
  ALine: PLineData;
  AString: WideString;
  AParser: TLineParser;
begin
  StartX := 0;
  StartY := Y;
  while StartY > 0 do begin
    ALine := FStr[StartY - 1];
    if ALine.LineEnd <> leWrap then break;
    dec(StartY);
    inc(StartX, ALine.Length);//leWrapLengthLineLength
  end;
  EndX := 0;
  EndY := Y;
  while true do begin
    ALine := FStr[EndY];
    inc(EndX, ALine.Length);
    if ALine.LineEnd <> leWrap then break;
    inc(EndY);
  end;
  ALine := FStr[StartY];
  AString := FStr.InternalSB.ToString(ALine.Start, EndX + StartX);
  AParser := TLineParser.Create(AString, StartX + X);
  try
    APrevious := ALine.Start + AParser.Previous;
    AStart := ALine.Start + AParser.Start;
    ALength := AParser.Count;
    canJump := AParser.CanJump;
  finally
    AParser.Free;
  end;
end;

procedure TNotePad.SetBoxSelectMode(Value: boolean);
begin
  if FStr.BoxMode = Value then exit;
  FStr.BoxMode := Value;
  Invalidate;
  if Assigned(FDocumentParent) then FDocumentParent.Invalidate;
  if Assigned(FDocumentChild) then FDocumentChild.Invalidate;
  DoViewChanged;
end;

procedure TNotePad.SetCaretLine(const Value: integer);
var
  p1, p2: TPoint;
begin
  //I𒆂łĂLbĝslȂ
  //sړړIƂ̂ŁAXقȂĂĂYȂړȂ
  p1 := FStr.LinesPosToWrapPos(0, Value);
  p2 := FStr.SelEndPos;
  if (p1.Y = p2.Y) then exit;
  FColKeeping := -1;
  FStr.BeginUpdate;
  FStr.SelStartPos := p1;
  FStr.SelEndtoSelStart;
  FStr.EndUpdate;
end;

procedure TNotePad.SetCodePage(const Value: TCodePage);
begin
  if FCodePage = Value then exit;
  FCodePage := Value;
  Invalidate;
end;

procedure TNotePad.SetDocumentParent(const Value: TNotePad);
begin
  //^u܂Ԃ̓^C~O̓Av
  if Value = nil then begin
    if Assigned(FDocumentParent) then begin
      FStr := FNotePadString;
      FDocumentParent.FDocumentChild := nil;
      FDocumentParent := nil;
      FFirstVisibleLine := 0;
      FFirstVisiblePix := 0;
      if Assigned(FSaveMode) then FCurrentMode := FSaveMode;
    end;
  end else begin
    FStr := Value.FNotePadString;
    //ɐeƂ͉Kv
    //eȂ_Ń[hۑĂ
    if Assigned(FDocumentParent) then
      FDocumentParent.FDocumentChild := nil
    else
      FSaveMode := FCurrentMode;
    Value.FDocumentChild := Self;
    FDocumentParent := Value;
    FFirstVisibleLine := Value.FFirstVisibleLine;
    FFirstVisiblePix := Value.FFirstVisiblePix;
    FCurrentMode := Value.FCurrentMode;
  end;
  CalcLeftBar;
  initScroll;
  Invalidate;
  FStr.BeginUpdate;
  FStr.EndUpdate;
end;

procedure TNotePad.SetEditMode(Value: THighlightMode);
var
  ANewMode: TNotePadMode;
  parser: TNotePadParser;
begin
  ANewMode := ObjectFromHighlight(Value);
  if FCurrentMode = ANewMode then exit;
  FCurrentMode := ANewMode;
  parser := FCurrentMode.NewParser(FStr);
  try
    parser.ParseRange(0, FStr.Count - 1);
    Invalidate;
  finally
    parser.Free;
  end;
  //p[X͓xvȂ̂œf[^ύX
  if Assigned(FDocumentParent) then FDocumentParent.FCurrentMode := ANewMode;
  if Assigned(FDocumentChild) then FDocumentChild.FCurrentMode := ANewMode;
end;

procedure TNotePad.SetModeExt(mode: THighlightMode;
  const Value: WideString);
begin
  ObjectFromHighlight(mode).FileExts := Value;
end;

procedure TNotePad.SetModeFromPath(pathName: WideString);
var
  ext: WideString;
  AMode: THighlightMode;
  function ExtractFileExtW(const FileName: widestring): widestring;
  var
    i, j: integer;
  begin
    i := 0;
    for j := 1 to Length(Filename) do
      if (FileName[j] = WideChar('\')) or (FileName[j] = WideChar(':'))
        or (FileName[j] = '.') then i := j;
    if (i > 0) and (FileName[i] = '.') then
      Result := System.Copy(FileName, i, MaxInt) else
      Result := '';
  end;
begin
  AMode := hmDefault;
  ext := ExtractFileExtW(pathName);
  if Length(ext) > 0 then begin
    if FModeHtml.HasExt(ext)        then AMode := hmHtml    else
    if FModeCSharp.HasExt(ext)      then AMode := hmCSharp  else
    if FModeDelphi.HasExt(ext)      then AMode := hmDelphi  else
    if FModeJScript.HasExt(ext)     then AMode := hmJScript else
    if FModeVBScript.HasExt(ext)    then AMode := hmVBScript else
    if FModePython.HasExt(ext)      then AMode := hmPython else
    if FModeCpp.HasExt(ext)         then AMode := hmCpp else
    if FModeJava.HasExt(ext)        then AMode := hmJava else
    if FModePerl.HasExt(ext)        then AMode := hmPerl else
    if FModeRuby.HasExt(ext)        then AMode := hmRuby else
    if FModeVisualBasic.HasExt(ext) then AMode := hmVisualBasic else
    if FModeCss.HasExt(ext)         then AMode := hmCss else
    if FModeBat.HasExt(ext)         then AMode := hmBat else
    ;
  end;
  if AMode = hmDefault then begin
    if (Length(pathName) > 3) and (pathName[Length(pathName)] = WideChar(']'))
      then AMode := hmHtml;
  end;
  EditMode := AMode;
end;

procedure TNotePad.SetModified(Value: boolean);
begin
  FModified := value;
  if not value then FStr.ClearModified;
end;

procedure TNotePad.SetOptions(const Value: TNotePadOptions);
begin
  FColKeeping := -1;
  FOptions := Value;
  if FOptions.ShowRuler then FShowRuler := 1 else FShowRuler := 0;
  if FOptions.ShowLeftBar then FShowLeftBar := 1 else FShowLeftBar := 0;
  FOptions.LineMargin := Min(16, Max(0, FOptions.LineMargin));
  FNotePadString.TabCount := FOptions.TabCount; //IWi̕ɓ`
  FFont.Assign(FOptions.Font);
  FFont.Charset := GetFontCharSet(FFont.Name);
  FFont.Pitch := fpDefault;
  CalcFont;
  CalcLeftBar;
  initScroll;
  //Lbg̑傫ςȂႢȂ
  //ŕςƕ\ȂȂႤ
  //Ȃ̂킩Ȃ̂ŁAtH[JXőΉ
  //Perform_݂Ȃ̂SendMessage
  {if Focused then begin
    SendMessage(Handle, WM_KILLFOCUS, 0, 0);
    SendMessage(Handle, WM_SETFOCUS, 0, 0);
  end;}
  if Focused then begin
    //ǂhideĂȂ̂ɏ鎖AȂ񂩂OhideĂȂƁA
    //Rɏ̂ɓł͏ĂȂƂɂȂă_ɂȂ炵B
    CaretHide;
    CaretReset;
  end;

  FAlphaBlend := FOptions.AlphaBlend;
  FNeedInitBMP := true;
  Invalidate;
end;

procedure TNotePad.SetSearchigCase(Value: boolean);
begin
  FSearchIgCase := Value;
  Invalidate;
  DoViewChanged;
end;

procedure TNotePad.SetSearchText(const Value: WideString);
begin
  FSearchText := Value;
  Invalidate;
  DoViewChanged;
end;

procedure TNotePad.SetSelectedText(Value: WideString);
begin
  FColKeeping := -1;
  FStr.SelText := Value;
end;

procedure TNotePad.SetSelectionLength(Value: integer);
begin
  BoxSelectMode := false; //sIہABOXł͂܂ꍇ邽
  FColKeeping := -1;
  FStr.BeginUpdate;
  FStr.SelEnd := Min(FStr.TextLength, (FStr.SelStart + Value));
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.SetSelectionStart(Value: integer);
begin
  FColKeeping := -1;
  if value <> SelectionStart then begin
    FStr.BeginUpdate;
    FStr.SelStart := Min(FStr.TextLength, Value);
    FStr.SelEndtoSelStart;
    FStr.RedrawCurrent;
    FStr.EndUpdate;
  end;
end;

procedure TNotePad.SetText(const Value: WideString);
begin
  FColKeeping := -1;
  FStr.Text := Value;
end;

procedure TNotePad.SetWordWrap(Value: TWordWrap);
begin
  FColKeeping := -1;
  FWordWrap := Value;
  initScroll;
  Invalidate;
end;

procedure TNotePad.SetWordWrapCount(Value: integer);
begin
  Value := Max(MIN_WRAPBYTE, Min(MAX_WRAPBYTE, Value));
  if Value = FWordWrapCount then exit;
  FColKeeping := -1;
  FWordWrapCount := Value;
  initScroll;
  Invalidate;
end;

procedure TNotePad.StringOnChanged(Sender: TObject);
var
  ACountChanged: boolean;
  ANewFirstLine, ANewFirstPix: integer;
  ASelEndY, ASelEndX: integer;
  AStart, AEnd: integer;
  ALine: PLineData;
  i: integer;
  DrawY: integer;
  r: TRect;
begin
  //œK̂ߏdĂяo
  //eqĂԁBt̓iVB
  //ViewChange͂ꂼ܂łɂĂ
  if Assigned(FDocumentChild) then FDocumentChild.StringOnChanged(Sender);
  r := Rect(0, 0, ClientWidth, FLineHeight);
  if FShowRuler > 0 then InvalidateRect(Handle, @r, false);
  ACountChanged := FStr.Count <> FStr.CountBeforeChanged;
  if ACountChanged then CalcLeftBar;
  initScroll;
  if FForceScroll or ((Focused or FCompleting) and (not FRulerDrag)) then begin
    FForceScroll := false;
    ANewFirstLine := FFirstVisibleLine;
    ANewFirstPix := FFirstVisiblePix;
    ASelEndY := FStr.SelEndPos.Y;
    ASelEndX := FStr.SelEndPos.X;
    ASelEndX := FStr.GetCaretX(ASelEndX, ASelEndY);
    if (ASelEndY < FFirstVisibleLine + 3) and (FFirstVisibleLine > 0) then
      ANewFirstLine := Max(0, ASelEndY - 2)
    else if (ASelEndY > FFirstVisibleLine + FVisibleLineCount - 4) and
            (FFirstVisibleLine < FStr.Count - FVisibleLineCount + 3) then
      ANewFirstLine := Max(0, ASelEndY + 4 - FVisibleLineCount);
    ANewFirstLine := Min(ANewFirstLine, FStr.Count - 1);
    if (ASelEndX < FFirstVisiblePix + FFontWidth * 3) and (FFirstVisiblePix > 0) then
      ANewFirstPix := Max(0, (ASelEndX div FFontWidth - 2) * FFontWidth)
    else if (ASelEndX > FFirstVisiblePix + (FVisibleByteCount - 4) * FFontWidth) and
            (FFirstVisiblePix div FFontWidth < FStr.WrapByte - FVisibleByteCount + 3) then
      ANewFirstPix := Max(0, (ASelEndX div FFontWidth + 4 - FVisibleByteCount) * FFontWidth);
    ANewFirstPix := Min(ANewFirstPix, (FStr.WrapByte - 1) * FFontWidth);
    if (FFirstVisibleLine <> ANewFirstLine) or (FFirstVisiblePix <> ANewFirstPix) then begin
      FFirstVisibleLine := ANewFirstLine;
      FFirstVisiblePix := ANewFirstPix;
      initScroll;
      Invalidate;
      if Assigned(FViewChanged) then FViewChanged(self);
      exit;
    end;
  end else begin
    //ANeBułACountȏɂȂȂ悤ɏcXN[͕Kv
    ANewFirstLine := FFirstVisibleLine;
    if ANewFirstLine > FStr.Count - 3 then begin
      ANewFirstLine := Max(0, FStr.Count - 3);
    end;
    //[[hbOȂXN[鑀
    ANewFirstPix := FFirstVisiblePix;
    if FRulerDrag then begin
      if (FStr.WrapByte * FFontWidth) < ANewFirstPix then ANewFirstPix := (FStr.WrapByte * FFontWidth);
      if (FStr.WrapByte - FVisibleByteCount + 3) * FFontWidth > ANewFirstPix then ANewFirstPix := (FStr.WrapByte - FVisibleByteCount + 3) * FFontWidth;
    end;
    if (ANewFirstLine <> FFirstVisibleLine) or (ANewFirstPix <> FFirstVisiblePix) then begin
      FFirstVisibleLine := ANewFirstLine;
      FFirstVisiblePix := ANewFirstPix;
      initScroll;
      Invalidate;
      if Assigned(FViewChanged) then FViewChanged(self);
      exit;
    end;
  end;

  ////XN[Ȃꍇ
  AStart := FStr.ChangedStart;
  AEnd := Min(FStr.ChangedEnd, FStr.Count - 1);
  if not ACountChanged then begin
    if FStr.ParseDone then begin
      //eLXgύXꂽ܂Ԃsꂽp[Xsꂽ
      //܂ԂȂsׂăp[XAʂႤƑSʖ
      //ʂł܂ԂsׂĂ`悵Ȃ΂Ȃ
      i := 0;
      repeat
        inc(i);
        ALine := FStr[AEnd + i - 1];
      until ALine.LineEnd <> leWrap;
      //eLXgύXiVJ[\ړ̎
    end else i := 1;
    DrawY := (AStart - FFirstVisibleLine + FShowRuler) * FLineHeight;
    //p[X͍s̍ŌコvĂΖωƂƂ炦鐫Ȃ̂ŁA
    //ω̂̍s̍sɖ߂悤ȃg[Nƍĕ`悳ȂƂB
    //dȂ̂ŏɈs`悷KvB
    r := Rect(0, DrawY, ClientWidth, FLineHeight * (AEnd - AStart + i + 1) + DrawY);
    IntersectRect(r, r, ClientRect);
    InvalidateRect(Handle, @r, false);
    if not FOptions.LeftBarCountWrap then begin
      r := Rect(0, 0, FLeftBarWidth + 3, ClientHeight);
      InvalidateRect(Handle, @r, false);
    end;
    if Assigned(FViewChanged) then FViewChanged(self);
    exit;
  end;
  initScroll;
  if Assigned(FViewChanged) then FViewChanged(self);
  Invalidate;
end;

procedure TNotePad.StringOnModified(Sender: TObject);
begin
  if not FModified then begin
    FModified := true;
    if Assigned(FModifiedChanged) then FModifiedChanged(self);
  end;
end;

procedure TNotePad.StringOnNeedParse(Sender: TObject);
var
  parser: TNotePadParser;
  AStart, AEnd, i: integer;
begin
  //eLXgɕύXꍇA
  //̉ӏ܂ސ܂ԂȂsSp[X
  //̌Av܂Ńp[X̂ŁAi͍ŏlAEnd + 1ɂȂ
  parser := FCurrentMode.NewParser(FStr);
  try
    AStart := Max(0, FStr.ChangedStart);
    AEnd := Min(FStr.Count - 1, FStr.ChangedEnd);
    while (AStart > 0) and (FStr[AStart - 1].LineEnd = leWrap) do dec(AStart);
    while FStr[AEnd].LineEnd = leWrap do inc(AEnd);
    i := parser.ParseRange(AStart, AEnd);
    if i > AEnd + 1 then begin
      Invalidate;
      if Assigned(FDocumentChild) then FDocumentChild.Invalidate;
    end;
  finally
    parser.Free;
  end;
end;

function TNotePad.StyleFromToken(const C: WideChar): TFontStyles;
begin
  case C of
    toControl:
      Result := FOptions.StyleControl;
    toHighlight:
      Result := FOptions.StyleHighlight;
    toSBSymbol,
    toHereStart,
    toSymbolNoDiv:
      Result := FOptions.StyleSymbol;
    toQString:
      Result := FOptions.StyleString;
    toSBInt,
    toHex:
      Result := FOptions.StyleInteger;
    toReserve:
      Result := FOptions.StyleReserve;
    toReserve2:
      Result := FOptions.StyleReserve2;
    toReserve3:
      Result := FOptions.StyleReserve3;
    toAnk,
    toAnkNoReserve:
      Result := FOptions.StyleAnk;
    toAlphabetic,
    toHangul:
      Result := FOptions.StyleAlphabetic;
    toDBSymbol:
      Result := FOptions.StyleDBSymbol;
    toDBInt:
      Result := FOptions.StyleDBInt;
    toDBAlph:
      Result := FOptions.StyleDBAlph;
    toDBHira:
      Result := FOptions.StyleDBHira;
    toDBKana:
      Result := FOptions.StyleDBKana;
    toDBKanji:
      Result := FOptions.StyleDBKanji;
    toSBKana,
    toSBKanaSymbol:
      Result := FOptions.StyleSBKana;
    toUnicode:
      Result := FOptions.StyleUnicode;
    toUrl,
    toUrlStart,
    toFtpStart,
    toHttpsStart,
    toRtspStart,
    toPnmStart,
    toMmsStart,
    toNewsStart,
    toNntpStart:
      Result := FOptions.StyleLink;
    toComment,
    toVBRem,
    toCommentBrock0,
    toCommentBrock1,
    toCommentBrock1Start,
    toCommentBrock2,
    toCommentBrock2Start,
    toCommentBrock3,
    toCommentBrock3Start,
    toPodComment,
    toPodCommentEnd:
      Result := FOptions.StyleComment;
    toSingleQuotation,
    toDoubleQuotation,
    toAtDoubleQuot,
    toAtDoubleQuotStart,
    toDelphiChar,
    toDelphiCharHex,
    toRegExp,
    toMatchStart,
    toSubStart,
    toSubMatch:
      Result := FOptions.StyleString;
    toTagStart,
    toTagEnd,
    toTagQuestStart,
    toTagQuestEnd,
    toInsideTag:
      Result := FOptions.StyleHtmlTag;
    toDirective:
      Result := FOptions.StylePrefix;
    toHereStartWord,
    toHereEnd:
      Result := FOptions.StyleHere;
    else Result := [];
  end;
end;

procedure TNotePad.SubSelect(AStart, AEnd: integer);
begin
  FSubSelectStart := AStart;
  FSubSelectEnd := AEnd;
  Invalidate;
end;

function TNotePad.ToPointer(index: integer): PWChar;
begin
  Result := FStr.InternalSB.ToPointer(index);
end;

procedure TNotePad.TrackMouseEventSub;
var
  struct: TTrackMouseEvent;
begin
  struct.cbSize := 16{sizeof(TTrackMouseEvent)};
  struct.dwFlags := TME_HOVER;
  struct.hwndTrack := Handle;
//  struct.dwHoverTime := HOVER_DEFAULT;
  SystemParametersInfo(SPI_GETKEYBOARDSPEED, 0, @(struct.dwHoverTime), 0);
  TrackMouseEvent(struct);
end;

procedure TNotePad.Undo;
begin
  FColKeeping := -1;
  FStr.Undo;
end;

procedure TNotePad.WMChar(var Message: TWMChar);
var
  p: TPoint;
  aIndentString: WideString;
  KeyChar: Word;
  AAutoIndent: boolean;
begin
  inherited;
  if (Message.CharCode = 255) and (Message.Unused <> 0) then
    KeyChar := Message.Unused
  else
    KeyChar := Message.CharCode;
  case KeyChar of
    $0020..$FFFF:begin
      InputChar(KeyChar);
    end;
    VK_RETURN: begin
      FColKeeping := -1;
      AAutoIndent := ((FOptions.AutoIndent = 1) and
        not(ssShift in KeyDataToShiftState(Message.KeyData))) or
        ((FOptions.AutoIndent = 0) and
        (ssShift in KeyDataToShiftState(Message.KeyData)));
      if AAutoIndent then begin
        if FStr.SelStart < FStr.SelEnd then
          p := FStr.SelStartPos
        else
          p := FStr.SelEndPos;
        aIndentString := FStr.GetIndentString(p.X, p.Y);
        FStr.SelText := WideString(#10) + aIndentString;
      end else begin
        FStr.SelText := #10;
      end;
    end;
    VK_TAB: begin
      FColKeeping := -1;
      if not FStr.DoTabIndent(ssShift in KeyDataToShiftState(Message.KeyData)) then
        FStr.SelText := #09;
    end;
    VK_BACK: begin
      //Ctrl+HƂłVK_BACKɗB
      //obNXy[XWMCHAR̂Delete͗Ȃ݂B
      FColKeeping := -1;
      FStr.InputBackKey;
    end;
  end;
end;

procedure TNotePad.WMCreate(var Message: TWMCreate);
begin
  inherited;
  RegisterDragDrop(Handle, self as IDropTarget);
end;

procedure TNotePad.WMDestroy(var Message: TWMDestroy);
begin
  RevokeDragDrop(Handle);
  inherited;
end;

procedure TNotePad.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result or DLGC_WANTCHARS or
    DLGC_WANTARROWS or DLGC_WANTTAB or DLGC_WANTALLKEYS;
end;

procedure TNotePad.WMHScroll(var Message: TWMHScroll);
var
  HScrollInc: integer;
  NewPos: integer;
begin
  inherited;
  case Message.ScrollCode of
    SB_LINELEFT: HScrollInc := -1;
    SB_LINERIGHT: HScrollInc := 1;
    SB_PAGELEFT: HScrollInc := Min(-1, -(FVisibleByteCount - 3));
    SB_PAGERIGHT: HScrollInc := Max(1, FVisibleByteCount - 3);
    SB_THUMBPOSITION,
    SB_THUMBTRACK: HScrollInc := Message.Pos - FFirstVisiblePix div FFontWidth;
    SB_LEFT: HScrollInc := - FFirstVisiblePix div FFontWidth;
    SB_RIGHT: HScrollInc := FStr.WrapByte - FVisibleByteCount + 3 - FFirstVisiblePix div FFontWidth;
    else exit;
  end;
  NewPos := Max(0, Min(FFirstVisiblePix div FFontWidth + HScrollInc, FStr.WrapByte - FVisibleByteCount + 3));
  if NewPos <> FFirstVisiblePix div FFontWidth then begin
    FFirstVisiblePix := NewPos * FFontWidth;
    initScroll;
    Invalidate;
  end;
  Message.Result := 0;
end;

procedure TNotePad.WMImeChar(var Message: TWMChar);
begin
  InputChar(Message.CharCode);
  Message.CharCode := 0;
  inherited;
end;

procedure TNotePad.WMImeRequest(var Msg: TMessage);
//󓮓Iĕϊ
//Ql
//http://www10.plala.or.jp/yaoshi/masa/comp/prog/reconv.html
//ŏLParam=nilāA
//͍\̃|C^炵B
//ɍ\̂ƑΏەȂēĂB
//ڂڂAResultɂ͂ȂTCYB
//AIDE̎s͉ɑԂꍇB
//http://www.osronline.com/ddkx/appendix/imeimes_0h2s.htm
//dwStrLen, dwCompStrLen, and dwTargetStrLen are the TCHAR count,
//dwStrOffset, dwCompStrOffset, and dwTargetStrOffset are the byte offset.
//dwSize͍\̂̃TCYƌƂA
//ATOKł͂̒l͖AMS-IMEł͍\+TCHARPʕTCYB
const
  IMR_RECONVERTSTRING = 4;
  IMR_CONFIRMRECONVERTSTRING = 5;
const
  SCS_QUERYRECONVERTSTRING = $00020000;
  SCS_CAP_SETRECONVERTSTRING = $00000004;
var
  i: integer;
  s : WideString;
  ss: packed array of WideChar;
  r: TReconvertString;
  H: HIMC;
  p:TPoint;
  X: integer;
  pr: Pointer;
begin
  inherited;
  if Msg.WParam = IMR_RECONVERTSTRING then begin
    if Msg.LParam = 0 then begin
      if SelectionLength = 0 then
      //SCS_QUERYRECONVERTSTRINGɑΉĂꍇ̂݁A
      //݊̍ĕϊ݂
      if (ImmGetProperty(GetKeyboardLayout(0),IGP_SETCOMPSTR) and SCS_CAP_SETRECONVERTSTRING) <> 0 then begin
        H := ImmGetContext(Handle);
        if H <> 0 then try
          p := FStr.SelEndPos;
          s := FStr.GetLineString(p.X, p.Y, X);
          SetLength(ss, Length(s) + 1);
          for i := 0 to Length(s) - 1 do ss[i] := s[i + 1];
          ss[Length(s)] := #0;
          r.dwSize := 32 + (Length(ss) * 2);//*sizeof(r)*/
          r.dwVersion := 0;
          r.dwStrOffset := 32;//*sizeof(r)*/
          r.dwStrLen := Length(s);
          r.dwCompStrOffset := X * 2;
          r.dwCompStrLen := 0;
          r.dwTargetStrOffset := r.dwCompStrOffset;//ŵsA
          r.dwTargetStrLen := 0;                   //CompƓlȂᓮȂ
          GetMem(pr, 32 + (Length(ss) * 2));
          try
            CopyMemory(pr, @r, 32{SizeOf(r)});
            CopyMemory(Pointer(integer(pr) + 32{SizeOf(r)}), PWChar(ss), Length(ss) * 2);
            ImmSetCompositionStringW(H, SCS_QUERYRECONVERTSTRING, pr, 32 + Length(ss) * 2 , nil, 0 );
            r := PReconvertString(pr)^;
          finally
            FreeMem(pr);
          end;
        finally
          ImmReleaseContext(Handle, H);
        end;
        SelectionStart := FStr.SelEnd - X + integer(r.dwCompStrOffset div 2);
        SelectionLength := r.dwCompStrLen;
      end;
      //ĕϊŎI镶݊ɂȂȂ
      if SelectionLength = 0 then SelectWordFromCaret;
      //MS-IME͉s܂łĕϊāACRLF𑗂Ă̂Ŗ肪N
      if not SelectedWithoutCRLF then exit;
      s := FStr.SelText;
      //J[\ʒuϊʒu擪Ɏčs߁AE獶ɑI悤ɂ
      if SelectionLength > 0 then begin
        i := SelectionLength;
        SelectionStart := SelectionStart + i;
        SelectionLength := - i;
        Update;
      end;
    end;
    s := FStr.SelText;
    SetLength(ss, Length(s) + 1);
    for i := 0 to Length(s) - 1 do ss[i] := s[i + 1];
    ss[Length(s)] := #0;
    Msg.Result := 32{sizeof(r)}+ Length(ss) * 2;
    if Msg.LParam = 0 then begin
    end else begin
      r.dwSize := Msg.Result;//*sizeof(r)*/
      r.dwVersion := 0;
      r.dwStrOffset := 32;//*sizeof(r)*/
      r.dwStrLen := Length(s{s});
      r.dwCompStrOffset := 0;
      r.dwCompStrLen := r.dwStrLen;//Length(s);
      r.dwTargetStrOffset := r.dwCompStrOffset;
      r.dwTargetStrLen := r.dwCompStrLen;
      CopyMemory(pointer(msg.LParam), @r, 32{SizeOf(r)});
      CopyMemory(Pointer(msg.LParam + 32{SizeOf(r)}), PWChar(ss), Length(ss) * 2);
    end;
  end else if Msg.WParam = IMR_CONFIRMRECONVERTSTRING then msg.Result := 0;
end;

procedure TNotePad.WMKeyDown(var Message: TWMKeyDown);
var
  Shift: TShiftState;
  p: TPoint;
  ALine: PLineData;
  NextX, NextY: integer;
  over: boolean;
  m: TMessage;
  i, APrevious, AStart, ALength: integer;
  cj: boolean;
begin
  inherited;
  Message.Result := 0;
  case Message.CharCode of
    VK_DOWN, VK_UP, VK_NEXT, VK_PRIOR:begin
    end;
    else FColKeeping := -1;
  end;
  Shift := KeyDataToShiftState(Message.KeyData);
  case Message.CharCode of
    VK_DELETE: if Shift = [ssShift] then Cut else FStr.InputDelKey;
//    VK_BACK, VK_F16: FStr.InputBackKey;//WMCHARŎ擾
    VK_INSERT: if Shift = [ssShift] then Paste else if Shift = [ssCtrl] then Copy;
    VK_CLEAR, VK_NUMPAD5: if Shift = [ssCtrl] then SelectAll;
    VK_RIGHT: begin
      p := FStr.SelEndPos;
      if ssCtrl in Shift then begin
        i := FStr.SelEnd;
        if i = FStr.TextLength then begin
          //eLXg̍Ō̕łɉEɈړ悤ƂƂA
          //͈͑I𒆂ɃVtgĉEȂIɂȂA
          //VtgĂꍇI𒆂̏ꍇ͏𔲂B
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
        SelectWordSub(p.X, p.Y, APrevious, AStart, ALength, cj);
        if (i = AStart + ALength) and (i < FStr.TextLength) then begin
          inc(ALength);
        end;
        BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
        FStr.BeginUpdate;
        FStr.SelEnd := AStart + ALength;
        if not (ssShift in Shift) then FStr.SelStart := AStart + ALength;
        FStr.RedrawCurrent;
        FStr.EndUpdate;
      end else begin
        ALine := FStr[p.y];
        if p.X < ALine.Length then begin
          inc(p.X);
        end else if p.Y < FStr.Count - 1 then begin
          inc(p.Y);
          p.X := 0;
        end else begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
        BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
        FStr.BeginUpdate;
        FStr.SelEndPos := p;
        if not (ssShift in Shift) then FStr.SelStartPos := p;
        FStr.RedrawCurrent;
        FStr.EndUpdate;
      end;
    end;
    VK_LEFT: begin
      p := FStr.SelEndPos;
      if ssCtrl in Shift then begin
        i := FStr.SelEnd;
        if i = 0 then begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
        SelectWordSub(p.X, p.Y, APrevious, AStart, ALength, cj);
        if i = AStart then AStart := APrevious;
        if (i = AStart) and (i > 0) then begin
          dec(AStart); 
        end;
        BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
        FStr.BeginUpdate;
        FStr.SelEnd := AStart;
        if not (ssShift in Shift) then FStr.SelStart := AStart;
        FStr.RedrawCurrent;
        FStr.EndUpdate;
      end else begin
        if p.X > 0 then begin
          dec(p.X);
        end else if p.Y > 0 then begin
          dec(p.Y);
          ALine := FStr[p.Y];
          p.X := ALine.Length;
        end else begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
        BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
        FStr.BeginUpdate;
        FStr.SelEndPos := p;
        if not (ssShift in Shift) then FStr.SelStartPos := p;
        FStr.RedrawCurrent;
        FStr.EndUpdate;
      end;
    end;
    VK_DOWN: begin
      p := FStr.SelEndPos;
      ALine := FStr[p.Y];
      if p.Y < FStr.Count - 1 then begin
        if p.Y >= FFirstVisibleLine then NextY := p.Y + 1 else
          NextY := Min(FStr.Count - 1, Max(p.Y, FFirstVisibleLine + 1) + 1);
        NextX := FStr.GetCaretX(p.X, p.Y);
        if FColKeeping < NextX then FColKeeping := NextX;
        if FOptions.KeepCol then
          p.X := FStr.FromCaretX(FColKeeping, NextY, over)
        else
          p.X := FStr.FromCaretX(NextX, NextY, over);
        p.Y := NextY;
      end else if (ssShift in Shift) and (p.X < ALine.Length) and (not(ssAlt in Shift)) then begin
        FColKeeping := -1;
        p.X := ALine.Length;
      end else begin
        if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
      end;
      BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
      FStr.BeginUpdate;
      FStr.SelEndPos := p;
      if not (ssShift in Shift) then FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
    VK_UP: begin
      p := FStr.SelEndPos;
      if p.Y > 0 then begin
        NextY := Max(0, Min(p.Y, FFirstVisibleLine + FVisibleLineCount - 3) - 1);
        NextX := FStr.GetCaretX(p.X, p.Y);
        if FColKeeping < NextX then FColKeeping := NextX;
        if FOptions.KeepCol then
          p.X := FStr.FromCaretX(FColKeeping, NextY, over)
        else
          p.X := FStr.FromCaretX(NextX, NextY, over);
        p.Y := NextY;
      end else if (ssShift in Shift) and (p.X > 0) and (not(ssAlt in Shift)) then begin
        FColKeeping := -1;
        p.X := 0;
      end else begin
        if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
      end;
      BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
      FStr.BeginUpdate;
      FStr.SelEndPos := p;
      if not (ssShift in Shift) then FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
    VK_NEXT: begin
      p := FStr.SelEndPos;
      ALine := FStr[p.Y];
      if p.Y < FStr.Count - 1 then begin
        if p.Y >= FFirstVisibleLine then
          NextY := Min(FStr.Count - 1, p.Y + FVisibleLineCount -3)
        else
          NextY := Min(FStr.Count - 1, Max(p.Y, FFirstVisibleLine + 2) + FVisibleLineCount -3);
        NextX := FStr.GetCaretX(p.X, p.Y);
        if FColKeeping < NextX then FColKeeping := NextX;
        if FOptions.KeepCol then
          p.X := FStr.FromCaretX(FColKeeping, NextY, over)
        else
          p.X := FStr.FromCaretX(NextX, NextY, over);
        p.Y := NextY;
      end else if (ssShift in Shift) and (p.X < ALine.Length) then begin
        FColKeeping := -1;
        p.X := ALine.Length;
      end else begin
        if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
      end;
      m.Msg := WM_VSCROLL;
      m.WParam := SB_PAGEDOWN;
      WndProc(m);
      BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
      FStr.BeginUpdate;
      FStr.SelEndPos := p;
      if not (ssShift in Shift) then FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
    VK_PRIOR: begin
      p := FStr.SelEndPos;
      if p.Y > 0 then begin
        NextY := Max(0, Min(p.Y, FFirstVisibleLine + FVisibleLineCount - 3) - FVisibleLineCount + 3);
        NextX := FStr.GetCaretX(p.X, p.Y);
        if FColKeeping < NextX then FColKeeping := NextX;
        if FOptions.KeepCol then
          p.X := FStr.FromCaretX(FColKeeping, NextY, over)
        else
          p.X := FStr.FromCaretX(NextX, NextY, over);
        p.Y := NextY;
      end else if (ssShift in Shift) and (p.X > 0) then begin
        FColKeeping := -1;
        p.X := 0;
      end else begin
        if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
      end;
      m.Msg := WM_VSCROLL;
      m.WParam := SB_PAGEUP;
      WndProc(m);
      BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
      FStr.BeginUpdate;
      FStr.SelEndPos := p;
      if not (ssShift in Shift) then FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
    VK_END: begin
      p := FStr.SelEndPos;
      if ssCtrl in Shift then begin
        ALine := FStr[FStr.Count - 1];
        if (p.Y < FStr.Count - 1) or (p.X < ALine.Length) then begin
          p.Y := FStr.Count - 1;
          p.X := ALine.Length;
        end else begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
      end else begin
        ALine := FStr[p.Y];
        if p.X < ALine.Length then begin
          p.X := ALine.Length;
        end else begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
      end;
      BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
      FStr.BeginUpdate;
      FStr.SelEndPos := p;
      if not (ssShift in Shift) then FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
    VK_HOME: begin
      p := FStr.SelEndPos;
      if ssCtrl in Shift then begin
        if (p.Y > 0) or (p.X > 0) then begin
          p.Y := 0;
          p.X := 0;
        end else begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
      end else begin
        if p.X > 0 then begin
          p.X := 0;
        end else begin
          if (ssShift in Shift) or ((FStr.SelStartPos.X = p.X) and (FStr.SelStartPos.Y = p.Y)) then exit;
        end;
      end;
      BoxSelectMode := (ssShift in Shift) and (ssAlt in Shift);
      FStr.BeginUpdate;
      FStr.SelEndPos := p;
      if not (ssShift in Shift) then FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
  end;
end;

procedure TNotePad.WMKillFocus(var Msg: TWMKillFocus);
begin
  inherited;
  CaretHide;
  //DestroyCaret;
  Msg.Result := 0;
end;

procedure TNotePad.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
  p1: TPoint;
  over: boolean;
begin
  inherited;
  //_uNbN͕ꏊłNȂ
  if Message.XPos < FLeftBarWidth + 3 then exit;
  p1 := PointToPos(Message.XPos, Message.YPos, false);//̐^񒆂͑O̕
  if p1.Y > FStr.Count - 1 then exit;
  p1.X := FStr.FromCaretX(p1.X, p1.Y, over);
  if over then exit;
  //FDoubleClickedƂtO͓ނ̈ӖŎg
  if not DoubleClickDown(p1.X) then begin
    //[hZNgsꂽƂAGAȂhbOɓ
    //GAOꍇtOĂȂƂxNbNĂ܂
    FDoubleClicked := true;
    WMLButtonDown(Message);
    //tOĖ߂Ă
    //upŃGAȂ߂xtOZbg
    FDoubleClicked := true;
  end;
end;

procedure TNotePad.WMLButtonDown(var Message: TWMLButtonDown);
var
  p, p1: TPoint;
  X, X2: integer;
  over: boolean;
  aSelArea: TPoint;
begin
  //MouseEnterɓ鏈̂inherited̑OMove
  //Moveł̓_uNbNLv`[肪̂ŃtB^
  if (not FDoubleClicked) and (not MouseCapture) then WMMouseMove(Message);
  inherited; // Windows.SetFocus cf DefaultHandler
  TrackMouseEventSub;
  FColKeeping := -1;
  FDDReady := false;
  FLineSelect := -1;
  FRulerDrag := false;
  //ȂƈꕶIɊԂɓȂ
  p := PointToPos(Message.XPos, Message.YPos, true);//̐^񒆂͎̕
  p1 := PointToPos(Message.XPos, Message.YPos, false);//̐^񒆂͑O̕
  if Message.YPos < (FShowRuler * FLineHeight) then begin
    FDoubleClicked := false;
    FRulerDrag := true;
    if FOptions.RulerEnabled and Assigned(FRulerDown) then FRulerDown(Self, P.X div FFontWidth);
    exit;
  end;
  if Message.XPos < FLeftBarWidth + 3 then begin
    BoxSelectMode := false;
    FDoubleClicked := false;
    if p.Y < FStr.Count - 1 then begin
      FLineSelect := p.Y;
      p.X := 0;
      fstr.BeginUpdate;
      FStr.SelEndPos := p;
      inc(p.Y);
      FStr.SelStartPos := p;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
      exit;
    end;
    p.Y := FStr.Count - 1;
    FLineSelect := p.Y;
    p.X := FStr[p.Y].Length;
    FStr.BeginUpdate;
    FStr.SelStartPos := p;
    p.X := 0;
    FStr.SelEndPos := p;
    FStr.RedrawCurrent;
    fstr.EndUpdate;
    exit;
  end;
  p.Y := Min(p.Y, FStr.Count - 1);
//  p.X := FStr.FromCaretX(p.X, p.Y, over);
//  p1.X := FStr.FromCaretX(p1.X, p.Y, over);
  //Sp̉ENbNւ̑Ή
  X := FStr.FromCaretX(p1.X, p.Y, over);
  X2 := FStr.FromCaretX(p.X, p.Y, over);
  {if X = FStr.FromCaretX(p1.X - 1, p.Y, over) then begin
    p.X := FStr.FromCaretX(p1.X + 1, p.Y, over);
    p1.X := X;
  end else begin }
    p.X := X2;
    p1.X := X;
  {end;}

  aSelArea := FStr.GetSelectedArea(p.Y);
  //p̓NbNŃJ[\ړꏊAp1̓hbOŊ荞ޏꏊ
  if (p.X > aSelArea.X) and (p1.X < aSelArea.Y) and (not over) then begin
    FDoubleClicked := false;
    FDDReady := true;
    exit;
  end;
  if FDoubleClicked then begin
    FDoubleClicked := false;
    exit;
  end;
  //ɂMessage.Keysɂ́AssAltĂȂBȂH
  BoxSelectMode := GetKeyState(VK_MENU) < 0;
  if ssShift in KeyDataToShiftState(Message.Keys) then begin
    FStr.BeginUpdate;
    FStr.SelEndPos := p;
    FStr.RedrawCurrent;
    FStr.EndUpdate;
    exit;
  end;
  FStr.BeginUpdate;
  FStr.SelStartPos := p;
  FStr.SelEndToSelStart;
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.WMLButtonUp(var Message: TWMLButtonUp);
var
  p, p1: TPoint;
  X, X2: integer;
  over: boolean;
  aSelArea: TPoint;
begin
//Todo: UpDownȂŋN邱ƂȂǂȂBBB̂H
//      悻̑MouseDownA䂪ƂMouseUp̎̓𒲂ׂ
//      Ԃ悻ɃLv`Ă邩炱ւ͗Ȃ
  inherited;
  FColKeeping := -1;
  FDDReady := false;
  //ȂƕςȂƂɓ
  p := PointToPos(Message.XPos, Message.YPos, true);//̐^񒆂͎̕
  p1 := PointToPos(Message.XPos, Message.YPos, false);//̐^񒆂͑O̕
//I͈͓downƂ̓hbOȂ̂
//upIɂȂƂ܂B
//DoubleClickupƂ͖Ă
  if FDoubleClicked then FDoubleClicked := false else
  if FRulerDrag then begin
    if FOptions.RulerEnabled and Assigned(FRulerUp) then FRulerUp(Self, p.X div FFontWidth);
    FRulerDrag := false;
  end else if FLineSelect = -1 then begin
    p.Y := Min(p.Y, FStr.Count - 1);
    X := FStr.FromCaretX(p1.X, p.Y, over);
    X2 := FStr.FromCaretX(p.X, p.Y, over);
    {if X = FStr.FromCaretX(p1.X - 1, p.Y, over) then begin
      p.X := FStr.FromCaretX(p1.X + 1, p.Y, over);
      p1.X := X;
    end else begin}
      p.X := X2;
      p1.X := X;
    {end;}
    aSelArea := FStr.GetSelectedArea(p.Y);
    //ɔɎĝ͍ŏIʒuiGj
    //p̓NbNŃJ[\ړꏊAp1̓hbOŊ荞ޏꏊ
    if (p.X > aSelArea.X) and (p.X < aSelArea.Y) and (not over) then begin
    {p.Y := Min(p.Y, FStr.Count - 1);
    p.X := FStr.FromCaretX(Max(0, p.X), p.Y, over);

    aSelArea := FStr.GetSelectedArea(p.Y);
    if (p.X > aSelArea.X) and (p.X < aSelArea.Y) then begin}
      FStr.BeginUpdate;
      FStr.SelStartPos := p;
      FStr.SelEndToSelStart;
      FStr.RedrawCurrent;
      FStr.EndUpdate;
    end;
  end;
end;

procedure TNotePad.WMMouseHover(var Message: TWMMouse);
var
  info1: TScrollInfo;
  b: boolean;
begin
  inherited;
  if MouseCapture and Focused then begin
    //{glbN͕`敔łA`悳MouseMoveŏɌĂ
    b := false;
    info1.cbSize := 28;//SizeOf(TScrollInfo);
    info1.fMask := 23;//SIF_ALL
    GetScrollInfo(Handle, SB_VERT, info1);
    if (info1.nPos > 0) and (Message.YPos < 0) then b := true;
    if (info1.nPos + integer(info1.nPage) < info1.nMax) and (Message.YPos > ClientHeight)
      then b := true;
    GetScrollInfo(Handle, SB_HORZ, info1);
    if (info1.nPos > 0) and (Message.XPos < 0) then b := true;
    if (info1.nPos + integer(info1.nPage) < info1.nMax) and (Message.XPos > ClientWidth)
      then b := true;
    if b then WMMouseMove(Message);
    //TrackňWM_HOVERȂ̂ŁAMouseMovesvłTrackAB
    //Lv`I͂ɗȂȂ̂TrackAIB
    TrackMouseEventSub;
  end;
end;

procedure TNotePad.WMMouseMove(var Message: TWMMouseMove);
var
  p, p1, p2, aSelArea: TPoint;
  X, X2: integer;
  over: boolean;
  DataObject: IDataObject;
  DropSource: IDropSource;
  AEffect: integer;
begin
  //}EXJ[\͔Lv`[̕ύX΂悢炵B
  //MouseDown̎̃J[\hbOpdlB
  inherited;
  FDoubleClicked := false;
  //ȂƈꕶIɊԂɓȂ
  p := PointToPos(Message.XPos, Message.YPos, true);//̐^񒆂͎̕
  p1 := PointToPos(Message.XPos, Message.YPos, false);//̐^񒆂͑O̕
  if Message.YPos < (FShowRuler * FLineHeight) then begin
    if MouseCapture then begin
    end else begin
      if FOptions.RulerEnabled then
        Cursor := crHSplit
      else
        Cursor := crDefault;
      exit;
    end;
  end;
  X2 := 0;
  if Message.XPos < FLeftBarWidth + 3 then begin
    Cursor := crArrow;
  end else begin
    p2.Y := Min(p.Y, FStr.Count - 1);
    //p2.X := FStr.FromCaretX(p.X, p2.Y, over);
    //p1.X := FStr.FromCaretX(p1.X, p2.Y, over);
    //Sp̉ENbNւ̑Ή
    X := FStr.FromCaretX(p1.X, p2.Y, over);
    X2 := FStr.FromCaretX(p.X, p2.Y, over);
    {if X = FStr.FromCaretX(p1.X - 1, p2.Y, over) then begin
      p2.X := FStr.FromCaretX(p1.X + 1, p2.Y, over);
      p1.X := X;
    end else begin }
      p2.X := X2;
      p1.X := X;
    {end;    }
    X2 := p2.X;
    aSelArea := FStr.GetSelectedArea(p2.Y);
    //p̓NbNŃJ[\ړꏊAp1̓hbOŊ荞ޏꏊ
    if (aSelArea.X < p2.X) and (p1.X < aSelArea.Y) and (not over) then begin
      Cursor := crArrow;
    end else begin
      Cursor := crIBeam;
    end;
  end;
  if not MouseCapture then exit;
  if FRulerDrag then begin
    if FOptions.RulerEnabled and Assigned(FRulerMove) then FRulerMove(Self, p.X div FFontWidth);
    exit;
  end;
  if FLineSelect > -1 then begin
    FStr.BeginUpdate;
    if p.Y > FLineSelect then begin
      if p.Y < FStr.Count - 1 then begin
        inc(p.Y);
        p.X := 0;
        FStr.SelEndPos := p;
        p.Y := FLineSelect;
      end else begin
        p.Y := FStr.Count - 1;
        p.X := FStr[p.Y].Length;
        FStr.SelEndPos := p;
        p.Y := FLineSelect;
        p.X := 0;
      end;
    end else begin
      p.X := 0;
      FStr.SelEndPos := p;
      if FLineSelect < FStr.Count - 1 then begin
        p.Y := FLineSelect + 1;
      end else begin
        p.Y := FLineSelect;
        p.X := FStr[FLineSelect].Length;
      end;
    end;
    FStr.SelStartPos := p;
    FStr.RedrawCurrent;
    FStr.EndUpdate;
    exit;
  end;
  if FDDReady and (FStr.SelLength <> 0) then begin
    FDDReady := false;
    FDragCancel := false;
    FDDString := TDDString.Create;
    FDDString.Text := FStr.SelTextIf;
    DataObject := FDDString;
    FDDSource := TDDSource.Create;
    DropSource := FDDSource;
    try
      DoDragDrop(DataObject, DropSource, DROPEFFECT_COPY or DROPEFFECT_MOVE, AEffect);
      FDraging := false;
      CaretHide;
      initCaretPosition;
      exit;
    finally
      //FDDString.Free;
      FDDString := nil;
      FDDSource := nil;
      DataObject := nil;
      DropSource := nil;
    end;
  end;
  p.Y := Min(p.Y, FStr.Count - 1);
  //p.X := FStr.FromCaretX(Max(0, p.X), p.Y, over);
  p.X := X2;
  FStr.BeginUpdate;
  FStr.SelEndPos := p;
  FStr.RedrawCurrent;
  FStr.EndUpdate;
end;

procedure TNotePad.WMMouseWheel(var Message: TWMMouseWheel);
var
  i, j, sl: SHORT;
begin
  Inherited;
  //Delta͗ݐςĂA120܂邲Ƃɏ
  inc(FMouseWheelDelta, Message.WheelDelta);
  SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, @sl, 0);
  if FMouseWheelDelta = 0 then begin
  end else if FMouseWheelDelta >= 120 then begin
    j := FMouseWheelDelta div 120;
    FMouseWheelDelta := 0;
    if sl = -1 then begin
      for i := 1 to j do
        SendMessageW(Handle, WM_VSCROLL, SB_PAGEUP, 0);
    end else begin
      for i := 1 to j * sl do
        SendMessageW(Handle, WM_VSCROLL, SB_LINEUP, 0)
    end;
  end else if FMouseWheelDelta <= -120 then begin
    j := FMouseWheelDelta div -120;
    FMouseWheelDelta := 0;
    if sl = -1 then begin
      for i := 1 to j do
        SendMessageW(Handle, WM_VSCROLL, SB_PAGEDOWN, 0);
    end else begin
      for i := 1 to j * sl do
        SendMessageW(Handle, WM_VSCROLL, SB_LINEDOWN, 0)
    end;
  end else begin
  end;
end;

procedure TNotePad.WMPaint(var Message: TWMPaint);
begin
  if CreateBitmap then exit;
  if NeedWrap then begin
    repeat until not NeedWrap;
    exit;
  end;
  FRgn := CreateRectRgn(0,0,0,0);
  try
    if GetUpdateRgn(Handle, FRgn, true) <> 0 then inherited;
  finally
    DeleteObject(FRgn);
  end;
end;

procedure TNotePad.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  CaretHide;
  CaretReset;
  Msg.Result := 0;
end;

procedure TNotePad.WMSize(var Msg: TWMSize);
begin
  inherited;
  Msg.Result := 0;
  if FLineHeight = 0 then exit;
  FColKeeping := -1;
  initScroll;
  Invalidate;
end;

procedure TNotePad.WMVScroll(var Message: TWMVScroll);
var
  VScrollInc: integer;
  NewPos: integer;
  info1: TScrollInfo;
begin
  inherited;
  case Message.ScrollCode of
    SB_LINEUP: VScrollInc := -1;
    SB_LINEDOWN: VScrollInc := 1;
    SB_PAGEUP: VScrollInc := Min(-1, -(FVisibleLineCount - 3));
    SB_PAGEDOWN: VScrollInc := Max(1, FVisibleLineCount - 3);
    //65536sڈȏɃXN[邽߂ɁA32rbgŋ߂
    SB_THUMBPOSITION,
    SB_THUMBTRACK:begin
      info1.cbSize := 28;//SizeOf(TScrollInfo);
      info1.fMask := 23;//SIF_ALL
      GetScrollInfo(Handle, SB_VERT, info1);
      VScrollInc := info1.nTrackPos - FFirstVisibleLine;
    end;
    SB_TOP: VScrollInc := - FFirstVisibleLine;
    SB_BOTTOM: VScrollInc := FStr.Count - FVisibleLineCount + 3 - FFirstVisibleLine;
    else exit;
  end;
  NewPos := Max(0, Min(FFirstVisibleLine + VScrollInc, FStr.Count - FVisibleLineCount + 3));
  if NewPos <> FFirstVisibleLine then begin
    FFirstVisibleLine := NewPos;
    initScroll;
    Invalidate;
  end;
  Message.Result := 0;
end;

procedure TNotePad.WndProc(var Message: TMessage);
const
  EM_CANPASTE                         = WM_USER + 50;
  EM_REDO                             = WM_USER + 84;
  EM_CANREDO                          = WM_USER + 85;

begin
  case Message.Msg of
    WM_ERASEBKGND: begin
      Message.Result := 0;
    end;
    WM_IME_STARTCOMPOSITION: begin
      inherited;
      FNeedInitImeFont := true;
      FStr.BeginUpdate;
      FStr.EndUpdate;
    end;
    EM_CANPASTE: begin
      Message.Result := byte(self.CanPaste);
    end;
    EM_CANREDO: begin
      Message.Result := byte(self.GetCanRedo);
    end;
    EM_CANUNDO: begin
      Message.Result := byte(self.GetCanUndo);
    end;
    EM_EMPTYUNDOBUFFER: begin
      inherited;
      self.ClearUndo;
    end;
    EM_GETFIRSTVISIBLELINE: begin
      Message.Result := FFirstVisibleLine;
    end;
    EM_REDO: begin
      if Self.CanRedo then begin
        Redo;
        Message.Result := -1;
      end else Message.Result := 0;
    end;
    EM_UNDO: begin
      if Self.CanUndo then begin
        Undo;
        Message.Result := -1;
      end else Message.Result := 0;
    end;

{
//TEditorRichEdit݊̃bZ[WMi߂ǂj
EMCharFromPos,
,
EMGetLine,
EMGetLineCount,
EMGetModify,
EMGetSel,
EMLineFromChar,
EMLineIndex,
EMLineLength,
EMPosFromChar,
EMReplaceSel,
EMScrollCaret,
EMSetModiry,
EMSetReadOnly,
,
EMSetSel,
WMChar,
WMClear,
WMCopy,
WMCut,
WMPaste,
WMGetDlgCode,
WMGetText,
WMGetTextLength,
WMHScroll,
WMImeComposition,
WMImeNotify,
WMKeyDown,
WMKeyUp,
WMKillFocus,
WMLButtonDown,
WMLButtonUp,
WMMouseMove,
WMSetFocus,
WMSetText,
WMSize,
}
    else inherited;
  end;
end;

initialization
  FInitialized := false;
  //YƁARegisterDragDropɎs܂B
  OleInitialize(nil);
finalization
  if Assigned(FModeDefault) then FModeDefault.Free;
  if Assigned(FModeHtml) then FModeHtml.Free;
  if Assigned(FModeCSharp) then FModeCSharp.Free;
  if Assigned(FModeDelphi) then FModeDelphi.Free;
  if Assigned(FModeJScript) then FModeJScript.Free;
  if Assigned(FModeVBScript) then FModeVBScript.Free;
  if Assigned(FModePython) then FModePython.Free;
  if Assigned(FModeCpp) then FModeCpp.Free;
  if Assigned(FModeJava) then FModeJava.Free;
  if Assigned(FModePerl) then FModePerl.Free;
  if Assigned(FModeRuby) then FModeRuby.Free;
  if Assigned(FModeVisualBasic) then FModeVisualBasic.Free;
  if Assigned(FModeCss) then FModeCss.Free;
  if Assigned(FModeBat) then FModeBat.Free;
  if Assigned(FBGWindow) then FBGWindow.Free;
  if Assigned(FTileBitmap) then FTileBitmap.Free;
  OleUninitialize;
end.
