unit PointerList;

interface

type
  TPointerList = class
  private
    FPointers : array of Pointer;
    FLength : integer;
    function GetBufferSize : integer;
    function Get(Index: Integer): Pointer;
    procedure Put(Index: Integer; const p: Pointer);
  public
    constructor Create(size : integer);
    destructor Destroy;override;
    procedure Add(const p: Pointer);
    procedure Insert(index: integer; const p: Pointer);
    procedure Remove(index, len: integer);
    procedure Clear;
    property Count: integer read FLength;
    property Capacity: integer read GetBufferSize;
    property Items[Index: Integer]: Pointer read Get write Put; default;
  end;

implementation

{ TPointerList }

procedure TPointerList.Add(const p: Pointer);
var
  i: integer;
//  Lens: integer;
begin
//  Lens := 1;
  i := Length(FPointers);
  while FLength + 1 + 1 > i do inc(i, i);
  if i > Length(FPointers) then SetLength(FPointers, i);
  FPointers[FLength] := p;
  inc(FLength{, Lens});
  FPointers[FLength] := nil;
end;

procedure TPointerList.Clear;
begin
  FLength := 0;
  FPointers[0] := nil;
end;

constructor TPointerList.Create(size: integer);
begin
  if size < 2 then size := 2;
  SetLength(FPointers, size);
  FPointers[0] := nil;
  FLength := 0;
end;

destructor TPointerList.Destroy;
begin
  SetLength(FPointers, 0);
  inherited;
end;

function TPointerList.Get(Index: Integer): Pointer;
begin
  result := FPointers[Index];
end;

function TPointerList.GetBufferSize: integer;
begin
  result := Length(FPointers);
end;

procedure TPointerList.Insert(index: integer; const p: Pointer);
var
  i : integer;
//  Lens :integer;
  ps, pd: Cardinal;
begin
  //Lens := 1;
  i := Length(FPointers);
  while FLength + 1 + 1 > i do inc(i, i);
  if i > Length(FPointers) then SetLength(FPointers, i);
  ps := Cardinal(@(FPointers[0])) + Cardinal(index) * 4;
  pd := ps + 4;
  Move(Pointer(ps)^, Pointer(pd)^, (FLength - index + 1) * 4);
  FPointers[index] := p;
  inc(FLength{, Lens});
end;

procedure TPointerList.Put(Index: Integer; const p: Pointer);
begin
  FPointers[Index] := p;
end;

procedure TPointerList.Remove(index, len: integer);
var
  ps, pd: Cardinal;
begin
  pd := Cardinal(@(FPointers[0])) + Cardinal(index) * 4;
  ps := pd + Cardinal(len) * 4;
  Move(Pointer(ps)^, Pointer(pd)^, (FLength  - len - index + 1) * 4);
  dec(FLength, len);
end;

end.
