unit DM;

interface

uses
  Windows,messages, SysUtils, Classes, ShlObj, ActiveX, ComObj,
  Graphics, ComCtrls, NotePadView, wNotePad, TextFileEncoding, HKeyMacro,
  TBXStatusBars, TntForms, wsMemIni;


type
  IManaMain = interface
    function MMGetPageControl: TPageControl;
    function MMCreateMDIChild(const PathName: widestring;code:TFileCode): TwNotePad;
    function MMAcceptFile(const PathName: WideString; code: TFileCode): TwNotePad;
    function MMGetHandle: HWND;
    function MMGetKeyMacro: TKeyMacro;
    function MMGetSubEditor: TwNotePad;
    procedure MMNotePadViewChanged(Sender: TObject);
    procedure MMLogAddString1(const PathName: WideString);
    procedure MMAppWriteHistINI;
    procedure MMNotePadFilesDroped(Sender: TObject; Files: array of WideString);
    procedure MMSetCaption(const Value: WideString);
    function MMGetStatusBar: TTBXStatusBar;
    function MMTabExists(const PathName: WideString): TTntForm;
    procedure MMNotePadRulerDown(Sender: TObject; Col: Integer);
    procedure MMNotePadRulerMove(Sender: TObject; Col: Integer);
    procedure MMNotePadRulerUp(Sender: TObject; Col: Integer);
    function MMGetActiveEditor: TwNotePad;
    procedure MMRestoreOption;
  end;


//function GetMenuFont: HFONT;
//function GetTitleFont: HFONT;
//function GetMessageFont: HFONT;
function HasClpBrd(hWindow:HWND): WideString; 
function MenuShort(const PathName: WideString): WideString; 
function TabShort(const PathName: WideString): WideString;
function ClipShort(const Value: WideString; Count:integer): WideString; 
function CmdIsLine(const value: WideString; var Line: integer): boolean; 
function IniFilePathNameA(var Value: WideString): boolean;
function LCMapStringSub(Value:PWChar; len: integer; Flag: Cardinal): WideString;
function HanKanaToZenKana(const Value: WideString): WideString;
function ZenKanaToHanKana(const Value: WideString): WideString;
function KataKanaToHiraKana(const Value: WideString): WideString;
function HiraKanaToKataKana(const Value: WideString): WideString;
function GetFileVersion: wideString;
function LowerCaseW(const Value: WideString) : WideString;
function SelectedTextToLine(const Value: WideString): WideString;
function FileExistsOrCreate(var Value: WideString; hWindow:HWND):boolean;
function ExtractStringFromLinkFile(Target: WideString ): WideString;
procedure UpdateHistIni;

const
  NotitleCaption: WideString ='VKeLXg';
  NotitleHint: widestring ='ۑ̐VKeLXg';

type
  TReplaceRange = (rrStartTop, rrStartHere, rrSelectedOnly);

var
  ManaMain: IManaMain;
  FNotePadOptions: TNotePadOptions;
  FWordWrapCount: integer;
  FWordWrap: TWordWrap;
  FEditorFont: TFont;
  FSearchText, FReplaceText, FHighlightText: WideString;
  FSearchDown, FSearchIgCase, FSearchConfirm,
  FSearchWithRegEx: boolean;
  FSearchLoop: boolean;
  FReplaceRange : TReplaceRange;
  DebuggerPresent: boolean;
  FAddExtTXT: boolean;
  FLastSavedFile: WideString;
  FHistIni: TMemIniFileW;

procedure UnitInitialize;

implementation

uses wsFileOp, StringBuffer;

function LowerCaseW(const Value: WideString) : WideString;
begin
  Result := Value;
  CharLowerBuffW(PWChar(Result), Length(Result));
end;

function HasClpBrd(hWindow:HWND): WideString;
var
  i:integer;
  Data: THandle;
begin
  Result := '';
  if OpenClipboard(hWindow) then begin
    try
      i := 0;
      repeat
        i := EnumClipboardFormats(i);
        if i = CF_TEXT then begin
          Data := GetClipboardData(CF_TEXT);
          if Data <> 0 then begin
            Result := PAnsiChar(GlobalLock(Data));
            GlobalUnlock(Data);
          end else Result := '';
          Break;
        end;
      until i = 0;
      i := 0;
      repeat
        i := EnumClipboardFormats(i);
        if i = CF_UNICODETEXT then begin
          Data := GetClipboardData(CF_UNICODETEXT);
          if Data <> 0 then begin
            Result := PWideChar(GlobalLock(Data));
            GlobalUnlock(Data);
          end else Result := '';
          Break;
        end;
      until i = 0;
    finally
      CloseClipboard;
    end;
  end;
end;

function GetMenuFont: HFONT;
var
  NonClientMetrics : TNonClientMetrics;
begin
  NonClientMetrics.cbSize := Sizeof( NonClientMetrics );
  if SystemParametersInfo( SPI_GETNONCLIENTMETRICS, 0,@NonClientMetrics, 0 ) then
    Result := CreateFontIndirect( NonClientMetrics.lfMenuFont )
  else
    Result := GetStockObject( SYSTEM_FONT );
end;

function GetTitleFont: HFONT;
var
  NonClientMetrics : TNonClientMetrics;
begin
  NonClientMetrics.cbSize := Sizeof( NonClientMetrics );
  if SystemParametersInfo( SPI_GETNONCLIENTMETRICS, 0,@NonClientMetrics, 0 ) then
    Result := CreateFontIndirect( NonClientMetrics.lfCaptionFont )
  else
    Result := GetStockObject( SYSTEM_FONT );
end;

function GetMessageFont: HFONT;
var
  NonClientMetrics : TNonClientMetrics;
begin
  NonClientMetrics.cbSize := Sizeof( NonClientMetrics );
  if SystemParametersInfo( SPI_GETNONCLIENTMETRICS, 0,@NonClientMetrics, 0 ) then
    Result := CreateFontIndirect( NonClientMetrics.lfMessageFont )
  else
    Result := GetStockObject( SYSTEM_FONT );
end;

function DupAmp(const s: WideString): WideString;
var
  i, j, Len: integer;
begin
  Result:='';
  Len := Length(s);
  for i := 1 to Len do if s[i] = WideChar('&') then inc(Len);
  SetLength(Result, Len);
  j := 1;
  for i := 1 to Length(s) do begin
    if s[i] = WideChar('&') then begin
      Result[j] := WideChar('&');
      inc(j);
    end;
    Result[j] := s[i];
    inc(j);
  end;
end;

function MenuShort(const PathName: WideString): WideString;
var
  s: AnsiString;
  i: integer;
  ws1, ws2: WideString;
begin
  s := PathName;
  if Length(s) <= 42 then begin
    Result := DupAmp(PathName);
    exit;
  end;
  i := 8;
  repeat
    inc(i);
    s := copy(PathName, 1, i);
  until Length(s) >= 19;
  ws1 := copy(PathName, 1, i) + '.';
  if Length(s) = 19 then ws1 := ws1 + '.';
  i := 8;
  repeat
    inc(i);
    s := copy(PathName, Length(Pathname) - i + 1,i);
  until Length(s) >= 19;
  ws2 := '.' + copy(PathName, Length(Pathname) - i + 1, i);
  if Length(s) = 19 then ws2 := '.' + ws2;
  Result := DupAmp(ws1 + ws2);
end;

function TabShort(const PathName: WideString): Widestring;
var
  i, j: integer;
begin
  result := ChangeFileExtW(ExtractFileNameW(PathName), '');
  j := 0;
  for i := 1 to Length(result) do begin
    inc(j, Length(AnsiString(result[i])));
    if j >= 9 then begin
      if i = Length(Result) then break;
      if (i = Length(Result) - 1) and (Length(AnsiString(result[i + 1])) = 1) then break;
      SetLength(Result, i);
      Result := Result + '~';
      break;
    end;
  end;
  Result := DupAmp(Result + ExtractFileExtW(PathName));
end;

function ClipShort(const Value: WideString; Count:integer): WideString;
var
  i, j: integer;
begin
  Result := value;
  j := 0;
  for i := 1 to Length(result) do begin
    inc(j, Length(AnsiString(Result[i])));
    if j >= Count then begin
      if i = Length(Result) then break;
      if (i = Length(Result) - 1) and (Length(AnsiString(Result[i])) = 1) then break;
      Result := Copy(Result, 1, i) + WideString('~');
      break;
    end;
  end;
  Result := DupAmp(Result);
  for i := 1 to Length(Result) do begin
    case Result[i] of
      #9: Result[i] := '>';
      #10,#13: Result[i] := '<';
    end;
  end;
end;

function CmdIsLine(const value: WideString; var Line: integer): boolean;
var
  s: WideString;
  E: integer;
begin
  Result := false;
  if Length(value) < 3 then exit;
  if (value[1] = WideChar('/')) and (value[2] = WideChar('j')) then begin
    s:=copy(value, 3, length(value) - 2);
    if s = WideString('e') then Line := -1 else begin
      try
        Val(s, Line, E);
        Result := E = 0;
      except
      end;
    end;
  end;
end;

function IniFilePathNameA(var Value: WideString): boolean;
begin
  Result:=false;
  //{݂̂ȂhCuɂ͏܂Ȃ
  if not FileExists2(ParamStrW(0)) then exit;
  Value := ChangeFileExtW(ParamStrW(0), '.ini');
  if Length(Value) > 0 then Result := true;
end;

function LCMapStringSub(Value:PWChar; len: integer; Flag: Cardinal): WideString;
var
  l: LCID;
  i: integer;
begin
  l := GetUserDefaultLCID;
  i := LCMapStringW(l, Flag, Value, len, nil, 0);
  SetLength(Result, i);
  i := LCMapStringW(l, Flag, Value, len, PWChar(Result), i);
  SetLength(Result, i);
  //ł͂
  //  Result := LCMapStringSub(PWChar(Value), Length(Value), LCMAP_FULLWIDTH);
end;

const
  wsHanKana: array[#$30A0..#$30FF] of WideString = (
				#0 ,'','','','','','','','','','','','','','','',
				'','','','','','','','','','','','','','','','',
				'','','','','','','','','','','','','','','','',
				'','','','','','','','','','','','','','','','',
				'','','','','','','','','','','','','','',#0 ,'',
				#0 ,#0 ,'','',#0 ,#0 ,#0 ,#0 ,#0 ,#0 ,#0 ,#0 ,'',#0 , #0, #0);
  wsZenKana: array[#$FF60..#$FF9F] of WideString = (
				#0 , 'B','u','v','A','E','','@','B','D','F','H','','','','b',
				'[','A','C','E','G','I','J','L','N','P','R','T','V','X','Z','\',
				'^','`','c','e','g','i','j','k','l','m','n','q','t','w','z','}',
				'~','','','','','','','','','','','','','','J','K');
var
  wcHanKana: array[#$30A0..#$30FF] of WideChar;
  wcZenKana: array[#$FF60..#$FF9F] of WideChar;
procedure initKana;
var
  i: WideChar;
begin
  for i := #$30A0 to #$30FF do wcHanKana[i] := wsHanKana[i][1];
  for i := #$FF60 to #$FF9F do wcZenKana[i] := wsZenKana[i][1];
end;

function HanKanaToZenKana(const Value: WideString): WideString;
var
  sb: TStringBuffer;
  i: integer;
  c: WideChar;
begin
  sb := TStringBuffer.Create(Length(Value));
  sb.Append(Value);
  try
    i := 0;
    while i < sb.StringLength do begin
      c := sb[i];
      case c of
        #$FF61..#$FF72,
        #$FF74..#$FF75,
        #$FF85..#$FF89,
        #$FF8F..#$FF9F: begin
          sb[i] := wcZenKana[C];
        end;
        #$FF73: begin
          if (i < sb.StringLength - 1) and (sb[i + 1] = #$FF9E) then begin
            sb[i] := #$30F4;
            sb.Remove(i + 1, 1);
          end else sb[i] := wcZenKana[C];
        end;
        #$FF76..#$FF84: begin
          if (i < sb.StringLength - 1) and (sb[i + 1] = #$FF9E) then begin
            sb[i] := WideChar(integer(wcZenKana[C]) + 1);
            sb.Remove(i + 1, 1);
          end else sb[i] := wcZenKana[C];
        end;
        #$FF8A..#$FF8E: begin
          if i < sb.StringLength - 1 then begin
            if sb[i + 1] = #$FF9E then begin
              sb[i] := WideChar(integer(wcZenKana[C]) + 1);
              sb.Remove(i + 1, 1);
            end else if sb[i + 1] = #$FF9F then begin
              sb[i] := WideChar(integer(wcZenKana[C]) + 2);
              sb.Remove(i + 1, 1);
            end else sb[i] := wcZenKana[C];
          end else sb[i] := wcZenKana[C];
        end;
      end;
      inc(i);
    end;
    Result := sb.ToString;
  finally
    sb.Free;
  end;
end;

function ZenKanaToHanKana(const Value: WideString): WideString;
var
  sb: TStringBuffer;
  i: integer;
  c: WideChar;
begin
  sb := TStringBuffer.Create(Length(Value));
  sb.Append(Value);
  try
    i := 0;
    while i < sb.StringLength do begin
      case sb[i] of
        #$30A2..#$30ED, #$30EF, #$30F2, #$30F3, #$30FC: begin
          c := wcHanKana[sb[i]];
          case c of
            #$FF9E: begin
              sb[i] := wcHanKana[WideChar(integer(sb[i]) - 1)];
              sb.Insert(i + 1, #$FF9E, 1);
              inc(i);
            end;
            #$FF9F: begin
              sb[i] := wcHanKana[WideChar(integer(sb[i]) - 2)];
              sb.Insert(i + 1, #$FF9F, 1);
              inc(i);
            end;
            else sb[i] := c;
          end;
        end;
        #$30F4: begin
          sb[i] := #$FF73;
          sb.Insert(i + 1, #$FF9E, 1);
          inc(i);
        end;
        #$30F7: begin
          sb[i] := #$FF9C;
          sb.Insert(i + 1, #$FF9E, 1);
          inc(i);
        end;
        #$30FA: begin
          sb[i] := #$FF66;
          sb.Insert(i + 1, #$FF9E, 1);
          inc(i);
        end;
      end;
      inc(i);
    end;
    Result := sb.ToString;
  finally
    sb.Free;
  end;
end;

function KataKanaToHiraKana(const Value: WideString): WideString;
var
  sb: TStringBuffer;
  i: integer;
begin
  sb := TStringBuffer.Create(Length(Value));
  sb.Append(Value);
  try
    for i := 0 to sb.StringLength - 1 do begin
      case sb[i] of
        #$30A1..#$30F4: begin
          sb[i] := WideChar(integer(sb[i]) - $60);
        end;
      end;
    end;
    Result := sb.ToString;
  finally
    sb.Free;
  end;
end;

function HiraKanaToKataKana(const Value: WideString): WideString;
var
  sb: TStringBuffer;
  i: integer;
begin
  sb := TStringBuffer.Create(Length(Value));
  sb.Append(Value);
  try
    for i := 0 to sb.StringLength - 1 do begin
      case sb[i] of
        #$3041..#$3094: begin
          sb[i] := WideChar(integer(sb[i]) + $60);
        end;
      end;
    end;
    Result := sb.ToString;
  finally
    sb.Free;
  end;
end;

function GetFileVersion: wideString;
var
  size: DWord;
  sizeFileInfo: DWord;
  ret: DWord;
  pData, pInfo: Pointer;
begin   
  Result := '';   
  size := GetFileVersionInfoSizeW(PWCHAR(paramstrw(0)), ret);
  GetMem(pData, size);
  try
    GetFileVersionInfoW(PWCHAR(paramstrw(0)), 0, size, pData);
    VerQueryValueW(pData,
      PWCHAR(widestring('\StringFileInfo\041103A4\FileVersion')), pInfo, sizeFileInfo);
    Result := PWCHAR(pInfo);
  finally
    FreeMem(pData);
  end;
end;

function SelectedTextToLine(const Value: WideString): WideString;
var
  i: integer;
begin
  for i := 1 to Length(Value) do begin
    if (integer(Value[i]) < $20) then begin
      Result := Copy(Value, 1, i - 1);
      exit;
    end;
  end;
  Result := Value;
end;

function FileExistsOrCreate(var Value: WideString; hWindow:HWND):boolean;
var
  parentDir: WideString;
  s: WideString;
  i: integer;
begin
  Result := false;
  if Length(Value) = 0 then exit;
  if Length(Value) > MAX_PATH then exit;
  for i := 1 to Length(value) do if Value[i] = '/' then exit;
  value := ExpandFileNameW(value);
  if not FileExists2(Value)  then begin
    if DirectoryExists2(Value) then exit; //tH_Ɠt@C͍Ȃ炵
    parentDir := ExtractFileDirW(Value);
    if not DirectoryExists2(parentDir) then exit;
    s := Value + #13#10 +'V쐬܂H';
    if MessageBoxW(hWindow,pwchar(s),'t@C܂',
      MB_OKCANCEL or MB_ICONQUESTION or MB_TASKMODAL) = idCancel then exit;
    if not SaveToFile(value,'',fcNONE) then begin
      MessageBoxW(hWindow,
       'fBXNςĂA'+ #13#10 +
       '܂͏݂łȂꏊ̂悤łB',
       't@C쐬ł܂I',
        MB_OK or MB_ICONINFORMATION or MB_TASKMODAL);
      exit;
    end;
  end;
  Result := true;
end;

//V[gJbgt@C
//http://www2.big.or.jp/~osamu/Delphi/Tips/key.cgi?key=6#0099.txt
const IID_IPersistFile: TGUID= (D1:$0000010B;D2:$0000;D3:$0000;D4:($C0,$00,$00,$00,$00,$00,$00,$46));

function ExtractStringFromLinkFile(Target: WideString ): WideString;
var
  R   : HRESULT;
  SL  : IShellLinkW;
  PF  : IPersistFile;
  Wrk : array[0..MAX_PATH] of WideChar;
  Path: array[0..MAX_PATH] of WideChar;
  WFD : WIN32_FIND_DATAW;
  i: integer;
begin
  Result := Target;
  CoInitialize(nil);
  R := CoCreateInstance(CLSID_ShellLink, nil, CLSCTX_INPROC_SERVER, IID_IShellLinkW, SL);
  if Succeeded(R) then begin
    R := SL.QueryInterface(IID_IPersistFile, PF );
    if Succeeded(R) then begin
      //      MultiByteToWideChar( CP_ACP, 0, PChar(Target), -1, Wrk, MAX_PATH );
      for i := 1 to Length(Target) do Wrk[i - 1] := Target[i];
      for i := Length(Target) to MAX_PATH do Wrk[i] := #0;
      R := PF.Load(Wrk, STGM_READ);
      if Succeeded(R) then begin
        R := SL.Resolve( 0{Application.Handle}, SLR_ANY_MATCH);
        if Succeeded(R) then begin
          SL.GetPath(Path, MAX_PATH, WFD, SLGP_SHORTPATH);
          Result := Path;
        end;
      end;
    end;
  end;
  CoUninitialize;
end;

procedure InitHistIni;
var
  fn: WideString;
begin
  if not IniFilePathNameA(fn) then exit;
  fn := ChangeFileExtW(fn, '.his');
  FHistIni := TMemIniFileW.Create(fn);
end;

procedure UpdateHistIni;
var
  fn: WideString;
  drive: WideChar;
begin
  if not Assigned(FHistIni) then exit;
  if not IniFilePathNameA(fn) then exit;
  fn := ChangeFileExtW(fn, '.his');
  if FileExists2(fn) and FileIsReadOnlyW(fn) then exit;
  drive := CharUpperW(PWChar(WideString(ParamStrW(0)[1])))^;
  if (drive = '\') or not(GetDriveTypeW(PWChar(WideString(drive + ':\'))) in [DRIVE_FIXED]) then exit;
  try
    FHistIni.UpdateFile;
  finally
  end;
end;

procedure UnitInitialize;
begin
  FEditorFont := TFont.Create;
  initKana;
  InitHistIni;
end;

function IsDebuggerPresent: LongBool; external Kernel32 name 'IsDebuggerPresent';

initialization
  DebuggerPresent := IsDebuggerPresent;
finalization
  if Assigned(FEditorFont) then FEditorFont.Free;
  if Assigned(FHistIni) then FHistIni.Free;
end.
