unit NotePadPrinter;

interface

uses Windows, Classes, SysUtils, Graphics, Printers,
  NotePadView, wNotePad, NotePadString;

const
  ColumnCountMin = 1;
  ColumnCountMax = 3;
  MarginMin = 0;
  MarginMax = 100;
  LastPageMax = 999999;
  PreviewMessage: WideString = 'vr[쐬';
  PrintMessage: WideString = '';

type
  TMetaFileList = class(TList)
  protected
    function GetMetaFile(Index: Integer): TMetaFile; virtual;
  public
    destructor Destroy; override;
    procedure Clear; override;
    function Add: TMetaFile;
    property MetaFiles[Index: Integer]: TMetaFile read GetMetaFile;
  end;

  PPaperItem = ^TPaperItem;
  TPaperItem = record
    Name: String;       // p̖O     dmFormName ex 'tHgJ[h 101.6 x 152.4 mm'
    Size: TPoint;       // p̒ƕ dmPaperLength, dmPaperWidth ( ~[gP )
    SizeIndex: Integer; // pTCY     dmPaperSize
  end;

  (*
    Size: TPoint ɂ́A1/10 ~[gPʂŒli[Axo邪
    TStringsPrinterInfo.FPaperInfoList: array[1..68] of TPaperInfo Ƃ̉
    ݊ۂ߁A~[gPʂƂĈĂB
    TPaperItemList.UpdateList ł́ADeviceCapabilities Ԃ 1/10 ~[g
    Pʂ̒ƕ~[gPʂɕϊĂB
  *)

  TPaperItemList = class(TList)
  protected
    function GetPaperItem(Index: Integer): PPaperItem; virtual;
  public
    destructor Destroy; override;
    procedure Clear; override;
    function Add: PPaperItem;
    procedure UpdateList;
    property Papers[Index: Integer]: PPaperItem read GetPaperItem;
  end;

  TPaperInfo = record
    Name: String;
    Size: TPoint;
  end;

  TStringsPrinterInfo = class
  protected
    FPrinterIndex: Integer;
    FPaperItemList: TPaperItemList;
    FPaperInfoList: array[1..68] of TPaperInfo;
    procedure CheckPrinterIndex; virtual;
    function GetPrinterName: String;
    function GetPaperName: String;
    function GetPaperIndex: Integer;
    function GetPaperSize: TPoint;
  public
    constructor Create;
    destructor Destroy; override;
    property PrinterName: String read GetPrinterName;
    property PaperName: String read GetPaperName;
    property PaperIndex: Integer read GetPaperIndex;
    property PaperSize: TPoint read GetPaperSize;
  end;

  TStringsPrinterHeaderAlign = (haNone, haLeft, haCenter, haRight);
  TStringsPrinterRangeOption = (roWhole, roLimited);


  {///////////////////////////////////////////////////////////////////

    TStringsPrinterPrintMode

    pmBW ............ ̔B
    pmThreeColor .... BkColor, Font.Color, HeaderColor ̂ݗLB
    pmBWFontStyle ... Bg[N Font.Style ̂ݗLB
    pmColorToken .... g[N̐FB

    pmBW, pmThreeColor           Fountain KvƂȂ
    pmBWFontStyle, pmColorToken  Fountain KvƂ

  ////////////////////////////////////////////////////////////////////}

  TStringsPrinterPrintMode = (pmBW, pmBWFontStyle, pmColorToken);

  TNotePadPrinter = class(TComponent)
  private
    FPageNumber: Integer;
    FOptions: TNotePadOptions;

    FNotePad: TwNotePad;
    FNotePadString: TNotePadString;
    FFont: TFont;
    FRangeOption: TStringsPrinterRangeOption;
    FPrinterInfo: TStringsPrinterInfo;
    FFirstPageNumber: integer;
    FFirstLineNumber: Integer;
    FStartPage: integer;
    FLastPage: integer;
    FMarginLeft: Integer;
    FMarginRight: Integer;
    FMarginBottom: Integer;
    FMarginTop: Integer;
    FHeader: WideString;
    FFooter: WideString;
    FHeaderAlign: TStringsPrinterHeaderAlign;
    FFooterAlign: TStringsPrinterHeaderAlign;
    FLineNumber: Boolean;
    FColumnCount: Integer;
    FColumnSpace: Integer;
    FMarginLine: Integer;
    FPrintLineNumber: Integer;
    FStartLine: Integer;
    FPrintMode: TStringsPrinterPrintMode;
    FFileName: WideString;
    FLineRangeOption: TStringsPrinterRangeOption;
    FLastLine: Integer;
    FAscent: integer;
    function DivideStringsToMetaFiles(MetaFileList: TMetaFileList; Mes: WideString): Boolean;
    procedure DrawLine(I: integer; MetaFileCanvas: TMetafileCanvas; Drw: TRect; Fw: integer);
    procedure SetNotePad(const Value: TwNotePad);
    procedure SetMarginBottom(Value: Integer);
    procedure SetMarginLeft(Value: Integer);
    procedure SetMarginRight(Value: Integer);
    procedure SetMarginTop(Value: Integer);
    procedure SetFont(Value: TFont);
    procedure SetColumnCount(Value: Integer);
    procedure SetColumnSpace(Value: Integer);
    procedure SetMarginLine(Value: Integer);
    procedure SetFirstLineNumber(Value: Integer);
    procedure SetFirstPageNumber(Value: Integer);
    procedure SetStartLine(Value: Integer);
    procedure SetStartPage(Value: Integer);
    procedure SetLastLine(const Value: Integer);
    procedure SetLastPage(const Value: Integer);
  protected
  public
    constructor Create(AOwner: TComponent);override;
    destructor Destroy;override;
    function PrintExecute: boolean;
    function PreviewExecute: boolean;
    function PrinterDot(MM: Double): Integer;
    function Ratio: Double;
    function PaperOffset: TRect;
    function ParseHeader(const S: WideString): WideString; 
    property NotePad: TwNotePad read FNotePad write SetNotePad;
    property PrinterInfo: TStringsPrinterInfo read FPrinterInfo;
    property MarginBottom: Integer read FMarginBottom write SetMarginBottom;
    property MarginLeft: Integer read FMarginLeft write SetMarginLeft;
    property MarginRight: Integer read FMarginRight write SetMarginRight;
    property MarginTop: Integer read FMarginTop write SetMarginTop;
    property Font: TFont read FFont write SetFont;
    property Footer: WideString read FFooter write FFooter;
    property Header: WideString read FHeader write FHeader;
    property FooterAlign: TStringsPrinterHeaderAlign read FFooterAlign write FFooterAlign;
    property HeaderAlign: TStringsPrinterHeaderAlign read FHeaderAlign write FHeaderAlign;
    property LineNumber: Boolean read FLineNumber write FLineNumber;
    property ColumnCount: Integer read FColumnCount write SetColumnCount;
    property ColumnSpace: Integer read FColumnSpace write SetColumnSpace;
    property MarginLine: Integer read FMarginLine write SetMarginLine;
    property FirstLineNumber: Integer read FFirstLineNumber write SetFirstLineNumber;
    property FirstPageNumber: Integer read FFirstPageNumber write SetFirstPageNumber;
    property RangeOption: TStringsPrinterRangeOption read FRangeOption write FRangeOption;
    property StartLine: Integer read FStartLine write SetStartLine;
    property StartPage: Integer read FStartPage write SetStartPage;
    property LastLine: Integer read FLastLine write SetLastLine;
    property LastPage: Integer read FLastPage write SetLastPage;
    property PrintMode: TStringsPrinterPrintMode read FPrintMode write FPrintMode;
    property FileName: WideString read FFileName write FFileName;
    property LineRangeOption: TStringsPrinterRangeOption read FLineRangeOption write FLineRangeOption;
  end;

function IsFixedFont(Value: TFont; h: HDC): boolean;
function IsFixedFontName(const Value: WideString; h: HDC): boolean;

implementation

uses WinSpool, Math, StringsPrinterPreview, StringsPrinterProgress,
  wsFileOp, NotePadMode, NotePadClasses, NotePadConsts, FontUtils, MaskUtils,
  Types, DM;

function IsFixedFontName(const Value: WideString; h: HDC): boolean;
var
  AFont: TFont;
begin
  AFont := TFont.Create;
  AFont.Charset := DEFAULT_CHARSET;
  AFont.Name := Value;
  Result := IsFixedFont(AFont, h);
  AFont.Free;
end;

function IsFixedFont(Value: TFont; h: HDC): boolean;
var
  TM: TEXTMETRICW;
begin
  SelectObject(h, Value.Handle);
  GetTextMetricsW(h, TM);
  Result := {((TM.tmPitchAndFamily and 1) <> 1) and }(TM.tmCharSet <> SYMBOL_CHARSET);
end;

var
  FFontWidthMap: array [$0000 .. $FFFF] of Word;

function GetGlyphWidth(const Value: WideChar; H: HDC): integer;
var
  W: array[0..0]of integer;
begin
  if FFontWidthMap[integer(Value)] > 0 then Result := FFontWidthMap[integer(Value)]
  else begin
    GetCharWidth32W(H, Cardinal(Value), Cardinal(Value), W);
    Result := Max(1,W[0]);
    FFontWidthMap[integer(Value)] := Result;
  end;
end;

procedure InitFontWidthMap(AMetafileCanvas: TMetafileCanvas);
var
  i: integer;
begin
  for i := $0 to $FFFF do FFontWidthMap[i] := 0;
  GetGlyphWidth(WideChar($3000), AMetafileCanvas.Handle);//ŕKĝŌ߂Ă
end;

{ TPaperItemList }

function TPaperItemList.GetPaperItem(Index: Integer): PPaperItem;
begin
  Result := PPaperItem(Items[Index]);
end;

destructor TPaperItemList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TPaperItemList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    Dispose(PPaperItem(Items[I]));
  inherited Clear; // SetCount(0); SetCapacity(0);
end;

function TPaperItemList.Add: PPaperItem;
var
  P: PPaperItem;
begin
  New(P);
  inherited Add(P);
  Result := P;
end;

procedure TPaperItemList.UpdateList;
(*
  ݂̃v^[T|[gĂṕuOvuƕvupTCYv
  擾B
  nifty:FDELPHI/MES/08/09033 Satobe ̔ɂĂB
  ƕ́ADeviceCapabilities Ԃ 1/10 ~[gPʂ̒l
  ~[gPʂɕϊĂB

  v^[CXg[ĂȂ UpdateList s
  XgNA邾ɂȂB
*)
type
  TPaperName = array[0..63] of Char;
  TPaperNameArray = array[0..999] of TPaperName;
  PPaperNameArray = ^TPaperNameArray;
  TSizeArray = array[0..999] of TPoint;
  PSizeArray = ^TSizeArray;
  TPapersArray = array[0..999] of WORD;
  PPapersArray = ^TPapersArray;
var
  Device, Driver, Port: array[0..79] of char;
  DeviceMode: THandle;
  DevMode: PDeviceMode;
  PapersCount: DWORD;
  PPaperNames: PPaperNameArray;    // dmFormName
  PSizes: PSizeArray;              // dmPaperLength, dmPaperWidth (1/10~[gP)
  PPapers: PPapersArray;           // dmPaperSize
  I: Integer;
  Info: PPaperItem;
begin
  Clear;
  try
    Printer.GetPrinter(Device, Driver, Port, DeviceMode);
    if DeviceMode <> 0 then
    begin
      DevMode := GlobalLock(DeviceMode);
      try
        PapersCount := DeviceCapabilities(Device, Port, DC_PAPERS, nil, DevMode);
        if PapersCount > 0 then
        begin
          GetMem(PPaperNames,PapersCount * SizeOf(TPaperName));
          GetMem(PSizes,     PapersCount * SizeOf(TPoint));
          GetMem(PPapers,    PapersCount * SizeOf(WORD));
          try
            // p̖ÕXg
            DeviceCapabilities(Device, Port, DC_PAPERNAMES, PChar(PPaperNames), Devmode);
            // p@i1/10~Pʁj TPoint ^̃Xg
            DeviceCapabilities(Device, Port, DC_PAPERSIZE, PChar(PSizes), Devmode);
            // pTCỸXg
            DeviceCapabilities(Device, Port, DC_PAPERS, PChar(PPapers), Devmode);
            for I := 0 to PapersCount - 1 do
            begin
              Info := Add;
              Info^.Name := StrPas(PPaperNames^[I]);
              Info^.Size := PSizes^[I];
              Info^.SizeIndex := PPapers^[I];
              // 1/10 ~[g -> ~[g ľܓ
              Info^.Size.X := Trunc(Info^.Size.X / 10 + 0.5);
              Info^.Size.Y := Trunc(Info^.Size.Y / 10 + 0.5);
            end;
          finally
            FreeMem(PPaperNames);
            FreeMem(PSizes);
            FreeMem(PPapers);
          end;
        end;
      finally
        GlobalUnlock(DeviceMode);
      end;
    end;
  except
    ; // ݂̃v^[ł̓XgNA邾
  end;
end;

{ TMetaFileList }

destructor TMetaFileList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TMetaFileList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    TMetaFile(Items[I]).Free;
  inherited Clear; // SetCount(0); SetCapacity(0);
end;

function TMetaFileList.Add: TMetaFile;
begin
  Result := TMetaFile.Create;
  inherited Add(Result);
end;

function TMetaFileList.GetMetaFile(Index: Integer): TMetaFile;
begin
  Result := TMetaFile(Items[Index]);
end;

{ TStringsPrinterInfo }

constructor TStringsPrinterInfo.Create;
begin
  inherited;

  FPrinterIndex := -1;
  FPaperItemList := TPaperItemList.Create;

  // ̎_ updatelist sƁÃ݂v^[ł́A
  // G[AɎŝōsȂ
  // FPaperItemList.UpdateList;

  // default
  FPaperInfoList[ 1].Name := 'DMPAPER_LETTER';
  FPaperInfoList[ 1].Size := Point(305, 279);
  FPaperInfoList[ 2].Name := 'DMPAPER_LETTERSMALL';
  FPaperInfoList[ 2].Size := Point(305, 279);
  FPaperInfoList[ 3].Name := 'DMPAPER_TABLOID';
  FPaperInfoList[ 3].Size := Point(279, 432);
  FPaperInfoList[ 4].Name := 'DMPAPER_LEDGER';
  FPaperInfoList[ 4].Size := Point(432, 279);
  FPaperInfoList[ 5].Name := 'DMPAPER_LEGAL';
  FPaperInfoList[ 5].Size := Point(305, 356);
  FPaperInfoList[ 6].Name := 'DMPAPER_STATEMENT';
  FPaperInfoList[ 6].Size := Point(305, 203);
  FPaperInfoList[ 7].Name := 'DMPAPER_EXECUTIVE';
  FPaperInfoList[ 7].Size := Point(356, 254);
  FPaperInfoList[ 8].Name := 'DMPAPER_A3';
  FPaperInfoList[ 8].Size := Point(297, 420);
  FPaperInfoList[ 9].Name := 'DMPAPER_A4';
  FPaperInfoList[ 9].Size := Point(210, 297);
  FPaperInfoList[10].Name := 'DMPAPER_A4SMALL';
  FPaperInfoList[10].Size := Point(210, 297);
  FPaperInfoList[11].Name := 'DMPAPER_A5';
  FPaperInfoList[11].Size := Point(148, 210);
  FPaperInfoList[12].Name := 'DMPAPER_B4';
  FPaperInfoList[12].Size := Point(250, 354);
  FPaperInfoList[13].Name := 'DMPAPER_B5';
  FPaperInfoList[13].Size := Point(182, 257);
  FPaperInfoList[14].Name := 'DMPAPER_FOLIO';
  FPaperInfoList[14].Size := Point(305, 330);
  FPaperInfoList[15].Name := 'DMPAPER_QUARTO';
  FPaperInfoList[15].Size := Point(215, 275);
  FPaperInfoList[16].Name := 'DMPAPER_10X14';
  FPaperInfoList[16].Size := Point(254, 356);
  FPaperInfoList[17].Name := 'DMPAPER_11X17';
  FPaperInfoList[17].Size := Point(279, 432);
  FPaperInfoList[18].Name := 'DMPAPER_NOTE';
  FPaperInfoList[18].Size := Point(305, 279);
  FPaperInfoList[19].Name := 'DMPAPER_ENV_9';
  FPaperInfoList[19].Size := Point(203, 203); // Envelope #9 3 78 x 8 78
  FPaperInfoList[20].Name := 'DMPAPER_ENV_10';
  FPaperInfoList[20].Size := Point(457, 229); // Envelope #10 4 18 x 9 12
  FPaperInfoList[21].Name := 'DMPAPER_ENV_11';
  FPaperInfoList[21].Size := Point(305, 254); // Envelope #11 4 12 x 10 38
  FPaperInfoList[22].Name := 'DMPAPER_ENV_12';
  FPaperInfoList[22].Size := Point(276, 279); // Envelope #12 4 \276 x 11
  FPaperInfoList[23].Name := 'DMPAPER_ENV_14';
  FPaperInfoList[23].Size := Point(127, 279); // Envelope #14 5 x 11 12
  FPaperInfoList[24].Name := 'DMPAPER_CSHEET';
  FPaperInfoList[24].Size := Point(100, 100); // C size sheet
  FPaperInfoList[25].Name := 'DMPAPER_DSHEET';
  FPaperInfoList[25].Size := Point(100, 100); // D size sheet
  FPaperInfoList[26].Name := 'DMPAPER_ESHEET';
  FPaperInfoList[26].Size := Point(100, 100); // E size sheet
  FPaperInfoList[27].Name := 'DMPAPER_ENV_DL';
  FPaperInfoList[27].Size := Point(110, 220);
  FPaperInfoList[28].Name := 'DMPAPER_ENV_C5';
  FPaperInfoList[28].Size := Point(162, 229);
  FPaperInfoList[29].Name := 'DMPAPER_ENV_C3';
  FPaperInfoList[29].Size := Point(324, 458);
  FPaperInfoList[30].Name := 'DMPAPER_ENV_C4';
  FPaperInfoList[30].Size := Point(229, 324);
  FPaperInfoList[31].Name := 'DMPAPER_ENV_C6';
  FPaperInfoList[31].Size := Point(114, 162);
  FPaperInfoList[32].Name := 'DMPAPER_ENV_C65';
  FPaperInfoList[32].Size := Point(114, 229);
  FPaperInfoList[33].Name := 'DMPAPER_ENV_B4';
  FPaperInfoList[33].Size := Point(250, 353);
  FPaperInfoList[34].Name := 'DMPAPER_ENV_B5';
  FPaperInfoList[34].Size := Point(176, 250);
  FPaperInfoList[35].Name := 'DMPAPER_ENV_B6';
  FPaperInfoList[35].Size := Point(176, 125);
  FPaperInfoList[36].Name := 'DMPAPER_ENV_ITALY';
  FPaperInfoList[36].Size := Point(110, 230);
  FPaperInfoList[37].Name := 'DMPAPER_ENV_MONARCH';
  FPaperInfoList[37].Size := Point(98, 190);
  FPaperInfoList[38].Name := 'DMPAPER_ENV_PERSONAL';
  FPaperInfoList[38].Size := Point(91, 155); // 6 34 Envelope 3 58 x 6 12 in
  FPaperInfoList[39].Name := 'DMPAPER_FANFOLD_US';
  FPaperInfoList[39].Size := Point(203, 279); // US Std Fanfold 14 78 x 11 in
  FPaperInfoList[40].Name := 'DMPAPER_FANFOLD_STD_GERMAN';
  FPaperInfoList[40].Size := Point(305, 305); // German Std Fanfold 8 12 x 12 in
  FPaperInfoList[41].Name := 'DMPAPER_FANFOLD_LGL_GERMAN';
  FPaperInfoList[41].Size := Point(305, 330); // German Legal Fanfold 8 12 x 13 in
  FPaperInfoList[42].Name := 'DMPAPER_ISO_B4';
  FPaperInfoList[42].Size := Point(250, 353);
  FPaperInfoList[43].Name := 'DMPAPER_JAPANESE_POSTCARD';
  FPaperInfoList[43].Size := Point(100, 148);
  FPaperInfoList[44].Name := 'DMPAPER_9X11';
  FPaperInfoList[44].Size := Point(229, 279);
  FPaperInfoList[45].Name := 'DMPAPER_10X11';
  FPaperInfoList[45].Size := Point(254, 279);
  FPaperInfoList[46].Name := 'DMPAPER_15X11';
  FPaperInfoList[46].Size := Point(381, 279);
  FPaperInfoList[47].Name := 'DMPAPER_ENV_INVITE';
  FPaperInfoList[47].Size := Point(220, 220);
  FPaperInfoList[48].Name := 'DMPAPER_RESERVED_48';
  FPaperInfoList[48].Size := Point(100, 100); // RESERVED--DO NOT USE
  FPaperInfoList[49].Name := 'DMPAPER_RESERVED_49';
  FPaperInfoList[49].Size := Point(100, 100); // RESERVED--DO NOT USE
  FPaperInfoList[50].Name := 'DMPAPER_LETTER_EXTRA';
  FPaperInfoList[50].Size := Point(275, 305);
  FPaperInfoList[51].Name := 'DMPAPER_LEGAL_EXTRA';
  FPaperInfoList[51].Size := Point(275, 381);
  FPaperInfoList[52].Name := 'DMPAPER_TABLOID_EXTRA';
  FPaperInfoList[52].Size := Point(297, 457);
  FPaperInfoList[53].Name := 'DMPAPER_A4_EXTRA';
  FPaperInfoList[53].Size := Point(235, 322);
  FPaperInfoList[54].Name := 'DMPAPER_LETTER_TRANSVERSE';
  FPaperInfoList[54].Size := Point(275, 279);
  FPaperInfoList[55].Name := 'DMPAPER_A4_TRANSVERSE';
  FPaperInfoList[55].Size := Point(210, 297);
  FPaperInfoList[56].Name := 'DMPAPER_LETTER_EXTRA_TRANSVERSE';
  FPaperInfoList[56].Size := Point(275, 305);
  FPaperInfoList[57].Name := 'DMPAPER_A_PLUS';
  FPaperInfoList[57].Size := Point(227, 356);
  FPaperInfoList[58].Name := 'DMPAPER_B_PLUS';
  FPaperInfoList[58].Size := Point(305, 487);
  FPaperInfoList[59].Name := 'DMPAPER_LETTER_PLUS';
  FPaperInfoList[59].Size := Point(216, 322);
  FPaperInfoList[60].Name := 'DMPAPER_A4_PLUS';
  FPaperInfoList[60].Size := Point(210, 330);
  FPaperInfoList[61].Name := 'DMPAPER_A5_TRANSVERSE';
  FPaperInfoList[61].Size := Point(148, 210);
  FPaperInfoList[62].Name := 'DMPAPER_B5_TRANSVERSE';
  FPaperInfoList[62].Size := Point(182, 257);
  FPaperInfoList[63].Name := 'DMPAPER_A3_EXTRA';
  FPaperInfoList[63].Size := Point(322, 445);
  FPaperInfoList[64].Name := 'DMPAPER_A5_EXTRA';
  FPaperInfoList[64].Size := Point(174, 235);
  FPaperInfoList[65].Name := 'DMPAPER_B5_EXTRA';
  FPaperInfoList[65].Size := Point(210, 276);
  FPaperInfoList[66].Name := 'DMPAPER_A2';
  FPaperInfoList[66].Size := Point(420, 594);
  FPaperInfoList[67].Name := 'DMPAPER_A3_TRANSVERSE';
  FPaperInfoList[67].Size := Point(297, 420);
  FPaperInfoList[68].Name := 'DMPAPER_A3_EXTRA_TRANSVERSE';
  FPaperInfoList[68].Size := Point(322, 445);
end;

destructor TStringsPrinterInfo.Destroy;
begin
  FPaperItemList.Free;
  inherited Destroy;
end;

procedure TStringsPrinterInfo.CheckPrinterIndex;
// v^[ςĂ΁AXgč\z
begin
  try
    if FPrinterIndex <> Printer.PrinterIndex then
    begin
      FPrinterIndex := Printer.PrinterIndex;
      FPaperItemList.UpdateList;
    end;
  except
    // ݂̃v^[ł́AXgNA
    FPaperItemList.Clear;
  end;
end;

function TStringsPrinterInfo.GetPrinterName: String;
begin
  try
    Result := Printer.Printers[Printer.PrinterIndex];
  except
    // ݂̃v^[ł́AsԂ
    Result := 's';
  end;
end;

function TStringsPrinterInfo.GetPaperName: String;
var
  Idx, I: Integer;
begin
  CheckPrinterIndex;
  Idx := GetPaperIndex;
  for I := 0 to FPaperItemList.Count - 1 do
    if FPaperItemList.Papers[I].SizeIndex = Idx then
    begin
      Result := FPaperItemList.Papers[I].Name;
      Exit;
    end;
  // ݂̃v^[ł́ACheckPrinterIndex ŃXgNÂ
  // ̃[v͉ȂB܂AGetpaperIndex  DMPAPER_A4 Ԃ̂
  // FPaperInfoList[DMPAPER_A4].Name ݒ肳
  try
    Result := FPaperInfoList[Idx].Name;
  except
    Result := '';
  end;
end;

function TStringsPrinterInfo.GetPaperSize: TPoint;
var
  Idx, I: Integer;
begin
  CheckPrinterIndex;
  Idx := GetPaperIndex;
  for I := 0 to FPaperItemList.Count - 1 do
    if FPaperItemList.Papers[I].SizeIndex = Idx then
    begin
      Result := FPaperItemList.Papers[I].Size;
      Exit;
    end;
  // ݂̃v^[ł́ACheckPrinterIndex ŃXgNÂ
  // ̃[v͉ȂB܂AGetpaperIndex  DMPAPER_A4 Ԃ̂
  // FPaperInfoList[DMPAPER_A4].Size ݒ肳
  try
    Result := FPaperInfoList[Idx].Size;
  except
    Result := Point(0, 0);
  end;
end;

function TStringsPrinterInfo.GetPaperIndex: Integer;
(*
  ݂̃v^[ɐݒ肳ĂpTCY̔ԍԂ
  ݂̃v^[ł́ADMPAPER_A4 Ԃ
*)
var
  ADevice, ADriver, APort: array[0..255] of Char;
  ADeviceMode: THandle;
  ADevMode : ^TDevMode;
begin
  try
    Printer.GetPrinter(ADevice, ADriver, APort, ADeviceMode);
    if ADeviceMode <> 0 then
    begin
      ADevMode := GlobalLock(ADeviceMode);
      try
        Result := ADevMode^.dmPaperSize;
      finally
        GlobalUnLock(ADeviceMode);
      end;
    end
    else
      Result := DMPAPER_A4;
  except
    // ݂̃v^[ł́ADMPAPER_A4 ftHgƂ
    Result := DMPAPER_A4;
  end;
end;

{ TNotePadPrinter }

constructor TNotePadPrinter.Create(AOwner: TComponent);
begin
  inherited;
  FPrinterInfo := TStringsPrinterInfo.Create;
  FFont := TFont.Create;
  FFont.Name := DefaultFontName;
  FNotePadString := TNotePadString.Create;
  FColumnCount := 1;
  FColumnSpace := 10;
  FLastPage := 1;//LastPageMax; //MAXĂƏɖ߂̂ʓ|
  FStartPage := 1;
  FStartLine := 1;
  FFirstPageNumber := 1;
  FFirstLineNumber := 1;
end;

destructor TNotePadPrinter.Destroy;
begin
  FNotePadString.Free;
  FPrinterInfo.Free;
  FFont.Free;
  inherited;
end;

procedure TNotePadPrinter.SetNotePad(const Value: TwNotePad);
begin
  FNotePad := Value;
  FNotePadString.Text := Value.Text;
  FOptions := Value.Options;
  FNotePadString.TabCount := FOptions.TabCount;
end;

function TNotePadPrinter.Ratio: Double;
(*
  ʉ𑜓xƃv^[𑜓x̔䗦ԂB
  ݂̃v^[ł́APԂAʏ DMPAPER_A4 TCY
  p\dlƂ
*)
var
  DC: HDC;
begin
  try
    DC := GetDC(0);
    try
      Result := GetDeviceCaps(DC, LOGPIXELSX) /
                GetDeviceCaps(Printer.Handle, LOGPIXELSX);
    finally
      ReleaseDC(0, DC);
    end;
  except
    // ݂̃v^[
    Result := 1;
  end;
end;

function TNotePadPrinter.PrinterDot(MM: Double): Integer;
(*
  v^[ mm v^[̃hbgɕϊ
  ݂̃v^[ł͉ʏ̃hbgɕϊ
*)
var
  DC: HDC;
begin
  try
    Result := Round(MM * GetDeviceCaps(Printer.Handle, LOGPIXELSX) / 25.4);
  except
    // ݂̃v^[ł́Aʏ̃hbgɕϊ
    DC := GetDC(0);
    try
      Result := Round(MM * GetDeviceCaps(DC, LOGPIXELSX) / 25.4);
    finally
      ReleaseDC(0, DC);
    end;
  end;
end;

(*
  MarginLeft                                              MarginRight
  /                                                       /
|     |                                                 |     |
+-------------------------------------------------------------+ <------
| 0,0                    Offset.Top                           |  Margin
| +---------------------------------------------------------+ |  Top
| |                                                         | |
| |   +-------------------------------------------------+   | | <------
| |   |                   Hr                            |   | |
| |   +----+--------+---+----+--------+---+----+--------+   | |
|O|   |    |        | C |    |        | C |    |        |   |O|
|f|   |    |        | o |    |        | o |    |        |   |f|
|f|   | Nr |  Dr    | l | Nr |  Dr    | l | Nr |  Dr    |   |f|
|s|   |    |        | u |    |        | u |    |        |   |s|
|e|   |    |        | m |    |        | m |    |        |   |e|
|t|   |    |        | S |    |        | S |    |        |   |t|
|.|   |    |        | p |    |        | p |    |        |   |.|
|L|   |    |        | a |    |        | a |    |        |   |R|
|e|   |    |        | c |    |        | c |    |        |   |i|
|f|   |    |        | e |    |        | e |    |        |   |g|
|t|   +----+--------+---+----+--------+---+----+--------+   |h|
| |   |                   Fr                            |   |t|
| |   +-------------------------(Cr.Right, Cr.Bottom)-->+   | | <------
| |                                                         | |  Margin
| +-------------------------------(PageWidth, PageHeight)-->+ |  Bottom
|                        Offset.Bottom                        |
+------------------------------------------(PaperSize.X, Y)-->+ <------
*)

function TNotePadPrinter.PaperOffset: TRect;
(*
   ݁AIĂy[p[TCYsNZlɕϊA
   PageWidth, PageHeight Ƃ̍ TRect lŕԂ

   ݂̃v^[ł́APrinter.PageWidth, Printer.PageHight 
   pTCYƂ̍lȂĂǂ̂ŁARect(0, 0, 0, 0) ԂB
*)
var
  X, Y: Integer;
begin
  try
    if Printer.Orientation = poPortrait then
    begin
      X := PrinterDot(FPrinterInfo.PaperSize.X);
      Y := PrinterDot(FPrinterInfo.PaperSize.Y);
    end
    else
    begin
      X := PrinterDot(FPrinterInfo.PaperSize.Y);
      Y := PrinterDot(FPrinterInfo.PaperSize.X);
    end;
    Result.Left := Max(0, (X - Printer.PageWidth) div 2);
    Result.Top := Max(0, (Y - Printer.PageHeight) div 2);
    Result.Right := Max(0, X - Printer.PageWidth - Result.Left);
    Result.Bottom := Max(0, Y - Printer.PageHeight - Result.Top);
  except
    // ݂̃v^[
    Result := Rect(0, 0, 0, 0);
  end;
end;

procedure TNotePadPrinter.SetMarginBottom(Value: Integer);
begin
  if FMarginBottom <> Value then
    FMarginBottom := Min(Max(Value, MarginMin), MarginMax);
end;

procedure TNotePadPrinter.SetMarginLeft(Value: Integer);
begin
  if FMarginLeft <> Value then
    FMarginLeft := Min(Max(Value, MarginMin), MarginMax);
end;

procedure TNotePadPrinter.SetMarginRight(Value: Integer);
begin
  if FMarginRight <> Value then
    FMarginRight := Min(Max(Value, MarginMin), MarginMax);
end;

procedure TNotePadPrinter.SetMarginTop(Value: Integer);
begin
  if FMarginTop <> Value then
    FMarginTop := Min(Max(Value, MarginMin), MarginMax);
end;

procedure TNotePadPrinter.SetFont(Value: TFont);
begin
  if Value <> nil then
    FFont.Assign(Font);
  FFont.Charset := GetFontCharSet(FFont.Name);
end;

procedure TNotePadPrinter.SetColumnCount(Value: Integer);
begin
  if FColumnCount <> Value then
    FColumnCount := Min(Max(Value, ColumnCountMin), ColumnCountMax);
end;

procedure TNotePadPrinter.SetColumnSpace(Value: Integer);
begin
  if FColumnSpace <> Value then
    FColumnSpace := Min(Max(Value, MarginMin), MarginMax);
end;

procedure TNotePadPrinter.SetMarginLine(Value: Integer);
begin
  if FMarginLine <> Value then
    FMarginLine := Min(Max(Value, MarginMin), MarginMax);
end;

procedure TNotePadPrinter.SetFirstLineNumber(Value: Integer);
begin
  if FFirstLineNumber <> Value then
    FFirstLineNumber := Min(Max(Value, 1), LastPageMax);
end;

procedure TNotePadPrinter.SetFirstPageNumber(Value: Integer);
begin
  if FFirstPageNumber <> Value then
    FFirstPageNumber := Min(Max(Value, 1), LastPageMax);
end;

procedure TNotePadPrinter.SetStartLine(Value: Integer);
begin
  if FStartLine <> Value then
    FStartLine := Min(Max(Value, 1), LastPageMax);
end;

procedure TNotePadPrinter.SetStartPage(Value: Integer);
begin
  if FStartPage <> Value then
    FStartPage := Min(Max(Value, 1), LastPageMax);
end;

function TNotePadPrinter.ParseHeader(const S: WideString): WideString;
(*
   %p    ݂̃y[Wԍ
   %f    t@Ci΃pXj
   %F    t@CiPt@Cj
   %y    ݂̐N
   %m    ݂̌
   %d    ݂̓
   %t    ݂̎
*)
var
  P: PWChar;
  Y, M, D: Word;
begin
  Result := '';
  DecodeDate(Now, Y, M, D);
  P := PWChar(S);
  while P^ <> #0 do
  begin
    if P^ = '%' then
    begin
      Inc(P);
      case P^ of
        'p': Result := Result + IntToStr(FPageNumber);
        'f': Result := Result + FFileName;
        'F': Result := Result + ExtractFileNameW(FFileName);
        'y': Result := Result + IntToStr(Y);
        'm': Result := Result + IntToStr(M);
        'd': Result := Result + IntToStr(D);
        't': Result := Result + TimeToStr(Now);
      else
        Result := Result + P^;
      end;
    end
    else
      Result := Result + P^;
    Inc(P);
  end;
end;

function TNotePadPrinter.DivideStringsToMetaFiles(
  MetaFileList: TMetaFileList; Mes: WideString): Boolean;
var
  TM: TTextMetric;
  Fh, Fw: Integer;
  I: Integer;
  CurrentColumn, ColumnOffset: Integer;
  LineMargin, ClientX, ClientY: Integer;
  Cr, Hr, Fr, Nr, Dr, Nrw, Drw, Drwl, Offset: TRect;
  Form: TFormStringsPrinterProgress;
  MetaFile: TMetaFile;
  MetaFileCanvas: TMetaFileCanvas;
  CurrentLineNumber: Integer; // 0 base
  parser: TNotePadParser;
  AWrapCount: integer;
  p:TSize;

  procedure NewMetaFileCanvas;
  begin
    MetaFile := MetaFileList.Add;
    MetaFile.Width := ClientX;
    MetaFile.Height := ClientY;
    MetaFileCanvas := TMetaFileCanvas.Create(MetaFile, 0);
    MetaFileCanvas.Font.Assign(Font);
    MetaFileCanvas.Font.Size := Round(Font.Size / Ratio);
    if PrintMode in [pmColorToken] then
      MetaFileCanvas.Brush.Color := FOptions.ColorBack
    else
      MetaFileCanvas.Brush.Color := clWhite;
    MetaFileCanvas.Brush.Style := bsSolid;
    MetaFileCanvas.FillRect(Rect(0, 0, ClientX, ClientY));
    MetaFileCanvas.Brush.Style := bsClear;
    // Font.Color, Font.Style ͕`̓sxݒ肳
  end;

  procedure InitDrawRects;
  var
    NewWrapByte: integer;
  begin
    FNotePadString.WrapByte := MAX_WRAPBYTE;
    repeat
      // LineNumber Rect
      if LineNumber then
        Nr := Rect(
                Cr.Left,
                Hr.Bottom,
                Cr.Left + Fw * (Length(IntToStr(FNotePadString.Count)) + 2),
                Fr.Top
              )
      else
        Nr := Rect(Cr.Left, Hr.Bottom, Cr.Left, Fr.Top);
      // Draw Rect
      Dr := Rect(
              Nr.Right,
              Hr.Bottom,
              Nr.Right + (
                Cr.Right - Nr.Right -
                (Nr.Right - Nr.Left + PrinterDot(ColumnSpace)) *
                (ColumnCount - 1)
              ) div ColumnCount,
              Fr.Top
            );
      NewWrapByte := (Dr.Right - Dr.Left  - GetGlyphWidth(WideChar($3000), 0)) div Fw;//Spꕶ̗]T
      if FNotePadString.WrapByte = NewWrapByte then break;
      FNotePadString.WrapByte := NewWrapByte;
    until FNotePadString.WrapByte <> NewWrapByte;
  end;

  procedure AssignHeaderColor;
  begin
    if PrintMode in [pmColorToken] then
    begin
      MetaFileCanvas.Font.Color := FOptions.ColorFore;
    end
    else
    begin
      MetaFileCanvas.Font.Color := clBlack;
    end;
    MetaFileCanvas.Brush.Style := bsClear;
    MetaFileCanvas.Font.Style := Font.Style;
  end;

  procedure DrawHeader(Align: TStringsPrinterHeaderAlign; const S: WideString;
    DrawRect: TRect);
  var
    R: TRect;
    Options: Integer;
  begin
    if (Align <> haNone) and (S <> '') then
    begin
      Options := DT_TOP or DT_SINGLELINE or DT_NOPREFIX;
      case Align of
        haLeft:
          Options := Options or DT_LEFT;
        haCenter:
          Options := Options or DT_CENTER;
        haRight:
          Options := Options or DT_RIGHT;
      end;
      R := DrawRect;
      AssignHeaderColor;
      // DT_LEFT, DT_CENTER, DT_RIGHT ̃IvV𗘗p̂ŁADrawText gp
      DrawTextW(MetaFileCanvas.Handle, PWChar(ParseHeader(S)), -1, R, Options);
    end;
  end;


begin
  Result := False;
  // Offset
  Offset := PaperOffset;
  // Client Rect
  try
    ClientX := Printer.PageWidth;
    ClientY := Printer.PageHeight;
  except
    // ݂̃v^[ł́ADMPAPER_A4 ̉ʏł̃sNZɐݒ肳
    ClientX := PrinterDot(PrinterInfo.PaperSize.X);
    ClientY := PrinterDot(PrinterInfo.PaperSize.Y);
  end;
  Cr := Rect(
          Max(0, PrinterDot(MarginLeft) - Offset.Left),
          Max(0, PrinterDot(MarginTop) - Offset.Top),
          Min(ClientX, ClientX - PrinterDot(MarginRight) + Offset.Right),
          Min(ClientY, ClientY - PrinterDot(MarginBottom) + Offset.Bottom)
        );
  // Font Size -> Fh, Fw
  MetaFile := TMetaFile.Create;
  try
    MetaFile.Width := ClientX;
    MetaFile.Height := ClientY;
    MetaFileCanvas := TMetaFileCanvas.Create(MetaFile, 0);
    try
      MetaFileCanvas.Font.Assign(Font);
      MetaFileCanvas.Font.Size := Round(Font.Size / Ratio);
      GetTextMetrics(MetaFileCanvas.Handle, TM);
      // Font Height
      Fh := TM.tmHeight + TM.tmExternalLeading;
      // Font Width
      GetTextExtentPoint32W(MetaFileCanvas.Handle, '0', 1, p);
      Fw := p.cx;//TM.tmAveCharWidth;
      FAscent := TM.tmAscent + TM.tmExternalLeading;
      InitFontWidthMap(MetaFileCanvas);
      FontUtils.InitFontWidthMap(MetaFileCanvas.Font, FW, (tm.tmPitchAndFamily and 1) <> 1);

    finally
      MetaFileCanvas.Free;
    end;
  finally
    MetaFile.Free;
  end;
  // Header Rect
  if (HeaderAlign <> haNone) and (Header <> '') then
    Hr := Rect(Cr.Left, Cr.Top, Cr.Right, Cr.Top + Fh * 3 div 2)
  else
    Hr := Rect(Cr.Left, Cr.Top, Cr.Right, Cr.Top);
  // Footer Rect
  if (FooterAlign <> haNone) and (Footer <> '') then
    Fr := Rect(Cr.Left, Cr.Bottom - Fh * 3 div 2, Cr.Right, Cr.Bottom)
  else
    Fr := Rect(Cr.Left, Cr.Bottom, Cr.Right, Cr.Bottom);
  // Nr, Dr
  InitDrawRects;
  ColumnOffset := Nr.Right - Nr.Left + Dr.Right - Dr.Left +
                  PrinterDot(ColumnSpace);
  LineMargin := PrinterDot(MarginLine);
  Nrw := Nr;
  Drw := Dr;
  CurrentColumn := 1; // ݂̃J
  // FirstLineNumber
  FPrintLineNumber := FirstLineNumber - 1;
  CurrentLineNumber := -1;
  // FirstPageNumber
//  if RangeOption = roWhole then
//    FPageNumber := FirstPageNumber
//  else
//    FPageNumber := FirstPageNumber - StartPage + 1;
  FPageNumber := 1;
  // fsItalic, fsBold ȏꍇ͉E[̂̕h߂ɁAP]T
  // Drwl ϐ́A܂Ԃ̂߂ɎgB
  Drwl := Drw;
//  if HasExtraFontStyle then
//    Drwl.Right := Drwl.Right - Fw;{ }
    Form := TFormStringsPrinterProgress.Execute(Self, Mes);
    try
      NewMetaFileCanvas;
      try
        // Header
        DrawHeader(HeaderAlign, Header, Hr);
        AWrapCount := 0;
        parser := FNotePad.CurrentMode.NewParser(FNotePadString);
        try
          for I := 0 to FNotePadString.Count - 1 do
          begin
            Form.Progress(Mes, FNotePadString.Count - 1, I);
            if Form.CanNotProgress then
              Exit;
            if I = 0 then AWrapCount := 0 else
              if FOptions.LeftBarCountWrap or
                 (FNotePadString[I - 1].LineEnd = leLF) then
                inc(AWrapCount);
            parser.ParseLineColor(i);

            // s͈͎wւ̑Ή
            if (LineRangeOption = roLimited) and
               ((AWrapCount < StartLine - 1) or (LastLine - 1 < AWrapCount)) then
              Continue;
            // ̈̍ŉiɓB̏
            if Drw.Top + Fh + LineMargin > Fr.Top then
              if CurrentColumn < ColumnCount then
              begin
                // OffsetRect Nr, Dr
                OffsetRect(Nr, ColumnOffset, 0);
                Nrw := Nr;
                OffsetRect(Dr, ColumnOffset, 0);
                Drw := Dr;
                Inc(CurrentColumn);
              end
              else
              begin
                DrawHeader(FooterAlign, Footer, Fr);
                MetaFileCanvas.Free;
                NewMetaFileCanvas;
                Inc(FPageNumber);
                DrawHeader(HeaderAlign, Header, Hr);
                // Init Nr, Dr
                InitDrawRects;
                Nrw := Nr;
                Drw := Dr;
                CurrentColumn := 1;
              end;
            // sԍ
            if LineNumber and (CurrentLineNumber <> AWrapCount) then
            begin
              CurrentLineNumber := AWrapCount;
              Inc(FPrintLineNumber);
              Nrw.Left := Nr.Left;
              Nrw.Right := Nr.Right;
              AssignHeaderColor;
              // DT_RIGHT IvV𗘗p̂ DrawText gp
              DrawText(
                MetaFileCanvas.Handle,
                PChar(IntToStr(FPrintLineNumber) + ':'), -1, Nrw,
                DT_RIGHT or DT_SINGLELINE or DT_NOPREFIX);
            end;
            // `̈̏
            Drw.Left := Dr.Left;
            // `
            {sum := 0;
            ALine := FNotePadString[I];
            ASb := FNotePadString.InternalSB.ToPointer(ALine.Start);
            for j := 0 to ALine.Length - 1 do begin
              C := Asb + j;
              case C^ of
                #$0009: begin
                  inc(sum, FNotePadString.TabCount - (sum mod FNotePadString.TabCount));
                end;
                #$0020: inc(sum);
                #$3000: inc(sum, 2);
                else begin
                  if PrintMode in [pmBWFontStyle, pmColorToken] then begin
                    MetaFileCanvas.Font.Style := FNotePad.StyleFromToken(LineTokens[j]);
                  end else begin
                    MetaFileCanvas.Font.Style := [];
                  end;
                  if PrintMode in [pmColorToken] then begin
                    MetaFileCanvas.Font.Color := FNotePad.ColorFromToken(LineTokens[j]);
                  end else begin
                    MetaFileCanvas.Font.Color := clBlack;
                  end;
                  TextOutW(MetaFileCanvas.Handle, Drw.Left + (sum * Fw),
                    Drw.Top, C, 1);
                  inc(sum, GetGlyphWidth(C^));
                end;
              end;
            end;}DrawLine(I, MetaFileCanvas, Drw, Fw);
            // ̍s֕`̈ړ
            Inc(Drw.Top, Fh + LineMargin);
            Nrw.Top := Drw.Top;
          end; // for I
        finally
          parser.Free;
        end;
        DrawHeader(FooterAlign, Footer, Fr);
      finally
        MetaFileCanvas.Free;
      end;
      Result := True;
    finally
      Form.Free;
      ManaMain.MMRestoreOption;
    end;
end;

procedure TNotePadPrinter.DrawLine(I: integer; MetaFileCanvas: TMetafileCanvas; Drw: TRect; Fw: integer);
var
  j: integer;
  ASb: PWChar;
  sum,suminc: integer;
  C: PWChar;
  ALine: PLineData;
  AStyle: TFontStyles;
begin
  sum := 0;
  ALine := FNotePadString[I];
  ASb := FNotePadString.InternalSB.ToPointer(ALine.Start);
  for j := 0 to ALine.Length - 1 do begin
    C := Asb + j;
    case C^ of
      #$0009: begin
        suminc := FNotePadString.TabCount * Fw - (sum mod (FNotePadString.TabCount * Fw));
        if suminc < Fw then inc(suminc, FNotePadString.TabCount * Fw);
        inc(sum, suminc);
      end;
      #$0020: inc(sum, GetGlyphWidth(WideChar($0020),MetaFileCanvas.Handle));
      #$3000: inc(sum, GetGlyphWidth(WideChar($3000),MetaFileCanvas.Handle));
      else begin
        if PrintMode in [pmBWFontStyle, pmColorToken] then begin
          AStyle := FNotePad.StyleFromToken(LineTokens[j]);
          MetaFileCanvas.Font.Style := AStyle - [fsUnderline];
        end else begin
          AStyle := [];
          MetaFileCanvas.Font.Style := [];
        end;
        if PrintMode in [pmColorToken] then begin
          MetaFileCanvas.Font.Color := FNotePad.ColorFromToken(LineTokens[j]);
        end else begin
          MetaFileCanvas.Font.Color := clBlack;
        end;
        TextOutW(MetaFileCanvas.Handle, Drw.Left + sum,
          Drw.Top, C, 1);
        if fsUnderline in AStyle then begin
          MetaFileCanvas.Pen.Color := MetaFileCanvas.Font.Color;
          MetaFileCanvas.MoveTo(Drw.Left + sum, Drw.Top + FAscent {Abs(MetaFileCanvas.Font.Height) - 1});
          MetaFileCanvas.LineTo(Drw.Left + sum + GetGlyphWidth(C^,MetaFileCanvas.Handle), Drw.Top + FAscent {Abs(MetaFileCanvas.Font.Height) - 1});
        end;
        inc(sum, GetGlyphWidth(C^,MetaFileCanvas.Handle));
      end;
    end;
  end;
end;

function TNotePadPrinter.PreviewExecute: boolean;
var
  MetaFileList: TMetaFileList;
begin
  Result := False;
  MetaFileList := TMetaFileList.Create;
  try
    if not DivideStringsToMetaFiles(MetaFileList, PreviewMessage) then
      Exit;
    TFormStringsPrinterPreview.Execute(Self, MetaFileList);
    Result := True;
  finally
    MetaFileList.Free;
  end;
end;

function TNotePadPrinter.PrintExecute: boolean;
var
  MetaFileList: TMetaFileList;
  I, C: Integer;
  ANeedNew: boolean;

  function CurrentPageNumber(Index: Integer): Integer;
  begin
    // TMetaFileList ̂Ox[X Index 猻݂̃y[WԍԂ
    if FRangeOption = roWhole then
      Result := Index + FFirstPageNumber
    else
      Result := Index + FFirstPageNumber - FStartPage + 1;
  end;

begin
  Result := False;
  ANeedNew := false;
  MetaFileList := TMetaFileList.Create;
  try
    if not DivideStringsToMetaFiles(MetaFileList, PrintMessage) then
      Exit;
    Printer.BeginDoc;
    try
      for I := 0 to MetaFileList.Count - 1 do
      begin
        (*
           LZꍇɂ
          EroLimited  I + 1  StartPage 菬 LastPage 傫ꍇ
          ECurrentPageNumber P菬lԂꍇ
          EpoOdd  Odd(CurrentPageNumber) Ȕꍇ
          EpoEven  Odd(CurrentPageNumber) ^̏ꍇ
        *)
        C := CurrentPageNumber(I);
        if ((FRangeOption = roLimited) and ((I + 1 < FStartPage) or (I + 1 > FLastPage))) or
           (C < 1) then
          Continue;
        //̃y[WKvǂ́AƂ߂Ă킩ƂύX  
        if ANeedNew then Printer.NewPage;
        Printer.Canvas.StretchDraw(Rect(0, 0, Printer.PageWidth, Printer.PageHeight), MetaFileList.Metafiles[I]);
        ANeedNew := true;
        //if I < MetaFileList.Count - 1 then
        //  Printer.NewPage;
      end;
    finally
      Printer.EndDoc;
    end;
    Result := True;
  finally
    MetaFileList.Free;
  end;
end;

procedure TNotePadPrinter.SetLastLine(const Value: Integer);
begin
  if FLastLine <> Value then
    FLastLine := Min(Max(Value, 1), LastPageMax);
end;

procedure TNotePadPrinter.SetLastPage(const Value: Integer);
begin
  if FLastPage <> Value then
    FLastPage := Min(Max(Value, 1), LastPageMax);
end;

end.
