unit wsStream;

interface

uses windows, messages, classes, sysutils;

type
  TFileStreamW = class(THandleStream)
  public
    constructor Create(const FileName: widestring; Mode: Word); overload;
    constructor Create(const FileName: widestring; Mode: Word; Rights: Cardinal); overload;
    destructor Destroy; override;
  end;

implementation

uses rtlconsts;

{ TFileStreamW }

constructor TFileStreamW.Create(const FileName: widestring; Mode: Word);
begin
  Create(Filename, Mode, 0);
end;

function FileCreate(const FileName: widestring; Rights: Integer): Integer; 
begin
  Result := Integer(CreateFileW(PwChar(FileName), GENERIC_READ or GENERIC_WRITE,
    0, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0));
end;

function FileOpen(const FileName: widestring; Mode: LongWord): Integer; 
const
  AccessMode: array[0..2] of LongWord = (
    GENERIC_READ,
    GENERIC_WRITE,
    GENERIC_READ or GENERIC_WRITE);
  ShareMode: array[0..4] of LongWord = (
    0,
    0,
    FILE_SHARE_READ,
    FILE_SHARE_WRITE,
    FILE_SHARE_READ or FILE_SHARE_WRITE);
begin
  Result := -1;
  if ((Mode and 3) <= fmOpenReadWrite) and
    ((Mode and $F0) <= fmShareDenyNone) then
    Result := Integer(CreateFileW(PWChar(FileName), AccessMode[Mode and 3],
      ShareMode[(Mode and $F0) shr 4], nil, OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL, 0));
end;

constructor TFileStreamW.Create(const FileName: widestring; Mode: Word;
  Rights: Cardinal);
begin
  if Mode = fmCreate then
  begin
    inherited Create(FileCreate(FileName, Rights));
    if FHandle < 0 then
      raise EFCreateError.CreateResFmt(@SFCreateErrorEx, [ExpandFileName(FileName), SysErrorMessage(GetLastError)]);
  end
  else
  begin
    inherited Create(FileOpen(FileName, Mode));
    if FHandle < 0 then
      raise EFOpenError.CreateResFmt(@SFOpenErrorEx, [ExpandFileName(FileName), SysErrorMessage(GetLastError)]);
  end;
end;

destructor TFileStreamW.Destroy;
begin
  if FHandle >= 0 then FileClose(FHandle);
  inherited Destroy;
end;

end.
